/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <iostream>
#include <fstream>
#include <cassert>
#include <stack>
#include "adjacency.hpp"

namespace p {

// ****************************************************************************
// 
// ****************************************************************************
adjacency::adjacency()
	: n(0), e(0), ee(0), data(0), nlist(0)
{
}

adjacency::adjacency(int n_vert)
	: n(0), e(0), ee(0), data(0), nlist(0)
{
	resize(n_vert);
}

adjacency::adjacency(const adjacency &src)
	: n(0), e(0), ee(0), data(0), nlist(0)
{
	operator=(src);
}

adjacency::~adjacency()
{
	clear();
}

void adjacency::clear()
{
	if (!n)
		return;

	n = 0;
	delete[] data;
	
	if (nlist) {
		delete[] nlist[0];
		delete[] nlist;
	}
}


void adjacency::resize(int n_vert)
{
	clear();
	
	n = n_vert;
	e = n * (n - 1) / 2;
	ee = (e + PXX_ADJ_BITS - 1) / PXX_ADJ_BITS;
	
	data = new PXX_ADJ_ELEMENT[ee];
}

int  adjacency::vertices() const
{
	return(n);
}

int  adjacency::pairs() const
{
	return(e);
}

int  adjacency::offset(int i, int j) const
{
	if (j > i) {
		int t = i; i = j; j = t;
	}
	
	assert(j >= 0);
	assert(i <  n);
	assert(j != i);
	
	return((i * (i - 1)) / 2 + j);
}

adjacency &adjacency::operator=(const adjacency &src)
{
	resize(src.n);
	
	for (int l = 0; l < ee; l++)
		data[l] = src.data[l];
	
	return(*this);
}

void adjacency::set(bool value)
{
	PXX_ADJ_ELEMENT f = (value) ? (PXX_ADJ_ELEMENT)-1 : 0;
	
	for (int l = 0; l < ee; l++)
		data[l] = f;
}

void adjacency::set(int i, int j, bool value)
{
	if (i == j)
		return;
	
	int o = offset(i, j);
	int l = o / PXX_ADJ_BITS;
	PXX_ADJ_ELEMENT b = (PXX_ADJ_ELEMENT)1 << (o % PXX_ADJ_BITS);
	
	if (value)
		data[l] |= b;
	else
		data[l] &= ~b;
}

void adjacency::complement()
{
	for (int l = 0; l < ee; l++)
		data[l] ^= (PXX_ADJ_ELEMENT)-1;
}

void adjacency::complement(int i, int j)
{
	if (i == j)
		return;
	
	int o = offset(i, j);
	int l = o / PXX_ADJ_BITS;
	PXX_ADJ_ELEMENT b = (PXX_ADJ_ELEMENT)1 << (o % PXX_ADJ_BITS);
	
	data[l] ^= b;
}

bool adjacency::get(int i, int j) const
{
	if (i == j)
		return(false);
	
	int o = offset(i, j);
	int l = o / PXX_ADJ_BITS;
	PXX_ADJ_ELEMENT b = (PXX_ADJ_ELEMENT)1 << (o % PXX_ADJ_BITS);
	
	return((data[l] & b) != 0);
}

bool adjacency::operator() (int i, int j) const
{
	return(get(i, j));
}

// ****************************************************************************
// 
// ****************************************************************************
adjacency &adjacency::pack()
{
	if (n <= 0)
		return(*this);
	
	if (nlist) {
		delete[] nlist[0];
		delete[] nlist;
	}
	
	int nz = 0;
	
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			if (get(i, j))
				nz++;
		}
	}
	
	nlist = new int *[n];
	int *head = new int[n + nz];

	for (int i = 0; i < n; i++) {
		nlist[i] = head;
		int k = 0;
		
		for (int j = 0; j < n; j++) {
			if (get(i, j)) {
				head[1 + k] = j;
				k++;
			}
		}
		head[0] = k;
		head += k + 1;
	}
	
	return(*this);
}

int  adjacency::pack_degree(int i) const
{
	return(nlist[i][0]);
}

const int *adjacency::pack_adj(int i) const
{
	return(&nlist[i][1]);
}

adjacency::it::it()
	: degree(0), offset(0), j(-1), adj(0)
{
}

adjacency::it::it(const adjacency &a, int i)
	: offset(0)
{
	degree = a.pack_degree(i);
	adj = a.pack_adj(i);
	
	if (degree > 0)
		j = adj[0];
	else
		j = -1;
}

adjacency::it::it(const it &src)
	: degree(src.degree), offset(src.offset), j(src.j), adj(src.adj)
{
}

adjacency::it::~it()
{
}

int adjacency::it::operator*() const
{
	return(j);
}

int adjacency::it::operator++()
{
	offset++;

	if (offset < degree) {
		j = adj[offset];
	} else {
		j = -1;
	}
	
	return(j);
}

int adjacency::it::operator++(int)
{
	int r = j;
	operator++();
	return(r);
}

adjacency::it &adjacency::it::operator=(const adjacency::it &src)
{
	degree = src.degree;
	offset = src.offset;
	j = src.j;
	adj = src.adj;
	return(*this);
}

void adjacency::it::of(const adjacency &a, int i)
{
	offset = 0;
	degree = a.pack_degree(i);
	adj = a.pack_adj(i);
	
	if (degree > 0)
		j = adj[0];
	else
		j = -1;
}

// ****************************************************************************
// 
// ****************************************************************************
int adjacency::read_simple(const std::string &path)
{
	if (path == "-")
		return(read_simple(std::cin));
	
	std::ifstream f(path.c_str());
	
	if (!f) {
		std::cerr << path << ": Could not open file for reading"
			<< std::endl;
		return(-1);
	}

	return(read_simple(f));
}

// ****************************************************************************
// 
// ****************************************************************************
int adjacency::read_simple(std::istream &f)
{
	int n = 0;
	int v = 0;
	
	f >> n;
	
	resize(n);
	
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			if (!f) {
				std::cerr << "adjacency::read_simple(): "
					<< "Read error" << std::endl;
				return(-1);
			}
			f >> v;
			set(i, j, v);
		}
	}

	return(0);
}

// ****************************************************************************
// 
// ****************************************************************************
int adjacency::write_simple(const std::string &path) const
{
	if (path == "-")
		return(write_simple(std::cout));
	
	std::ofstream f(path.c_str());

	if (!f) {
		std::cerr << path << ": Could not open file for writing"
			<< std::endl;
		return(-1);
	}
		
	return(write_simple(f));
}

// ****************************************************************************
// 
// ****************************************************************************
int adjacency::write_simple(std::ostream &f) const
{
	f << n << std::endl;

	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			f << ((get(i, j)) ? " 1" : " 0");
		}
		f << std::endl;
		
		if (!f) {
			std::cerr << "adjacency::write_simple(): Write error"
				<< std::endl;
			return(-1);
		}
	}

	return(0);
}


// ****************************************************************************
// 
// ****************************************************************************
std::ostream &operator<<(std::ostream &os, const adjacency &a)
{
	for (int i = 0; i < a.vertices(); i++) {
		for (int j = 0; j < a.vertices(); j++) {
			os << ((a.get(i, j)) ? " 1" : " 0");
		}
		os << std::endl;
	}
	
	return(os);
}

// ****************************************************************************
// 
// ****************************************************************************
adjacency_pre::adjacency_pre(const adjacency &src)
	: a(src), ncomp(0), comp(src.vertices())
{
	int n = a.vertices();
	
	int unsorted = n;
	std::stack<int> front;
	std::fill(comp.begin(), comp.end(), -1);
	
	while (unsorted > 0) {
		int i = 0;
		while (comp[i] != -1)
			i++;
		
		comp[i] = ncomp;
		unsorted--;
		front.push(i);
		
		while(!front.empty()) {
			i = front.top();
			front.pop();
		
			for (adjacency::it j(a, i); *j != -1; j++) {
				if (comp[*j] == -1) {
					comp[*j] = ncomp;
					unsorted--;
					front.push(*j);
				}
			}
		}
		
		ncomp++;
	}
}

adjacency_pre::~adjacency_pre()
{
}

int adjacency_pre::components() const
{
	return(ncomp);
}

void adjacency_pre::build_component(int p,
	adjacency &d, std::vector<int> &orig_idx)
{
	int n = a.vertices();
	
	orig_idx.clear();
	
	for (int i = 0; i < n; i++) {
		if (comp[i] == p) {
			orig_idx.push_back(i);
		}
	}
	
	int q = orig_idx.size();
	
	d.resize(q);
	
	for (int i = 0; i < q; i++) {
		int oi = orig_idx[i];
		for (int j = 0; j < i; j++) {
			int oj = orig_idx[j];
			d.set(i, j, a.get(oi, oj));
		}
	}
	
	d.pack();
}

// ****************************************************************************
// 
// ****************************************************************************
} // namespace

