/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <vector>
#include "check.hpp"
#include "combinations.hpp"

namespace p {

// **************************************************************************
// 
// **************************************************************************
choose::choose()
	: k(0), m(0), ended(true)
{
}

choose::choose(int k0, int m0)
{
	init(k0, m0);
}

choose::~choose()
{
}

// **************************************************************************
// 
// **************************************************************************
void choose::init(int k0, int m0)
{
	k = k0;
	m = m0;
	idx.resize(k);

	ended = !((k >= 1) && (k <= m));
	
	for (int i = 0; i < k; i++)
		idx[i] = i;
}

bool choose::finished() const
{
	return(ended);
}

bool choose::pending() const
{
	return(!ended);
}

// **************************************************************************
// 
// **************************************************************************
bool choose::next()
{
	int p = k - 1;
	while ((p >= 0) && (idx[p] == m - k + p)) p--;
	
	if (p < 0) {
		ended = true;
		return(false);
	}

	int v = idx[p] + 1;
	for (int i = p; i < k; i++)
		idx[i] = v++;
	
	return(true);
}

// **************************************************************************
// 
// **************************************************************************
int choose::operator[] (int i) const
{
	return(idx[i]);
}

// **************************************************************************
// 
// **************************************************************************
permute::permute()
	: n(0), idx(), ended(true)
{
}

permute::permute(int n0)
	: n(n0), idx(n), ended(false)
{
	id();
}

permute::permute(int n0, int i, int j)
	: n(n0), idx(n), ended(false)
{
	idswap(i, j);
}

permute::permute(const permute &a)
	: n(a.n), idx(a.idx), ended(a.ended)
{
}

permute::~permute()
{
}

// **************************************************************************
// 
// **************************************************************************
void permute::id(int n0)
{
	n = n0;
	idx.resize(n);
	ended = false;
	
	for (int i = 0; i < n; i++)
		idx[i] = i;
}

void permute::id()
{
	ended = false;

	for (int i = 0; i < n; i++)
		idx[i] = i;
}

void permute::idswap(int i, int j)
{
	id();
	swap(i, j);
}

void permute::swap(int i, int j)
{
	int t = idx[i];
	idx[i] = idx[j];
	idx[j] = t;
}


bool permute::finished() const
{
	return(ended);
}

bool permute::pending() const
{
	return(!ended);
}

int permute::operator[] (int i) const
{
	//if ((i < 0) || (i >= n)) {
	//	PXX_ABORT("permute::operator[%d] out of bounds (size %d)\n",
	//		i, n);
	//}
	return(idx[i]);
}

int &permute::operator[] (int i)
{
	return(idx[i]);
}

permute &permute::operator=(const permute &a)
{
	n = a.n;
	idx = a.idx;
	ended = a.ended;
	
	return(*this);
}

// **************************************************************************
// 
// **************************************************************************
bool permute::next()
{
	bool f = false;
	int k;
	
	for (k = n - 2; k >= 0; k--) {
		if (idx[k] < idx[k + 1]) {
			f = true;
			break;
		}
	}
	
	if (!f) {
		ended = true;
		return(false);
	}
	
	int l = n - 1;
	while (idx[l] <= idx[k])
		l--;

	swap(k, l);
	
	int i = k + 1;
	int j = n - 1;
	while (i < j) {
		swap(i, j);
		i++;
		j--;
	}
	
	return(true);	
}

permute &permute::operator++()
{
	next();
	return(*this);
}

permute permute::operator++(int)
{
	permute r = *this;
	next();
	return(r);
}


// **************************************************************************
// 
// **************************************************************************

} // namespace
