/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cmath>
#include "tolerance.hpp"
#include "dense_vector.hpp"
#include "sparse_vector.hpp"
#include "constraint.hpp"

namespace p {

// *************************************************************************
// 
// *************************************************************************
constraint::constraint()
	: name(""), sense('E'), rhs(0.0)
{
}

constraint::constraint(const svec &srcrow, char srcsense, double srcrhs)
	: name(""), row(srcrow), sense(srcsense), rhs(srcrhs)
{
}

constraint::constraint(const constraint &src)
	: name(src.name), row(src.row), sense(src.sense), rhs(src.rhs)
{
}

constraint::~constraint()
{
}

constraint &constraint::operator=(const constraint &src)
{
	name = src.name;
	row = src.row;
	sense = src.sense;
	rhs = src.rhs;
	
	return(*this);
}

void constraint::set_name(const std::string &str)
{
	name = str;
}

void constraint::setup(const svec &srcrow, char srcsense, double srcrhs)
{
	row = srcrow;
	sense = srcsense;
	rhs = srcrhs;
}

// *************************************************************************
// 
// *************************************************************************
const char *constraint::cstr_name() const
{
	if (name.empty())
		return(NULL);
	
	return(name.c_str());
}

// *************************************************************************
// 
// *************************************************************************
double constraint::lhs(const dvec &x) const
{
	return(row * x);
}

double constraint::violation(const dvec &x) const
{
	return(-slack(x));
}

double constraint::slack(const dvec &x) const
{
	double v = lhs(x);
	
	switch(sense) {
	case 'G':
		return(v - rhs);
	case 'L':
		return(rhs - v);
	case 'E':
	default:
		return(fabs(v - rhs));
	}
}

double constraint::norm() const
{
	return(row.norm());
}

double constraint::dynamism(const tolerance &tol) const
{
	if (row.nz() == 0)
		return(1.0);
	
	double vmin, vmax;
	vmin = vmax = row.value(0);
	
	for (int l = 1; l < row.nz(); l++) {
		double v = fabs(row.value(l));
		
		if (tol.nonzero(v)) {
			vmin = std::min(vmin, v);
			vmax = std::max(vmax, v);
		}
	}
	
	// rhs does not count for vmin, as >= 0 has no scale
	// vmin = std::min(vmin, fabs(rhs));
	vmax = std::max(vmax, fabs(rhs));
	
	return(vmax / vmin);
}

// *************************************************************************
// 
// *************************************************************************
}

