/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cassert>
#include <cstring>
#include <cmath>
#include "check.hpp"
#include "strprintf.hpp"
#include "tolerance.hpp"
#include "cplex_model.hpp"


namespace p {

// *************************************************************************
// Handle CPLEX error code returns
// *************************************************************************
static int cpxret(CPXENVptr env,
	const char *file, int line, const char *func, const char *call, int err)
{
	if (err == 0)
		return(0);
	
	char buf[4096];
	const char *errstr;
	
	errstr = CPXgeterrorstring(env, err, buf);
	
	if (!errstr) {
		errstr = "Unknown error code.\n";
	} else if (errstr == buf) {
		for (size_t o = 0; o < sizeof(buf); o++) {
			if (buf[o] == 0)
				break;
			else if ((buf[o] == '\n') || (buf[o] == '\r'))
				buf[o] = ' ';
		}
	}
	
	printf("%s:%d: %s(): %s: CPLEX Error %5d.\n\t%s\n",
		file, line, func, call, err, errstr);
	
	return(err);
}

#define PXX_CPX_RET(c)	cpxret(env, __FILE__, __LINE__, __func__, #c, c)
#define PXX_CPX(c)	PXX_CHECK(!PXX_CPX_RET(c))

// *************************************************************************
// 
// *************************************************************************
cplex_env::cplex_env()
	: env(0), refcount(0)
{
}

cplex_env::~cplex_env()
{
	if (refcount != 0) {
		std::cerr << "cplex_env: WARNING: " << refcount
			<< " dangling references at destruction" << std::endl;
	}
}

void cplex_env::enter()
{
	refcount++;
	
	if (refcount > 1)
		return;
	
	int status;
	
	env = CPXopenCPLEX(&status);
	
	PXX_CPX(status);
	PXX_CHECK(env);
	
	setintparam(CPX_PARAM_SCRIND, CPX_OFF);
	// setintparam(CPX_PARAM_DATACHECK, CPX_ON);
	setintparam(CPX_PARAM_THREADS, 1);
}

void cplex_env::leave()
{
	refcount--;
	
	if (refcount != 0)
		return;
	
	CPXcloseCPLEX(&env);
}

CPXENVptr cplex_env::getptr() const
{
	return(env);
}

int cplex_env::getintparam(int which)
{
	int v;
	PXX_CPX(CPXgetintparam(env, which, &v));
	return(v);
}

double cplex_env::getdblparam(int which)
{
	double v;
	PXX_CPX(CPXgetdblparam(env, which, &v));
	return(v);
}

void cplex_env::setintparam(int which, int value)
{
	PXX_CPX(CPXsetintparam(env, which, value));
}

void cplex_env::setdblparam(int which, double value)
{
	PXX_CPX(CPXsetdblparam(env, which, value));
}

void cplex_env::setparam(int which, int value)
{
	setintparam(which, value);
}

void cplex_env::setparam(int which, double value)
{
	setdblparam(which, value);
}

void cplex_env::setlpcallbackfunc(
	int (CPXPUBLIC *callback) (CPXCENVptr, void *, int, void *),
	void *cbhandle)
{
	PXX_CPX(CPXsetlpcallbackfunc(env, callback, cbhandle));
}


void cplex_env::setinfocallbackfunc(
	int (CPXPUBLIC *callback) (CPXCENVptr, void *, int, void *),
	void *cbhandle)
{
	PXX_CPX(CPXsetinfocallbackfunc(env, callback, cbhandle));
}

void cplex_env::getcallbackinfo(void *cbdata, int wherefrom, int whichinfo,
	void *result_p)
{
	PXX_CPX(CPXgetcallbackinfo(env, cbdata, wherefrom, whichinfo,
		result_p));
}

int cplex_env::getcallbackinfo_failsafe(void *cbdata, int wherefrom,
	int whichinfo, void *result_p)
{
	return(CPXgetcallbackinfo(env, cbdata, wherefrom, whichinfo,
		result_p));
}

// *************************************************************************
// 
// *************************************************************************
cplex_env global_cplex_env;

// *************************************************************************
// 
// *************************************************************************
cplex_problem::cplex_problem(cplex_env &cpxenv)
	: env_handler(cpxenv)
{
	env_handler.enter();

	int status;
	
	env = env_handler.getptr();
	lp = CPXcreateprob(env, &status, "pxx");
	
	PXX_CPX(status);
	PXX_CHECK(lp);
}

cplex_problem::cplex_problem(CPXCLPptr src, cplex_env &cpxenv)
	: env_handler(cpxenv)
{
	env_handler.enter();
	
	int status;
	
	env = env_handler.getptr();
	lp = CPXcloneprob(env, src, &status);

	PXX_CPX(status);
	PXX_CHECK(lp);
}

cplex_problem::cplex_problem(const cplex_problem &src)
	: env_handler(src.env_handler)
{
	env_handler.enter();
	
	int status;
	
	env = env_handler.getptr();
	lp = CPXcloneprob(env, src.lp, &status);

	PXX_CPX(status);
	PXX_CHECK(lp);
}

cplex_problem::~cplex_problem()
{
	PXX_CPX(CPXfreeprob(env, &lp));
	
	env_handler.leave();
}

cplex_env &cplex_problem::getenv()
{
	return(env_handler);
}

void cplex_problem::clear()
{
	int status;
	
	PXX_CPX(CPXfreeprob(env, &lp));
	lp = CPXcreateprob(env, &status, "pxx");

	PXX_CPX(status);
	PXX_CHECK(lp);
}

// *************************************************************************
// low-level wrappers: access results
// *************************************************************************
void cplex_problem::getbase(int *cstat, int *rstat)
{
	PXX_CPX(CPXgetbase(env, lp, cstat, rstat));
}

void cplex_problem::getdj(double *dj, int begin, int end)
{
	PXX_CPX(CPXgetdj(env, lp, dj, begin, end));
}

int cplex_problem::getitcnt()
{
	return(CPXgetitcnt(env, lp));
}

double cplex_problem::getobjval()
{
	double v;
	PXX_CPX(CPXgetobjval(env, lp, &v));
	return(v);
}

void cplex_problem::getpi(double *pi, int begin, int end)
{
	PXX_CPX(CPXgetpi(env, lp, pi, begin, end));
}

void cplex_problem::getslack(double *slack, int begin, int end)
{
	PXX_CPX(CPXgetslack(env, lp, slack, begin, end));
}

int cplex_problem::getstat()
{
	return(CPXgetstat(env, lp));
}

void cplex_problem::getx(double *x, int begin, int end)
{
	PXX_CPX(CPXgetx(env, lp, x, begin, end));
}

// *************************************************************************
// low-level wrappers: advanced
// *************************************************************************
void cplex_problem::binvacol(int j, double *x)
{
	PXX_CPX(CPXbinvacol(env, lp, j, x));
}

void cplex_problem::binvarow(int i, double *z)
{
	PXX_CPX(CPXbinvarow(env, lp, i, z));
}

void cplex_problem::binvcol(int j, double *x)
{
	PXX_CPX(CPXbinvcol(env, lp, j, x));
}

void cplex_problem::binvrow(int i, double *z)
{
	PXX_CPX(CPXbinvrow(env, lp, i, z));
}

void cplex_problem::getbhead(int *head, double *x)
{
	PXX_CPX(CPXgetbhead(env, lp, head, x));
}

void cplex_problem::getray(double *ray)
{
	PXX_CPX(CPXgetray(env, lp, ray));
}

void cplex_problem::presolve(int method)
{
	PXX_CPX(CPXpresolve(env, lp, method));
}

CPXCLPptr cplex_problem::getredlp()
{
	CPXCLPptr rlp;
	
	PXX_CPX(CPXgetredlp(env, lp, &rlp));
	PXX_CHECK(rlp);
	
	return(rlp);
}

CPXCLPptr cplex_problem::getredlp_failsafe()
{
	CPXCLPptr rlp;
	
	PXX_CPX(CPXgetredlp(env, lp, &rlp));
	
	return(rlp);
}

CPXCLPptr cplex_problem::getredlp_or_self()
{
	CPXCLPptr rlp;
	
	PXX_CPX(CPXgetredlp(env, lp, &rlp));
	
	if (!rlp)
		return(lp);
	
	return(rlp);
}

double cplex_problem::getobjoffset()
{
	double objoffs;
	
	PXX_CPX(CPXgetobjoffset(env, lp, &objoffs));
	
	return(objoffs);
}

// *************************************************************************
// low-level wrappers: modify
// *************************************************************************
void cplex_problem::addrows(int ccnt, int rcnt, int nzcnt,
	const double *rhs, const char *sense,
	const int *rmatbeg, const int *rmatind, const double *rmatval,
	char **colname, char **rowname)
{
	PXX_CPX(CPXaddrows(env, lp, ccnt, rcnt, nzcnt, rhs, sense,
		rmatbeg, rmatind, rmatval, colname, rowname));
}

void cplex_problem::chgbds(int cnt, const int *indices, const char *lu,
	const double *bd)
{
	PXX_CPX(CPXchgbds(env, lp, cnt, indices, lu, bd));
}

void cplex_problem::cplex_problem::chgctype(int cnt, const int *indices,
	const char *xctype)
{
	PXX_CPX(CPXchgctype(env, lp, cnt, indices, xctype));
}

void cplex_problem::chgobj(int cnt, const int *indices, const double *values)
{
	PXX_CPX(CPXchgobj(env, lp, cnt, indices, values));
}

void cplex_problem::chgobjsen(int maxormin)
{
	CPXchgobjsen(env, lp, maxormin);
}

void cplex_problem::chgprobtype(int type)
{
	PXX_CPX(CPXchgprobtype(env, lp, type));
}

void cplex_problem::chgrhs(int cnt, const int *indices, const double *values)
{
	PXX_CPX(CPXchgrhs(env, lp, cnt, indices, values));
}

void cplex_problem::delrows(int begin, int end)
{
	PXX_CPX(CPXdelrows(env, lp, begin, end));
}

void cplex_problem::newcols(int ccnt, const double *obj,
	const double *lb, const double *ub,
	const char *xctype, char **colname)
{
	PXX_CPX(CPXnewcols(env, lp, ccnt, obj, lb, ub, xctype, colname));
}

void cplex_problem::copybase(int *cstat, int *rstat)
{
	PXX_CPX(CPXcopybase(env, lp, cstat, rstat));
}

// *************************************************************************
// low-level wrappers: optimizers
// *************************************************************************
void cplex_problem::primopt()
{
	PXX_CPX(CPXprimopt(env, lp));
}

void cplex_problem::dualopt()
{
	PXX_CPX(CPXdualopt(env, lp));
}

void cplex_problem::lpopt()
{
	PXX_CPX(CPXlpopt(env, lp));
}

int cplex_problem::lpopt_failsafe()
{
	return(PXX_CPX_RET(CPXlpopt(env, lp)));
}

void cplex_problem::mipopt()
{
	PXX_CPX(CPXmipopt(env, lp));
}

int  cplex_problem::mipopt_failsafe()
{
	return(PXX_CPX_RET(CPXmipopt(env, lp)));
}

void cplex_problem::baropt()
{
	PXX_CPX(CPXbaropt(env, lp));
}

// *************************************************************************
// low-level wrappers: query
// *************************************************************************
void cplex_problem::getcolname(char **name, char *namestore, int storespace,
	int *surplus_p, int begin, int end)
{
	int status;
	
	status = CPXgetcolname(env, lp, name, namestore,
		storespace, surplus_p, begin, end);
	
	if (status == CPXERR_NEGATIVE_SURPLUS)
		return;
	
	PXX_CPX(status);
}

void cplex_problem::getlb(double *lb, int begin, int end)
{
	PXX_CPX(CPXgetlb(env, lp, lb, begin, end));
}

int cplex_problem::getnumcols()
{
	return(CPXgetnumcols(env, lp));
}

int cplex_problem::getnumnz()
{
	return(CPXgetnumnz(env, lp));
}

int cplex_problem::getnumrows()
{
	return(CPXgetnumrows(env, lp));
}

void cplex_problem::getobj(double *obj, int begin, int end)
{
	PXX_CPX(CPXgetobj(env, lp, obj, begin, end));
}

int cplex_problem::getobjsen()
{
	return(CPXgetobjsen(env, lp));
}

void cplex_problem::getprobname(char *buf_str, int bufspace, int *surplus_p)
{
	PXX_CPX(CPXgetprobname(env, lp, buf_str, bufspace, surplus_p));
}

int cplex_problem::getprobtype()
{
	return(CPXgetprobtype(env, lp));
}

void cplex_problem::getrhs(double *rhs, int begin, int end)
{
	PXX_CPX(CPXgetrhs(env, lp, rhs, begin, end));
}

void cplex_problem::getrowname(char **name, char *namestore, int storespace,
	int *surplus_p, int begin, int end)
{
	int status = CPXgetrowname(env, lp,
		name, namestore, storespace, surplus_p, begin, end);
	
	if (status == CPXERR_NEGATIVE_SURPLUS)
		return;

	PXX_CPX(status);
}

void cplex_problem::getrows(int *nzcnt_p, int *rmatbeg, int *rmatind,
	double *rmatval, int rmatspace, int *surplus_p, int begin, int end)
{
	int status = CPXgetrows(env, lp, nzcnt_p, rmatbeg, rmatind,
		rmatval, rmatspace, surplus_p, begin, end);
	
	if (status == CPXERR_NEGATIVE_SURPLUS)
		return;

	PXX_CPX(status);
}

void cplex_problem::getrows_exact(int *nzcnt_p, int *rmatbeg, int *rmatind,
	double *rmatval, int rmatspace, int begin, int end)
{
	int CPXgetrows_surplus;
	
	getrows(nzcnt_p, rmatbeg, rmatind, rmatval, rmatspace,
		&CPXgetrows_surplus, begin, end);
	
	PXX_CHECK(CPXgetrows_surplus == 0);
}

void cplex_problem::getsense(char *sense, int begin, int end)
{
	PXX_CPX(CPXgetsense(env, lp, sense, begin, end));
}

void cplex_problem::getub(double *ub, int begin, int end)
{
	PXX_CPX(CPXgetub(env, lp, ub, begin, end));
}

void cplex_problem::getctype(char *xctype, int begin, int end)
{
	PXX_CPX(CPXgetctype(env, lp, xctype, begin, end));
}

// *************************************************************************
// low-level wrappers: solution pool
// *************************************************************************
int cplex_problem::readcopyprob(const char *name, const char *type)
{
	return(PXX_CPX_RET(CPXreadcopyprob(env, lp, name, type)));
}

// *************************************************************************
// low-level wrappers: solution pool
// *************************************************************************
int cplex_problem::getsolnpoolnumsolns()
{
	return(CPXgetsolnpoolnumsolns(env, lp));
}

double cplex_problem::getsolnpoolobjval(int soln)
{
	double v;
	PXX_CPX(CPXgetsolnpoolobjval(env, lp, soln, &v));
	return(v);
}

void cplex_problem::getsolnpoolx(int soln, double *x, int begin, int end)
{
	PXX_CPX(CPXgetsolnpoolx(env, lp, soln, x, begin, end));
}

void cplex_problem::populate()
{
	PXX_CPX(CPXpopulate(env, lp));
}

int cplex_problem::populate_failsafe()
{
	return(PXX_CPX_RET(CPXpopulate(env, lp)));
}

// *************************************************************************
// low-level wrappers: write files	
// *************************************************************************
void cplex_problem::writeprob(const char *name, const char *type)
{
	PXX_CPX(CPXwriteprob(env, lp, name, type));
}

#define CASE_STR(s)	case s: return(#s);
const char *cplex_problem::cplex_problem::stat_str(int status)
{
	switch(status) {
		CASE_STR(CPX_STAT_ABORT_DUAL_OBJ_LIM);
		CASE_STR(CPX_STAT_ABORT_IT_LIM);
		CASE_STR(CPX_STAT_ABORT_OBJ_LIM);
		CASE_STR(CPX_STAT_ABORT_PRIM_OBJ_LIM);
		CASE_STR(CPX_STAT_ABORT_TIME_LIM);
		CASE_STR(CPX_STAT_ABORT_USER);
		CASE_STR(CPX_STAT_CONFLICT_ABORT_CONTRADICTION);
		CASE_STR(CPX_STAT_CONFLICT_ABORT_IT_LIM);
		CASE_STR(CPX_STAT_CONFLICT_ABORT_MEM_LIM);
		CASE_STR(CPX_STAT_CONFLICT_ABORT_NODE_LIM);
		CASE_STR(CPX_STAT_CONFLICT_ABORT_OBJ_LIM);
		CASE_STR(CPX_STAT_CONFLICT_ABORT_TIME_LIM);
		CASE_STR(CPX_STAT_CONFLICT_ABORT_USER);
		CASE_STR(CPX_STAT_CONFLICT_FEASIBLE);
		CASE_STR(CPX_STAT_CONFLICT_MINIMAL);
		CASE_STR(CPX_STAT_FEASIBLE);
		CASE_STR(CPX_STAT_FEASIBLE_RELAXED_INF);
		CASE_STR(CPX_STAT_FEASIBLE_RELAXED_QUAD);
		CASE_STR(CPX_STAT_FEASIBLE_RELAXED_SUM);
		CASE_STR(CPX_STAT_INFEASIBLE);
		CASE_STR(CPX_STAT_INForUNBD);
		CASE_STR(CPX_STAT_NUM_BEST);
		CASE_STR(CPX_STAT_OPTIMAL);
		CASE_STR(CPX_STAT_OPTIMAL_FACE_UNBOUNDED);
		CASE_STR(CPX_STAT_OPTIMAL_INFEAS);
		CASE_STR(CPX_STAT_OPTIMAL_RELAXED_INF);
		CASE_STR(CPX_STAT_OPTIMAL_RELAXED_QUAD);
		CASE_STR(CPX_STAT_OPTIMAL_RELAXED_SUM);
		CASE_STR(CPX_STAT_UNBOUNDED);
		CASE_STR(CPXMIP_ABORT_FEAS);
		CASE_STR(CPXMIP_ABORT_INFEAS);
		CASE_STR(CPXMIP_ABORT_RELAXED);
		CASE_STR(CPXMIP_FAIL_FEAS);
		CASE_STR(CPXMIP_FAIL_FEAS_NO_TREE);
		CASE_STR(CPXMIP_FAIL_INFEAS);
		CASE_STR(CPXMIP_FAIL_INFEAS_NO_TREE);
		CASE_STR(CPXMIP_FEASIBLE);
		CASE_STR(CPXMIP_FEASIBLE_RELAXED_INF);
		CASE_STR(CPXMIP_FEASIBLE_RELAXED_QUAD);
		CASE_STR(CPXMIP_FEASIBLE_RELAXED_SUM);
		CASE_STR(CPXMIP_INFEASIBLE);
		CASE_STR(CPXMIP_INForUNBD);
		CASE_STR(CPXMIP_MEM_LIM_FEAS);
		CASE_STR(CPXMIP_MEM_LIM_INFEAS);
		CASE_STR(CPXMIP_NODE_LIM_FEAS);
		CASE_STR(CPXMIP_NODE_LIM_INFEAS);
		CASE_STR(CPXMIP_OPTIMAL);
		CASE_STR(CPXMIP_OPTIMAL_INFEAS);
		CASE_STR(CPXMIP_OPTIMAL_POPULATED);
		CASE_STR(CPXMIP_OPTIMAL_POPULATED_TOL);
		CASE_STR(CPXMIP_OPTIMAL_RELAXED_INF);
		CASE_STR(CPXMIP_OPTIMAL_RELAXED_QUAD);
		CASE_STR(CPXMIP_OPTIMAL_RELAXED_SUM);
		CASE_STR(CPXMIP_OPTIMAL_TOL);
		CASE_STR(CPXMIP_POPULATESOL_LIM);
		CASE_STR(CPXMIP_SOL_LIM);
		CASE_STR(CPXMIP_TIME_LIM_FEAS);
		CASE_STR(CPXMIP_TIME_LIM_INFEAS);
		CASE_STR(CPXMIP_UNBOUNDED);
		default:
			return("?");
	}
}

// *************************************************************************
// model: object handling
// *************************************************************************
cplex_model::cplex_model(cplex_env &cpxenv)
	: cplex_problem(cpxenv)
{
}

cplex_model::cplex_model(CPXCLPptr src, cplex_env &cpxenv)
	: cplex_problem(src, cpxenv)
{
}

cplex_model::cplex_model(const cplex_model &src)
	: cplex_problem(src)
{
	lower = src.lower;
	upper = src.upper;
	obj = src.obj;
	ctype = src.ctype;
	xnames = src.xnames;
}

cplex_model::~cplex_model()
{
	target.clear();
}

void cplex_model::clear()
{
	cplex_problem::clear();
	
	lower.clear();
	upper.clear();
	obj.clear();
	ctype.clear();
	xnames.clear();
	xstr.clear();
	
	target.clear();
}

// *************************************************************************
// model: columns: construction
// *************************************************************************
void cplex_model::prepare_columns(int n)
{
	lower.resize(n);
	upper.resize(n);
	obj.resize(n);
	ctype.resize(n);
	xnames.resize(n);
}

void cplex_model::commit_columns()
{
	int n = obj.size();
	assert(n == (int)lower.size());
	assert(n == (int)upper.size());
	assert(n == (int)ctype.size());
	assert(n == (int)xnames.size());
	
	bool is_mip = false;
	for (int j = 0; j < n; j++) {
		if (ctype[j] != 'C') {
			is_mip = true;
			break;
		}
	}
	
	newcols(n, &obj[0], &lower[0], &upper[0],
		(is_mip) ? &ctype[0] : NULL, &xnames[0]);
}

void cplex_model::fetch_columns()
{
	int n = getnumcols();
	
	prepare_columns(n);
	
	getlb(&lower[0], 0, n - 1);
	getub(&upper[0], 0, n - 1);
	getobj(&obj[0], 0, n - 1);
	
	if (getprobtype() == CPXPROB_MILP) {
		getctype(&ctype[0], 0, n - 1);
	} else {
		std::fill(ctype.begin(), ctype.end(), 'C');
	}
	
	xstr.resize(n);
	for (int j = 0; j < n; j++) {
		char *name, *namestore;
		int l, surplus;
		
		getcolname(&name, NULL, 0, &surplus, j, j);
		l = -surplus;
		
		namestore = new char[l];
		
		getcolname(&name, namestore, l, &surplus, j, j);
		PXX_CHECK(surplus >= 0);
		
		/*if (surplus > 0) {
			std::cout << "Warning: cplex says it needs " << l
				<< " bytes for '" << name
				<< "' then leaves " << surplus << " unused."
				<< std::endl;
		}*/
		
		xstr[j] = name;
		xnames[j] = (char *)xstr[j].c_str();
		
		delete[] namestore;
	}
}

// *************************************************************************
// model: columns: bounds and objective
// *************************************************************************
void cplex_model::set_objective(const dvec &c)
{
	int n = c.size();
	std::vector<int> idx(n);
	
	for (int j = 0; j < n; j++)
		idx[j] = j;
	
	chgobj(n, &idx[0], &c[0]);
}

void cplex_model::set_objective(int idx, double c)
{
	chgobj(1, &idx, &c);
}

void cplex_model::set_bounds(const dvec &lb, const dvec &ub)
{
	int n = lb.size();
	
	assert(n == (int)ub.size());
	
	std::vector<int> idx(n);
	std::vector<char> clb(n), cub(n);
	
	for (int j = 0; j < n; j++)
		idx[j] = j;
	std::fill(clb.begin(), clb.end(), 'L');
	std::fill(cub.begin(), cub.end(), 'U');
	
	chgbds(n, &idx[0], &clb[0], &lb[0]);
	chgbds(n, &idx[0], &cub[0], &ub[0]);
}

void cplex_model::set_lb(int idx, double lb)
{
	char clb = 'L';
	
	chgbds(1, &idx, &clb, &lb);
}

void cplex_model::set_ub(int idx, double ub)
{
	char cub = 'U';
	
	chgbds(1, &idx, &cub, &ub);
}

void cplex_model::set_bounds(int idx, double lb, double ub)
{
	set_lb(idx, lb);
	set_ub(idx, ub);
}

void cplex_model::set_ctype(const char *ctype)
{
	int n = this->ctype.size();
	std::vector<int> idx(n);
	
	for (int j = 0; j < n; j++)
		idx[j] = j;
	
	chgctype(n, &idx[0], ctype);
}

void cplex_model::set_ctype(int idx, char ctype)
{
	chgctype(1, &idx, &ctype);
}


// *************************************************************************
// model: columns: macro ops
// *************************************************************************
bool cplex_model::at_lb(const dvec &x, int idx)
{
	return(sparse_tol.inclusive_le(x[idx], lower[idx]));
}

bool cplex_model::at_ub(const dvec &x, int idx)
{
	return(sparse_tol.inclusive_ge(x[idx], upper[idx]));
}

bool cplex_model::at_bound(const dvec &x, int idx)
{
	return(at_lb(x, idx) || at_ub(x, idx));
}


void cplex_model::fix(int idx, double value)
{
	set_bounds(idx, value, value);
}

void cplex_model::unfix(int idx)
{
	set_bounds(idx, lower[idx], upper[idx]);
}

void cplex_model::unfix()
{
	set_bounds(lower, upper);
}

void cplex_model::make_LP()
{
	chgprobtype(CPXPROB_LP);
}

void cplex_model::make_MIP()
{
	set_ctype(&ctype[0]);
}

int cplex_model::lookup(const char *name) const
{
	for (int j = 0; j < (int)xnames.size(); j++) {
		if (!strcmp(name, xnames[j]))
			return(j);
	}
	
	PXX_ABORT("column name not found: %s\n", name);
	return(-1);
}

int cplex_model::lookup(const std::string &name) const
{
	return(lookup(name.c_str()));
}

// *************************************************************************
// model: rows: construction
// *************************************************************************
int cplex_model::add_constraint(const constraint &c)
{
	return(add_constraint(c.name, c.row, c.sense, c.rhs));
}

int cplex_model::add_constraint(
	const std::string &name, const svec &row, char sense, double rhs)
{
	int nz = row.nz();
	int rmatbeg = 0;
	std::vector<int> rmatind(nz);
	std::vector<double> rmatval(nz);

	row.get(&rmatind[0], &rmatval[0]);
	
	const char *rowname = ((name.empty()) ? NULL : name.c_str());
	const char **rownames = ((rowname) ? &rowname : NULL);
	
	addrows(0, 1, nz, &rhs, &sense,
		&rmatbeg, &rmatind[0], &rmatval[0], NULL, (char **)rownames);
	
	return(target.add());
}

int cplex_model::add_constraint(const char *rowname,
	int nz, int *idx, double *val, char sense, double rhs)
{
	int rmatbeg = 0;

	if ((rowname) && (rowname[0] == 0))
		rowname = NULL;
	
	addrows(0, 1, nz, &rhs, &sense,
		&rmatbeg, idx, val, NULL,
		(char **)((rowname) ? &rowname : NULL));
	
	return(target.add());
}


void cplex_model::remove_constraint(int idx)
{
	delrows(idx, idx);
	target.remove(idx);
}


// *************************************************************************
// model: rows: query constraints
// *************************************************************************
void cplex_model::get_constraint(int i, constraint &c)
{
	get_constraint(i, c.name, c.row, c.sense, c.rhs);
}

void cplex_model::get_constraint(int i, std::string &name,
	svec &row, char &sense, double &rhs)
{
	int l, surplus;

	// name
	char *cname, *namestore;
	
	getrowname(&cname, NULL, 0, &surplus, i, i);
	l = -surplus;
	namestore = new char[l];
	getrowname(&cname, namestore, l, &surplus, i, i);
	PXX_CHECK(surplus >= 0);
	/*if (surplus > 0) {
		std::cout << "Warning: cplex says it needs " << l
			<< " bytes for '" << cname
			<< "' then leaves " << surplus << " unused."
			<< std::endl;
	}*/
	name = cname;
	delete[] namestore;
	
	// row
	int matbeg, nz;
	int *matind;
	double *matval;
	
	getrows(&nz, &matbeg, NULL, NULL, 0, &surplus, i, i);
	l = -surplus;
	
	matind = new int[l];
	matval = new double[l];
	
	getrows(&nz, &matbeg, matind, matval, l, &surplus, i, i);
	PXX_CHECK((surplus == 0) && (nz == l));
	
	row.resize(getnumcols());
	row.set(l, matind, matval);
	getrhs(&rhs, i, i);
	getsense(&sense, i, i);
	
	delete[] matval;
	delete[] matind;
}

// *************************************************************************
// model: rows: cut pool
// *************************************************************************
void cplex_model::add_cut(cut &c)
{
	if (c.tracker.target)
		return;
	
	int i = add_constraint(c);
	target.set_tracker(i, &c.tracker);
}

void cplex_model::remove_cut(cut &c)
{
	if (c.tracker.target == &this->target)
		remove_constraint(c.tracker.index);
}

int cplex_model::add_all(cut_pool &p)
{
	int n_added = 0;
	
	for (std::list<cut>::iterator c = p.pool.begin();
	c != p.pool.end(); c++) {
		if (!c->tracker.target) {
			c->age = 0;
			add_cut(*c);
			n_added++;
		}
	}
	
	return(n_added);
}

void cplex_model::purge_all(cut_pool &p)
{
	for (std::list<cut>::iterator c = p.pool.begin();
	c != p.pool.end(); c++) {
		if (c->tracker.target == &this->target) {
			remove_cut(*c);
		}
	}
}

int cplex_model::add_purge(cut_pool &p, const dvec &x,
	double min_violation_add, double max_slack_keep,
	int max_age)
{
	int n_added = 0;
	
	for (std::list<cut>::iterator c = p.pool.begin();
	c != p.pool.end(); c++) {
		if ((c->tracker.target == &this->target) && (c->age >= 0)) {
			if (c->slack(x) <= max_slack_keep) {
				c->age = 0;
			} else {
				c->age++;
				if (c->age > max_age)
					remove_cut(*c);
			}
		} else if (!c->tracker.target) {
			if (c->violation(x) >= min_violation_add) {
				if (c->age >= 0)
					c->age = 0;
				add_cut(*c);
				n_added++;
			}
		}
	}
	
	return(n_added);
}

// *************************************************************************
// 
// *************************************************************************
void cplex_model::import_model(const model &q)
{
	// int m = q.constraints.size();
	int n = q.obj.size();
	
	clear();
	
	obj = q.obj;
	lower = q.lower;
	upper = q.upper;
	ctype = q.ctype;
	xstr = q.cname;
	
	xnames.resize(n);
	for (int j = 0; j < n; j++)
		xnames[j] = (char *)xstr[j].c_str();
	
	commit_columns();
	
	for (std::list<constraint>::const_iterator it = q.constraints.begin();
	it != q.constraints.end(); it++) {
		add_constraint(*it);
	}
}

void cplex_model::export_model(model &q)
{
	int m = getnumrows();
	
	fetch_columns();
	
	q.obj = obj;
	q.lower = lower;
	q.upper = upper;
	q.ctype = ctype;
	q.cname = xstr;
	
	q.constraints.clear();
	
	for (int i = 0; i < m; i++) {
		constraint &c = q.add_constraint();
		get_constraint(i, c);
	}
}

void cplex_model::export_tableau(model &q, const tolerance &tol)
{
	int m = getnumrows();
	int n = getnumcols();
	double *b = new double[m];
	double *rm = new double[m];
	double *dd = new double[n];

	fetch_columns();
	
	getrhs(b, 0, m - 1);
	
	q.obj.resize(n);
	getdj(&q.obj[0], 0, n - 1);
	
	q.lower = lower;
	q.upper = upper;
	q.ctype = ctype;
	q.cname = xstr;
	
	q.constraints.clear();
	
	for (int i = 0; i < m; i++) {
		constraint &c = q.add_constraint();
		
		format(c.name, "_tab%03d_", i);

		binvarow(i, dd);
		c.row.gather(n, dd, tol);
		c.sense = 'E';
		c.rhs = 0.0;

		binvrow(i, rm);
		for (int k = 0; k < m; k++)
			c.rhs += rm[k] * b[k];
	}
	
	delete[] b;
	delete[] rm;
	delete[] dd;
}


// *************************************************************************
// 
// *************************************************************************
int cplex_model::freenb_driveout(double max_step, int max_iter)
{
	//int old_preind = getenv().getintparam(CPX_PARAM_PREIND);
	int old_itlim = getenv().getintparam(CPX_PARAM_ITLIM);

	//getenv().setintparam(CPX_PARAM_PREIND, CPX_OFF);
	getenv().setintparam(CPX_PARAM_ITLIM, 200);

	int m = getnumrows();
	int n = getnumcols();
	dvec xb(m), aj(m);
	dvec lower(n), upper(n);
	std::vector<int> bhead(m);
	std::vector<int> rstat(m);
	std::vector<int> cstat(n);
	
	getlb(&lower[0], 0, n - 1);
	getub(&upper[0], 0, n - 1);
	
	int it = 0;
	
	while (it++ < max_iter) {
		getbhead(&bhead[0], &xb[0]);
		getbase(&cstat[0], &rstat[0]);

		std::cout << "\t\t---------- Driveout Iteration "
			<< it << std::endl;
		int ej = -1;
		int lj = -1;
		int dir = 0;
		
		for (int j = 0; j < n; j++)
		if (cstat[j] == CPX_FREE_SUPER) {
			std::cout << "\t\tx" << j << " free non-basic"
				<< std::endl;
			binvacol(j, &aj[0]);
			
			double minstep = CPX_INFBOUND;
			int mini = -1;
			
			for (int i = 0; i < m; i++)
			if (aj[i] != 0.0) {
				double lb, ub;
				int bj = bhead[i];
				
				if (bj >= 0) {
					lb = lower[bj];
					ub = upper[bj];
				} else {
					lb = 0.0;
					ub = CPX_INFBOUND;
				}
				
				double aij = fabs(aj[i]);
				
				if (lb > -CPX_INFBOUND) {
					double step = (xb[i] - lb) / aij;
			
					if (step < minstep) {
						minstep = step;
						mini = i;
						dir = -1;
					}
				}
				
				if (ub <  CPX_INFBOUND) {
					double step = (ub - xb[i]) / aij;
					
					if (step < minstep) {
						minstep = step;
						mini = i;
						dir = 1;
					}
				}
			}
			
			std::cout << "\t\t\tmin step length: " << minstep
				<< std::endl;
			
			if (minstep <= max_step) {
				ej = j;
				lj = bhead[mini];
				std::cout << "\t\t\tentering x" << ej;
				if (lj >= 0)
					std::cout << ", leaving x" << lj;
				else
					std::cout << ", leaving r" << -lj - 1;
				std::cout << " from " << xb[mini] << " to ";
				if (lj < 0)
					std::cout << "zero";
				else if (dir > 0)
					std::cout << upper[lj];
				else
					std::cout << lower[lj];
				
				std::cout << ", in row " << mini << std::endl;
				break;
			}
		}
		
		if (ej == -1) {
			std::cout << "\t\t---------- Driveout performed "
				<< it << " iterations" << std::endl;
			//getenv().setintparam(CPX_PARAM_PREIND, old_preind);
			getenv().setintparam(CPX_PARAM_ITLIM, old_itlim);
			return(0);
		}
		
		cstat[ej] = CPX_BASIC;
		
		if (lj >= 0)
			cstat[lj] = (dir > 0) ? CPX_AT_UPPER : CPX_AT_LOWER;
		else
			rstat[-lj - 1] = CPX_AT_LOWER;
		
		copybase(&cstat[0], &rstat[0]);
		
		//std::cout << "\t\t\tlpopt()..." << std::endl;
		primopt();
		//std::cout << stat_str(getstat()) << std::endl;
		//std::cout << "\t\t\tDone." << std::endl;
		
		int stat = getstat();
		if ((stat != CPX_STAT_OPTIMAL)
		&&  (stat != CPX_STAT_OPTIMAL_INFEAS)) {
			//std::cout << "\t\t\tlpopt() failed: "
			//	<< stat_str(stat) << std::endl;
			//getenv().setintparam(CPX_PARAM_PREIND, old_preind);
			getenv().setintparam(CPX_PARAM_ITLIM, old_itlim);
			return(-1);
		}
	}
	
	//getenv().setintparam(CPX_PARAM_PREIND, old_preind);
	getenv().setintparam(CPX_PARAM_ITLIM, old_itlim);
	return(-1);
}

// *************************************************************************
// 
// *************************************************************************
}
