/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef PXX_CPLEX_MODEL_HPP
#define PXX_CPLEX_MODEL_HPP
#include "ilcplex/cplex.h"
#include "tolerance.hpp"
#include "constraint.hpp"
#include "model.hpp"
#include "index_tracker.hpp"
#include "cut_pool.hpp"

namespace p {


// ************************************************
class cplex_env {

public:
	cplex_env();
	~cplex_env();

	void enter();
	void leave();
	CPXENVptr getptr() const;
	
	int getintparam(int which);
	double getdblparam(int which);
	void setintparam(int which, int value);
	void setdblparam(int which, double value);
	
	void setparam(int which, int value);
	void setparam(int which, double value);
	
	void setlpcallbackfunc(
		int (CPXPUBLIC *callback) (CPXCENVptr, void *, int, void *),
		void *cbhandle);
	void setinfocallbackfunc(
		int (CPXPUBLIC *callback) (CPXCENVptr, void *, int, void *),
		void *cbhandle);
	
	void getcallbackinfo(void *cbdata, int wherefrom, int whichinfo,
		void *result_p);
	int getcallbackinfo_failsafe(void *cbdata, int wherefrom,
		int whichinfo, void *result_p);
	
private:
	cplex_env(const cplex_env &);
	cplex_env &operator=(const cplex_env &);

private:
	CPXENVptr env;
	int refcount;
};

extern cplex_env global_cplex_env;

// ************************************************
class cplex_problem {

public:
	// object handling
	cplex_problem(cplex_env &cpxenv = global_cplex_env);
	cplex_problem(CPXCLPptr src, cplex_env &cpxenv = global_cplex_env);
	cplex_problem(const cplex_problem &src);
	~cplex_problem();
	
	cplex_env &getenv();
	void clear();
	
	// low-level wrappers: access results
	void getbase(int *cstat, int *rstat);
	void getdj(double *dj, int begin, int end);
	int getitcnt();
	double getobjval();
	void getpi(double *pi, int begin, int end);
	void getslack(double *slack, int begin, int end);
	int getstat();
	void getx(double *x, int begin, int end);
	
	// low-level wrappers: advanced
	void binvacol(int j, double *x);
	void binvarow(int i, double *z);
	void binvcol(int j, double *x);
	void binvrow(int i, double *z);
	void getbhead(int *head, double *x);
	void getray(double *ray);
	
	void presolve(int method = CPX_ALG_NONE);
	CPXCLPptr getredlp();
	CPXCLPptr getredlp_failsafe();
	CPXCLPptr getredlp_or_self();
	double getobjoffset();
	
	// low-level wrappers: modify
	void addrows(int ccnt, int rcnt, int nzcnt,
		const double *rhs, const char *sense,
		const int *rmatbeg, const int *rmatind, const double *rmatval,
		char **colname, char **rowname);
	void chgbds(int cnt, const int *indices, const char *lu,
		const double *bd);
	void chgctype(int cnt, const int *indices, const char *xctype);
	void chgobj(int cnt, const int *indices, const double *values);
	void chgobjsen(int maxormin);
	void chgprobtype(int type);
	void chgrhs(int cnt, const int *indices, const double *values);
	void delrows(int begin, int end);
	void newcols(int ccnt, const double *obj,
		const double *lb, const double *ub,
		const char *xctype, char **colname);

	void copybase(int *cstat, int *rstat);

	// low-level wrappers: optimizers
	void primopt();
	void dualopt();
	void lpopt();
	int lpopt_failsafe();
	void mipopt();
	int mipopt_failsafe();
	void baropt();

	// low-level wrappers: query general problem and query mip
	void getcolname(char **name, char *namestore, int storespace,
		int *surplus_p, int begin, int end);
	void getlb(double *lb, int begin, int end);
	int getnumcols();
	int getnumnz();
	int getnumrows();
	void getobj(double *obj, int begin, int end);
	int getobjsen();
	void getprobname(char *buf_str, int bufspace, int *surplus_p);
	int getprobtype();
	void getrhs(double *rhs, int begin, int end);	
	void getrowname(char **name, char *namestore, int storespace,
		int *surplus_p, int begin, int end);
	void getrows(int *nzcnt_p, int *rmatbeg, int *rmatind, double *rmatval,
		int rmatspace, int *surplus_p, int begin, int end);
	void getrows_exact(int *nzcnt_p, int *rmatbeg, int *rmatind,
		double *rmatval, int rmatspace, int begin, int end);
	void getsense(char *sense, int begin, int end);
	void getub(double *ub, int begin, int end);
	
	void getctype(char *xctype, int begin, int end);
	
	// low-level wrappers: read files
	int readcopyprob(const char *name, const char *type);
	
	// low-level wrappers: solution pool
	int getsolnpoolnumsolns();
	double getsolnpoolobjval(int soln);
	void getsolnpoolx(int soln, double *x, int begin, int end);
	void populate();
	int populate_failsafe();

	// low-level wrappers: write files	
	void writeprob(const char *name, const char *type);

	// stringify status returned by getstat()
	const char *stat_str(int status);

private:
	cplex_env &env_handler;
	CPXENVptr env;
	CPXLPptr lp;
};

// ************************************************
/*
    A cplex_model is a modifiable problem containing three things:
    - an actual CPXLP problem, on which modifications can be applied
    - some column information describing the original problem
    - some row information describing where constraints come from
    
    Available model alterations include:
    - changing bounds/objective/ctype and reverting to original formulation
    - adding/removing cuts with consistent tracking
*/
class cplex_model : public cplex_problem {

public:
	// object handling
	cplex_model(cplex_env &cpxenv = global_cplex_env);
	cplex_model(CPXCLPptr src, cplex_env &cpxenv = global_cplex_env);
	cplex_model(const cplex_model &src);
	~cplex_model();
	
	void clear();
	
	// columns: model construction (object data)
	void prepare_columns(int n);
	void commit_columns();
	void fetch_columns();

	// columns: bounds and objective function (parameter data)
	void set_objective(const dvec &c);
	void set_objective(int idx, double c);
	void set_bounds(const dvec &lb, const dvec &ub);
	void set_bounds(int idx, double lb, double ub);
	void set_lb(int idx, double lb);
	void set_ub(int idx, double ub);
	void set_ctype(const char *ctype);
	void set_ctype(int idx, char ctype);
	
	// columns: macro ops (object data)
	bool at_lb(const dvec &x, int idx);
	bool at_ub(const dvec &x, int idx);
	bool at_bound(const dvec &x, int idx);
	void fix(int idx, double value);
	void unfix(int idx);
	void unfix();
	void make_LP();
	void make_MIP();
	int lookup(const char *name) const;
	int lookup(const std::string &name) const;

	// rows: model construction
	int add_constraint(const constraint &c);
	int add_constraint(const std::string &name,
		const svec &row, char sense, double rhs);
	int add_constraint(const char *name,
		int nz, int *idx, double *val, char sense, double rhs);
	void remove_constraint(int idx);
	
	// rows: query model
	void get_constraint(int i, constraint &c);
	void get_constraint(int i, std::string &name,
		svec &row, char &sense, double &rhs);
	
	// rows: cut pool
	void add_cut(cut &c);
	void remove_cut(cut &c);
	
	int add_all(cut_pool &p);
	void purge_all(cut_pool &p);

	int add_purge(cut_pool &p, const dvec &x,
		double min_violation_add, double max_slack_keep,
		int max_age);

	// model
	void import_model(const model &q);
	void export_model(model &q);
	void export_tableau(model &q, const tolerance &tol = sparse_tol);

	// postprocessing
	int freenb_driveout(double max_step, int max_iter);

public:
	// user-visible fields
	dvec lower, upper;
	dvec obj;
	std::vector<char> ctype;
	std::vector<char *> xnames;

	// optional string storage for xnames
	std::vector<std::string> xstr;

	// row tracker target
	index_target target;
};


// ************************************************
} // namespace

#endif

