/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cmath>
#include "sparse_vector.hpp"
#include "dense_vector.hpp"

namespace p {

// *************************************************************************
// 
// *************************************************************************
dvec::dvec()
{
}

dvec::dvec(const dvec &src)
	: std::vector<double>(src)
{
}

dvec::dvec(int n)
	: std::vector<double>(n)
{
}

dvec::dvec(int n, const double *val)
{
	copy(n, val);
}

dvec::dvec(const svec &src)
{
	src.spread(*this);
}

dvec::dvec(const std::vector<double> &src)
	: std::vector<double>(src)
{
}

void dvec::copy(int n, const double *val)
{
	resize(n);
	std::copy(val, val + n, this->begin());
}

void dvec::copy(int n, double val)
{
	resize(n);
	set(val);
}

void dvec::set(double val)
{
	set(0, size() - 1, val);
}

void dvec::set(int i0, int i1, double val)
{
	for (int i = i0; i <= i1; i++)
		(*this)[i] = val;
}

void dvec::set(int i0, int i1, const dvec &src, int j0)
{
	for (int i = i0; i <= i1; i++, j0++)
		(*this)[i] = src[j0];
}

void dvec::set(int i0, int i1, const double *val, int j0)
{
	for (int i = i0; i <= i1; i++, j0++)
		(*this)[i] = val[j0];
}

void dvec::pad(int n, double val)
{
	int n0 = size();
	
	if (n0 >= n)
		return;
	
	resize(n);
	for (int i = n0; i < n; i++)
		(*this)[i] = val;
}

// *************************************************************************
// 
// *************************************************************************
double dvec::sqnorm() const
{
	int n = size();
	double v = 0.0;
	
	for (int i = 0; i < n; i++)
		v += (*this)[i] * (*this)[i];
	
	return(v);
}

double dvec::norm() const
{
	return(sqrt(sqnorm()));
}

double dvec::sum() const
{
	int n = size();
	double v = 0.0;
	
	for (int i = 0; i < n; i++)
		v += (*this)[i];
	
	return(v);
}


// *************************************************************************
// 
// *************************************************************************
dvec dvec::operator+(const dvec &src) const
{
	int n0 = std::min(size(), src.size());
	int n  = std::max(size(), src.size());

	dvec r(n);
	
	for (int i = 0; i < n0; i++)
		r[i] = (*this)[i] + src[i];
	
	for (int i = n0; i < (int)size(); i++)
		r[i] = (*this)[i];
		
	for (int i = n0; i < (int)src.size(); i++)
		r[i] = src[i];
	
	return(r);
}

dvec dvec::operator-(const dvec &src) const
{
	int n0 = std::min(size(), src.size());
	int n  = std::max(size(), src.size());

	dvec r(n);
	
	for (int i = 0; i < n0; i++)
		r[i] = (*this)[i] - src[i];
	
	for (int i = n0; i < (int)size(); i++)
		r[i] = (*this)[i];
		
	for (int i = n0; i < (int)src.size(); i++)
		r[i] = -src[i];
	
	return(r);
}

double dvec::operator*(const dvec &src) const
{
	int n = std::min(size(), src.size());
	double v = 0.0;
	
	for (int i = 0; i < n; i++)
		v += (*this)[i] * src[i];
	
	return(v);
}

double dvec::operator*(const svec &src) const
{
	return(src.operator*(*this));
}


dvec dvec::operator+() const
{
	return(*this);
}

dvec dvec::operator-() const
{
	dvec r(size());
	
	for (int i = 0; i < (int)size(); i++)
		r[i] = -(*this)[i];
	
	return(r);
}

dvec dvec::operator+(double scalar) const
{
	dvec r(size());
	
	for (int i = 0; i < (int)size(); i++)
		r[i] = (*this)[i] + scalar;
	
	return(r);
}

dvec dvec::operator-(double scalar) const
{
	dvec r(size());
	
	for (int i = 0; i < (int)size(); i++)
		r[i] = (*this)[i] - scalar;
	
	return(r);
}

dvec dvec::operator*(double scalar) const
{
	dvec r(size());
	
	for (int i = 0; i < (int)size(); i++)
		r[i] = (*this)[i] * scalar;
	
	return(r);
}

dvec dvec::operator/(double scalar) const
{
	dvec r(size());
	
	for (int i = 0; i < (int)size(); i++)
		r[i] = (*this)[i] / scalar;
	
	return(r);
}


// *************************************************************************
// 
// *************************************************************************
dvec &dvec::operator=(const dvec &src)
{
	std::vector<double>::operator=(src);
	return(*this);
}

dvec &dvec::operator=(const svec &src)
{
	src.spread(*this);
	return(*this);
}

dvec &dvec::operator+=(const dvec &src)
{
	int n0 = std::min(size(), src.size());
	
	for (int i = 0; i < n0; i++)
		(*this)[i] += src[i];

	if (n0 < (int)src.size()) {
		resize(src.size());

		for (int i = n0; i < (int)src.size(); i++)
			(*this)[i] = src[i];
	}
	
	return(*this);
}

dvec &dvec::operator-=(const dvec &src)
{
	int n0 = std::min(size(), src.size());
	
	for (int i = 0; i < n0; i++)
		(*this)[i] -= src[i];

	if (n0 < (int)src.size()) {
		resize(src.size());

		for (int i = n0; i < (int)src.size(); i++)
			(*this)[i] = -src[i];
	}

	return(*this);
}


dvec &dvec::operator+=(double scalar)
{
	for (int i = 0; i < (int)size(); i++)
		(*this)[i] += scalar;

	return(*this);
}

dvec &dvec::operator-=(double scalar)
{
	for (int i = 0; i < (int)size(); i++)
		(*this)[i] -= scalar;

	return(*this);
}

dvec &dvec::operator*=(double scalar)
{
	for (int i = 0; i < (int)size(); i++)
		(*this)[i] *= scalar;

	return(*this);
}

dvec &dvec::operator/=(double scalar)
{
	for (int i = 0; i < (int)size(); i++)
		(*this)[i] /= scalar;

	return(*this);
}

// *************************************************************************
// 
// *************************************************************************
std::ostream &operator<<(std::ostream &os, const dvec &v)
{
	for (int i = 0; i < (int)v.size(); i++)
		os << " " << v[i];
	
	return(os);
}

std::istream &operator>>(std::istream &is, dvec &v)
{
	for (int i = 0; i < (int)v.size(); i++)
		is >> v[i];
	
	return(is);
}

// *************************************************************************
// 
// *************************************************************************

} // namespace




