/*
    This file is part of libp -- linear algebra tools for optimization
    Copyright (C) 2010-2012  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libp.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "setup.h"
#include "gcd.h"
#include "rationals.h"
#include "diophantine.h"

#define DEBUG(s...)	fprintf(stdout, s)

// ***************************************************************************
// 
// ***************************************************************************
static int p_overflows_int(p_long a)
{
	p_long b = a >> (P_LONG_BITS - P_INT_BITS);
	return( (b != 0) && (b != -1) );
}

// ***************************************************************************
// 
// ***************************************************************************
int p_diophantine(p_int *x1, p_int *x2, p_int *s1, p_int *s2,
	p_int A, p_int B, p_int C)
{
	p_int gcd;
	p_int Ap, Bp, Cp, k1, k2;
	p_ulong lm;
	
	gcd = p_gcd(A, B);
	
	if (C % gcd)
		return(1);

	Ap = A / gcd;
	Bp = B / gcd;
	Cp = C / gcd;
	
	p_gcdext(&k1, &k2, Ap, Bp);
	
	if (x1) {
		lm = (p_long)k1 * (p_long)Cp;
		if (p_overflows_int(lm))
			return(-1);
		*x1 = lm;
	}
	
	if (x2) {
		lm = (p_long)k2 * (p_long)Cp;
		if (p_overflows_int(lm))
			return(-1);
		*x2 = lm;
	}
	
	if (s1)
		*s1 = Bp;
	if (s2)
		*s2 = -Ap;
	
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
p_int p_z2lp_segment_div(p_int v_r[2], const p_int a[2], const p_int b[2])
{
	p_int u[2];
	p_int gcd;
	
	u[0] = b[0] - a[0];
	u[1] = b[1] - a[1];
	
	gcd = p_gcd(u[0], u[1]);
	
	if (v_r) {
		v_r[0] = u[0] / gcd;
		v_r[1] = u[1] / gcd;
	}
	
	return(gcd);
}


p_int p_z2lp_inside_segment(p_int x_r[2], p_int y_r[2],
	const p_int a[2], const p_int b[2])
{
	p_int v[2];
	p_int seg;
	
	seg = p_z2lp_segment_div(v, a, b);
	
	if (x_r) {
		x_r[0] = a[0] + v[0];
		x_r[1] = a[1] + v[1];
	}
	
	if (y_r) {
		y_r[0] = b[0] - v[0];
		y_r[1] = b[1] - v[1];
	}
	
	return(seg - 1);
}

// ***************************************************************************
// 
// ***************************************************************************
int p_z2lp_colinear(p_int a[2], p_int b[2], p_int c[2])
{
	p_int u[2], v[2];
	p_long d;
	
	u[0] = b[0] - a[0];
	u[1] = b[1] - a[1];
	v[0] = c[0] - a[0];
	v[1] = c[1] - a[1];
		
	// d = det([u | v])
	d = ((p_long)u[0] * (p_long)v[1]) - ((p_long)u[1] * (p_long)v[0]);
	
	return(d == 0);
}

// ***************************************************************************
// 
// ***************************************************************************
/*
Here we have a compromise between
 - more accuracy (more bits assigned to rational precision)
 - less overflows (less bits assigned to rational precision)
The problem comes from triple multiplications of coefficients of rational
numbers. If (P_Z2LP_RZ_BITS <= P_INT_BITS / 3), then no overflow will occur.
But this leaves few bits to coefficients and leads to poor precision.
*/
#define P_Z2LP_RZ_BITS	(P_INT_BITS / 3)
#define P_Z2LP_RZ_BOUND	((1 << P_Z2LP_RZ_BITS) - 1)

int p_z2lp_on_ray(p_int x_r[2], p_int v_r[2],
	const double f_d[2], const double r_d[2])
{
	double f_frac[2];
	p_int f_floor[2];
	p_rational f[2], R;
	p_long alpha, beta, gamma1, gamma2, gamma;
	p_long ta, tb, jN, jD, j;
	
	p_int t, k0, k1, s0, s1;
	
	// Rationalize input data (without overflow checking)
	f_floor[0] = floor(f_d[0]);
	f_floor[1] = floor(f_d[1]);
	f_frac[0] = f_d[0] - f_floor[0];
	f_frac[1] = f_d[1] - f_floor[1];
	
	f[0] = p_rationalize(f_frac[0], P_Z2LP_RZ_BOUND);
	f[1] = p_rationalize(f_frac[1], P_Z2LP_RZ_BOUND);
	
	if (fabs(r_d[0]) > fabs(r_d[1])) {
		R = p_rationalize(r_d[0] / r_d[1], P_Z2LP_RZ_BOUND);
	} else {
		R = p_rationalize(r_d[1] / r_d[0], P_Z2LP_RZ_BOUND);
		t = R.num; R.num = R.den; R.den = t;
	}
	
	// Build equation (with overflow checking)
	ta = (p_long)f[1].den * (p_long)R.den;
	tb = (p_long)f[0].den * (p_long)R.num;
	alpha = (p_long)f[0].den * ta;
	beta = -(p_long)f[1].den * tb;
	gamma1 =  f[0].num * ta;
	gamma2 = -f[1].num * tb;
	gamma = gamma1 + gamma2;
	
	if ((p_overflows_int(ta)) || (p_overflows_int(tb))
	||  (p_overflows_int(alpha)) || (p_overflows_int(beta))
	||  (p_overflows_int(gamma1)) || (p_overflows_int(gamma2))
	||  (p_overflows_int(gamma)))
		return(0);
	
	// Solve the equation: alpha k0 + beta k1 = gamma
	if (p_diophantine(&k0, &k1, &s0, &s1, alpha, beta, gamma))
		return(0);
	
	// Give (s0, s1) the direction of r
	if (((s0 > 0) && (r_d[0] < 0.0))
	||  ((s0 < 0) && (r_d[0] > 0.0))
	||  ((s1 > 0) && (r_d[1] < 0.0))
	||  ((s1 < 0) && (r_d[1] > 0.0))) {
		s0 = -s0;
		s1 = -s1;
	}
	
	// Find first point touched by the ray
	if (s0 != 0) {
		jN = (p_long)f[0].num - (p_ulong)f[0].den * k0;
		jD = (p_long)s0 * f[0].den;
	} else {
		jN = (p_long)f[1].num - (p_ulong)f[1].den * k1;
		jD = (p_long)s1 * f[1].den;
	}
	
	// Ceil division: We know that the quotient is not integer, and that
	// in C99, divide rounds towards zero. Therefore, we just add 1 to
	// positive quotients.
	j = jN / jD;
	if ((jN > 0) == (jD > 0))
		j++;

	k0 += j * s0;
	k1 += j * s1;
	
	// Translate back to initial origin
	k0 += f_floor[0];
	k1 += f_floor[1];
	
	if (x_r) {
		x_r[0] = k0;
		x_r[1] = k1;
	}
	
	if (v_r) {
		v_r[0] = s0;
		v_r[1] = s1;
	}
	
	return(1);
}

// ***************************************************************************
// 
// ***************************************************************************
static p_int p_z2lp_inside_tri_gen(p_int u[2], p_int v[2], p_int d)
{
	p_int la[2], L[2], ja, J, k;
	p_int num, den;
	
	// Swap in order to try get a point with small coordinates
	/* if (u[0] + u[1] < v[0] + v[1]) {
		p_int t;
		
		// DEBUG("\t    Swapping u and v\n");
		t = u[0]; u[0] = v[0]; v[0] = t;
		t = u[1]; u[1] = v[1]; v[1] = t;
		d = -d;
	}*/
	
	if (p_diophantine(&la[0], NULL, &L[0], NULL, u[0], -d, -v[0])) {
		return(-1);
	}
	
	if (p_diophantine(&la[1], NULL, &L[1], NULL, u[1], -d, -v[1])) {
		return(-1);
	}
	
	if (p_diophantine(&ja, NULL, &J, NULL, L[0], -L[1], la[1] - la[0])) {
		return(-1);
	}
	
	// DEBUG("\t\tj1 = %d +  k %d\n", ja, J);
	// DEBUG("\t\tl1 = %d + j1 %d\n", la[0], L[0]);
	// DEBUG("\t\tl  = %d + (%d + k %d) %d\n", la[0], ja, J, L[0]);
	// DEBUG("\t\t   = %d + %d %d + k %d %d\n", la[0], ja, L[0], J, L[0]);
	// DEBUG("\t\t   = %d + k %d\n", la[0] + ja * L[0], J * L[0]);
	// DEBUG("\t\tx  = {l (%d %d) + (%d %d)} / |%d|\n",
	// 	u[0], u[1], v[0], v[1], d);

	num = -la[0] - L[0] * ja;
	den = J * L[0];
	
	// Floor division: We know that the quotient is not integer, and that
	// in C99, divide rounds towards zero. Therefore, we just substract 1
	// to negative quotients.
	k = num / den;
	
	if ((num > 0) != (den > 0))
		k--;
	
	// In case (J * L[0] > 0), we needed ceil instead of floor
	if (den > 0)
		k++;

	// DEBUG("\t\tk  = %s(%d / %d) = %d\n",
	// 	(den > 0) ? "ceil" : "floor", num, den, k);
	// DEBUG("\t\tl  = %d\n", la[0] + ja * L[0] + k * J * L[0]);
	
	return(la[0] + ja * L[0] + k * J * L[0]);
}

// ***************************************************************************
// 
// ***************************************************************************
p_int p_z2lp_inside_triangle(p_int x_r[2],
	const p_int a[2], const p_int b[2], const p_int c[2])
{
	const p_int *o;
	p_int u[2], v[2], w[2], x[2];
	p_int segAB, segAC, segBC, segU, segV, segX;
	p_long d, absd, n;
	p_int lambda;
	int f;
	
	// ***** find the number of lattice points:
	u[0] = b[0] - a[0]; u[1] = b[1] - a[1];
	v[0] = c[0] - a[0]; v[1] = c[1] - a[1];
	w[0] = c[0] - b[0]; w[1] = c[1] - b[1];
	segAB = p_gcd(u[0], u[1]);
	segAC = p_gcd(v[0], v[1]);
	segBC = p_gcd(w[0], w[1]);
			
	// d = det([u | v])
	d = ((p_long)u[0] * (p_long)v[1]) - ((p_long)u[1] * (p_long)v[0]);
	
	if (d > 0) {
		absd = d;
	} else if (d < 0) {
		absd = -d;
	} else {
		return(0);
	}

	n = (absd - segAB - segAC - segBC + 2) / 2;
	
	// DEBUG("\tTriangle (%d %d) (%d %d) (%d %d)\n",
	// 	a[0], a[1], b[0], b[1], c[0], c[1]);
	// DEBUG("\t  segs are AB (%d %d), AC (%d %d), BC (%d %d)\n",
	// 	u[0], u[1], v[0], v[1], w[0], w[1]);
	// DEBUG("\t  segs have gcd AB %d, AC %d, BC %d\n",
	// 	segAB, segAC, segBC);
	// DEBUG("\t  det %ld, %ld interior points\n", d, n);
	
	if (n == 0)
		return(0);
	
	if (n > P_INT_MAX)
		n = P_INT_MAX;
	
	if (!x_r)
		return(n);
	
	// ***** there is exactly one lattice points each edges
	//       ("extra-degenerate case"):
	if ((segAB == 2) && (segAC == 2) && (segBC == 2)) {
		u[0] = a[0] + (u[0] >> 1);
		u[1] = a[1] + (u[1] >> 1);
		v[0] = a[0] + (v[0] >> 1);
		v[1] = a[1] + (v[1] >> 1);
		
		return(p_z2lp_inside_triangle(x_r, a, u, v));
	}

	// ***** there are two (or more) lattice points on 2 or 3 edges
	//       ("degenerate case"):
	f = 0;
	
	if ((!f) && (segAB > 1) && (segAC > 1) && (segAB + segAC > 4)) {
		o = a;
		segU = segAB;
		segV = segAC;
		f = 1;
	}
	
	if ((!f) && (segAB > 1) && (segBC > 1) && (segAB + segBC > 4)) {
		o = b;
		segU = segAB;
		segV = segBC;
		u[0] = -u[0];
		u[1] = -u[1];
		v[0] = w[0];
		v[1] = w[1];
		f = 1;
	}
	
	if ((!f) && (segAC > 1) && (segBC > 1) && (segAC + segBC > 4)) {
		o = c;
		segU = segBC;
		segV = segAC;
		u[0] = -w[0];
		u[1] = -w[1];
		v[0] = -v[0];
		v[1] = -v[1];
		f = 1;
	}
	
	if (f) {
		x[0] = (u[0] / segU) + (v[0] / segV);
		x[1] = (u[1] / segU) + (v[1] / segV);
		segX = p_gcd(x[0], x[1]);
		x_r[0] = o[0] + (x[0] / segX);
		x_r[1] = o[1] + (x[1] / segX);
		return(n);
	}
	
	// ***** find suitable u and v not containing lattice points
	if (segAB > 1) {
		o = c;
		v[0] = -v[0];
		v[1] = -v[1];
		u[0] = -w[0];
		u[1] = -w[1];
		d = -d;
	} else if (segAC > 1) {
		o = b;
		u[0] = -u[0];
		u[1] = -u[1];
		v[0] = w[0];
		v[1] = w[1];
		d = -d;
	} else {
		o = a;
	}
	
	if (p_overflows_int(d)) {
		DEBUG("\t  Determinant overflows p_int\n");
		return(-1);
	}
	
	// ***** find a lattice point (general case):
	lambda = p_z2lp_inside_tri_gen(u, v, d);
	
	if (lambda <= 0)
		return(-1);
	
	// ***** verify
	x[0] = lambda * u[0] + v[0];
	x[1] = lambda * u[1] + v[1];
	if ((x[0] % absd != 0) || (x[1] % absd != 0)
	||  (lambda <= 0) || (lambda >= absd)) {
		// DEBUG("Returned point is not in the interior\n");
		return(-1);
	}
	x_r[0] = o[0] + x[0] / absd;
	x_r[1] = o[1] + x[1] / absd;
	
	return(n);
}


// ***************************************************************************
// Test for p_z2lp_on_ray()
// ***************************************************************************
/*
#define FRAC	8192

int main()
{
	double fd[2], rd[2];
	p_int x[2], v[2];
	int i;
	
	// srand(time(NULL));
	i = 0;
	while(1) {
		fd[0] = (double)(rand() % FRAC) / (double)FRAC;
		fd[1] = (double)(rand() % FRAC) / (double)FRAC;
		rd[0] = (double)(rand() % FRAC) / (double)FRAC;
		rd[1] = (double)(rand() % FRAC) / (double)FRAC;
	
		printf("data%d\n", i++);
		printf("f = [ %20.12g %20.12g ]\nr = [ %20.12g %20.12g ]\n",
			fd[0], fd[1], rd[0], rd[1]);
	
		if (!p_z2lp_on_ray(x, v, fd, rd)) {
			printf("No integer point on the line\n");
		} else {
			printf("Solution: [ %d %d ] + j [ %d %d ]\n",
				x[0], x[1],
				v[0], v[1]);
			return(0);
		}
	}
	
	return(0);
}
// */

// ***************************************************************************
// Test for p_z2lp_inside_triangle()
// ***************************************************************************
/*
#define M	1024
int main()
{
	p_int a[2] = { 0, 0 };
	p_int b[2] = { 1, 1 };
	p_int c[2] = { 9, 3 };
	p_int x[2];
	p_int r;
	int i;
	double dx, dy, dt = 0.0;
	
	for (i = 0; i < 100000; i++)
	{
		if (i > 0)
		{
			a[0] = rand() % M;
			a[1] = rand() % M;
			b[0] = rand() % M;
			b[1] = rand() % M;
			c[0] = rand() % M;
			c[1] = rand() % M;
		}

		printf("Triangle (%d %d) (%d %d) (%d %d)\n",
			a[0], a[1], b[0], b[1], c[0], c[1]);	
	
		r = p_z2lp_inside_triangle(x, a, b, c);
	
		if (r == 0) {
			printf("  lattice-free\n");
		} else if (r > 0) {
			printf("  has %d lattice points\n", r);
			printf("  x = ( %d, %d )\n", x[0], x[1]);
			dx = x[0] - a[0];
			dy = x[1] - a[1];
			dt += sqrt(dx * dx + dy * dy);
		} else {
			printf("  overflowed\n");
		}
	}
	
	printf("Average: %g\n", dt / (double)i);
	return(0);
}
// */

