/*
    This file is part of libp -- linear algebra tools for optimization
    Copyright (C) 2010-2012  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libp.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef __DIOPHANTINE_H
#define __DIOPHANTINE_H
#include "setup.h"

// **************************************************************************
//  Diophantine equations
// **************************************************************************

/*
  ***************************************************************************
  p_diophantine() solves the diophantine equation
    A x1 + B x2 = C
  It returns a particular solution x1* and x2* and the corresponding steps
  such that
  
  x1(j) = x1* + j s1
  x2(j) = x2* + j s2
  
  Return value:
     0 if a solution is found
     1 if no solution exists
    -1 if an overflow occurs
  ***************************************************************************
*/
int p_diophantine(p_int *x1, p_int *x2, p_int *s1, p_int *s2,
	p_int A, p_int B, p_int C);

// **************************************************************************
//  Closed-form Z^2 lattice points queries through diophantine systems
// **************************************************************************

/*
  ***************************************************************************
  p_z2lp_segment_div() reduces a line segment u given by two integer points
  A and B to its GCD.

  Return value:
    gcd(u1, u2)
  
  Return argument 'v' (can be NULL):
    v = u / gcd(u1, u2)
  ***************************************************************************
*/
p_int p_z2lp_segment_div(p_int v_r[2], const p_int a[2], const p_int b[2]);

/*
  ***************************************************************************
  p_z2lp_inside_segment() counts the number of lattice points inside a segment
  u defined by two integer vertices A and B, excluding these vertices.

  Return value:
    The number of lattice points
  
  Return arguments 'x' and 'y' (can be NULL):
    x = a + u / gcd(u1, u2)
    y = b - u / gcd(u1, u2)
  ***************************************************************************
*/
p_int p_z2lp_inside_segment(p_int x_r[2], p_int y_r[2],
	const p_int a[2], const p_int b[2]);

/*
  ***************************************************************************
  p_z2lp_colinear() returns nonzero if the three given points are colinear.
  ***************************************************************************
*/
int p_z2lp_colinear(p_int a[2], p_int b[2], p_int c[2]);

/*
  ***************************************************************************
  p_z2lp_on_ray() looks for lattice points on a rational line.
  The returned points are obviously guaranteed to be lattice points.
  However, given the potential overflows and inaccuracies in the
  double/rational conversion, they are not guaranteed to lie precisely
  or remotely on the specified line.
  
  Arguments f, r:
    'f' is a rational point of the line, 'r' is a direction vector.
  
  Return value:
    Nonzero if a lattice point is returned.
  
  Return arguments 'x', 'v':
    'x' is a particular solution and 'v' is the 'step' between the
    solutions. They are chosen such that
    - 'v' points in the same direction as 'r'
    - 'f' lies between 'x - v' and 'x', i.e. 'x' is the first lattice point
      encountered by the ray 'r' starting from the point 'f'.
  ***************************************************************************
*/
int p_z2lp_on_ray(p_int x_r[2], p_int v_r[2],
	const double f[2], const double r[2]);

/*
  ***************************************************************************
  p_z2lp_inside_triangle() finds lattice points in the interior of a triangle
  defined by three integer vertices.
  
  Return value:
   - Zero if the triangle is proven lattice-free
   - Negative if an overflow occurs
   - Positive if such lattice point is found
  
  Return argument 'x':
    A lattice point in the interior of the triangle. If possible, that point
    is such that the lattice formed by basis vectors (b - a) and (x - a) is
    unimodular.
  ***************************************************************************
*/
p_int p_z2lp_inside_triangle(p_int x_r[2],
	const p_int a[2], const p_int b[2], const p_int c[2]);



#endif


