#include <iostream>
#include <vector>
#include <queue>
#include <cmath>
#include "pxx/strprintf.hpp"
#include "pxx/timer.hpp"
#include "pxx/value.hpp"
#include "pxx/triangular.hpp"
#include "pxx/mipdb.hpp"
#include "pxx/cplex_model.hpp"
#include "pxx/onerow.hpp"
#include "pxx/tworow.hpp"

// ************************************************************************
// 
// ************************************************************************
#define DEBUG_ROWS
//#define DEBUG_MODELS
//#define DEBUG_CUT
//#define DEBUG_VIOLATION
//#define DEBUG_SPECIFIC_VALIDITY
//#define DEBUG_LOCAL_VALIDITY
//#define DEBUG_GLOBAL_VALIDITY

#if defined(DEBUG_SPECIFIC_VALIDITY) || defined(DEBUG_LOCAL_VALIDITY) \
	|| defined(DEBUG_GLOBAL_VALIDITY)
#define DEBUG_VIOLATION
#endif

// ************************************************************************
// Heuristic scoring for a pair of rows
// ************************************************************************
class model_score {
public:
	model_score();
	model_score(int row1, int row2, const p::svec &a, const p::svec &b);
	model_score(const model_score &b);
	int eval(const p::svec &a, const p::svec &b) const;
	bool operator<(const model_score &b) const;
	model_score &operator=(const model_score &b);

public:
	int i1, i2, score;
};

model_score::model_score()
	: i1(0), i2(0), score(0)
{
}

model_score::model_score(int row1, int row2, const p::svec &a, const p::svec &b)
	: i1(row1), i2(row2)
{
	score = eval(a, b);
}

model_score::model_score(const model_score &b)
	: i1(b.i1), i2(b.i2), score(b.score)
{
}

int model_score::eval(const p::svec &a, const p::svec &b) const
{
	int la = 0;
	int lb = 0;
	int common = 0;
	int diff = 0;
	
	while ((la < a.nz()) && (lb < b.nz())) {
		int ia = a.index(la);
		int ib = b.index(lb);
		
		if (ia == ib) {
			common++;
			la++;
			lb++;
		} else if (ia < ib) {
			la++;
		} else { // if (ia > ib)
			lb++;
		}
	}
	
	diff = a.nz() + b.nz() - 2 * common;
	
	return(common - diff);
}

bool model_score::operator<(const model_score &b) const
{
	if (score != b.score)
		return(score > b.score);
	
	if (i1 != b.i1)
		return(i1 < b.i1);

	if (i2 != b.i2)
		return(i2 < b.i2);
	
	return(false);
};

model_score &model_score::operator=(const model_score &b)
{
	i1 = b.i1;
	i2 = b.i2;
	score = b.score;
	return(*this);
}

// ************************************************************************
// Check rhs with a mip
// ************************************************************************
double best_rhs(p::cplex_model &initial, const p::svec &erow,
	p::dvec &x)
{
	int n = initial.getnumcols();
	
	initial.obj = erow;
	initial.obj.resize(n);
	initial.set_objective(initial.obj);
	initial.mipopt();

	int istat = initial.getstat();
	if ((istat != CPXMIP_OPTIMAL) && (istat != CPXMIP_OPTIMAL_INFEAS)
	&&  (istat != CPXMIP_OPTIMAL_TOL)) {
		std::cout << "best_rhs(): " << initial.stat_str(istat)
			<< std::endl;
		return(-CPX_INFBOUND);
	}
	
	x.resize(n);
	initial.getx(&x[0], 0, n - 1);
	
	return(initial.getobjval());
}

// ************************************************************************
// Loop for a given tableau
// ************************************************************************
static int tableau_iter(p::cplex_model &initial, p::cplex_model &qc,
	const p::elim_ruleset &el, const p::dvec &x_opt,
	p::model &tab, const p::mipdb_entry &info, int rank)
{
	// get basic info & config
	int m = tab.constraints.size();
	int n = tab.obj.size();

	int row_use_max = p::values["config"]["row_use_max"];
	int models_max = p::values["config"]["models_max"];
	int models_max_mem = p::values["config"]["models_max_mem"];
	double row_density_max = p::values["config"]["row_density_max"];
	int splits_enable = p::values["config"]["splits_enable"];
	
	// store row pointers for random access
	std::vector<p::constraint *> rowp(n);
	
	std::list<p::constraint>::iterator it = tab.constraints.begin();
	for (int i = 0; i < m; i++) {
		rowp[i] = &(*it);
		it++;
	}

	// retrieve solution and basis header
	p::dvec x0(n);
	p::dvec xB(m);
	std::vector<int> bhead(m);
	
	qc.getx(&x0[0], 0, n - 1);
	qc.getbhead(&bhead[0], &xB[0]);

	// select suitable rows
	std::vector<bool> suitint(m), suitfrac(m), suitsparse(m);
	int scount = 0, sfcount = 0, splcount = 0;
	
	for (int i = 0; i < m; i++) {
		double frac = fabs(xB[i] - floor(xB[i] + 0.5));

		suitsparse[i] = ((double)rowp[i]->row.nz() / (double)n
			<= row_density_max);
		
		if (bhead[i] < 0)
			suitint[i] = false;
		else
			suitint[i] = (tab.ctype[bhead[i]] != 'C');

		suitfrac[i] = (frac >= 2.0e-2);
		
		if (suitsparse[i] && suitint[i])
			scount++;
		if (suitsparse[i] && suitint[i] && suitfrac[i])
			sfcount++;
		if (suitint[i] && suitfrac[i])
			splcount++;
	}

	// select two-row models
	if (row_use_max > models_max)
		row_use_max = models_max;
	
	if (row_use_max > m)
		row_use_max = m;
	
	std::priority_queue<model_score> globq;

	if (row_use_max > 0)
	for (int i = 0; i < m; i++)
	if (suitsparse[i] && suitint[i]) {
		std::priority_queue<model_score> rowq;
		
		for (int j = i + 1; j < m; j++)
		if (suitsparse[j] && suitint[j]
		&&  (suitfrac[i] || suitfrac[j])) {
			model_score pair(i, j, rowp[i]->row, rowp[j]->row);
			
			if ((int)rowq.size() < row_use_max) {
				rowq.push(pair);
			} else if (pair < rowq.top()) {
				rowq.pop();
				rowq.push(pair);
			}
		}
		
		while (!rowq.empty()) {
			if ((int)globq.size() < models_max) {
				globq.push(rowq.top());
			} else if (rowq.top() < globq.top()) {
				globq.pop();
				globq.push(rowq.top());
			}
			
			rowq.pop();
		}
	}
	
	// collect results
	int nsep = globq.size();
	std::vector<model_score> sel(nsep);

	for (int k = nsep - 1; k >= 0; k--) {
		sel[k] = globq.top();
		globq.pop();
	}
	
	// build two-row models
	int nsep_mem = (nsep < models_max_mem) ? nsep : (models_max_mem + 1);

	p::values["timer"]["sep_total"].shadow_start();
	p::values["timer"]["sep_init"].start();
	
	p::tworow sep[nsep_mem];
	
	for (int k = 0; k < nsep_mem; k++) {
		sep[k].init(tab, x0, *rowp[sel[k].i1], *rowp[sel[k].i2],
			bhead[sel[k].i1], bhead[sel[k].i2]);
	}
	p::values["timer"]["sep_init"].stop();
	p::values["timer"]["sep_total"].shadow_stop();

	std::cout << p::strprintf("R%02d: ", rank)
		<< "Built " << nsep << " two-row models ("
		<< nsep_mem << " resident) on " << m << " rows ("
		<< scount << " suitable, " << sfcount << " fractional); "
		<< splcount << " one-row models"
		<< std::endl;
		
#ifdef DEBUG_ROWS
	std::cout << "model map:" << std::endl;
	for (int k = 0; k < nsep; k++) {
		int sel_i = sel[k].i1;
		int sel_j = sel[k].i2;
		int score = sel[k].score;
		
		std::cout << "(" << sel_i << ", " << sel_j << "): score " <<
			score << "; nz " << rowp[sel_i]->row.nz() << ", "
			<< rowp[sel_j]->row.nz() << "; densities "
			<< (double)rowp[sel_i]->row.nz() / (double)n << ", "
			<< (double)rowp[sel_j]->row.nz() / (double)n
			<< std::endl;
	}
#endif

	// state
	p::cut_pool pool;
	p::dvec xstar(x0);
	
	// separate splits
	p::onerow spl;
	int splcomp = 0;
	int splsep = 0;
	
	if (splits_enable)
	for (int i = 0; i < m; i++)
	if (suitint[i] && suitfrac[i]) {
		p::constraint c;
		std::string cname;
		
		splcomp++;
		
		p::format(cname, "R%02dS%03d", rank, i);
		
		p::values["timer"]["split"].start();
		int r = spl.separate(c, tab, x0, *rowp[i], bhead[i], cname);
		p::values["timer"]["split"].stop();
		
		if (r == 0) {
			p::values["splits"]["computed"]++;
#ifdef DEBUG_MODELS
				std::cout << rowp[i]->row << " = "
					<< rowp[i]->rhs << std::endl;
#endif
#ifdef DEBUG_CUT
			std::cout << c.row << " >= " << c.rhs << std::endl;
#endif
			el.eliminate(c);
#ifdef DEBUG_CUT
			std::cout << c.row << " >= " << c.rhs << std::endl;
#endif
			double fxv = c.violation(xstar);
#ifdef DEBUG_GLOBAL_VALIDITY
			p::dvec xmip;
			double erhs = best_rhs(initial, c.row, xmip);
#endif
#ifdef DEBUG_VIOLATION
			std::cout << "v=" << fxv << "; "
#endif
#ifdef DEBUG_SPECIFIC_VALIDITY
				<< "s=" << c.slack(x_opt) << "; "
#endif
#ifdef DEBUG_GLOBAL_VALIDITY
				<< "erhs=" << c.rhs
				<< " (max " << erhs << ")"
#endif
#ifdef DEBUG_VIOLATION
				<< std::endl;
#endif
#ifdef DEBUG_GLOBAL_VALIDITY
			if (c.rhs > erhs + 2.0e-5) {
				std::cout << c.name
					<< " is invalid_mip"
					<< std::endl;
				std::cout << "x" << bhead[i]
					<< " = " << xmip[bhead[i]]
					<< " (frac " << xstar[bhead[i]]
					<< ");" << std::endl;
				for (int l = 0; l < rowp[i]->row.nz(); l++) {
					if (xmip[rowp[i]->row.index(l)] != 0.0)
					std::cout
						<< xmip[rowp[i]->row.index(l)]
						<< "*" << rowp[i]->row.value(l)
						<< " ";
				}
				std::cout << " =? " << rowp[i]->rhs << std::endl;
				for (int l = 0; l < c.row.nz(); l++) {
					std::cout
						<< xmip[c.row.index(l)]
						<< "*" << c.row.value(l)
						<< " ";
				}
				std::cout << " >=? " << c.rhs << std::endl;
				exit(1);
			}
#endif

			if ((c.dynamism(p::tolerance(0.0)) <= 1e6)
			&&  (fxv >= 1e-6)) {
				p::values["splits"]["separating"]++;
				pool.add(c);
				splsep++;
			}
		}
	}
	
	// Reopt after splits
	if (splsep > 0) {
		qc.add_all(pool);
		p::values["timer"]["lp"].start();
		qc.lpopt();
		p::values["timer"]["lp"].stop();
	
		int stat = qc.getstat();
	
		if ((stat != CPX_STAT_OPTIMAL)
		&&  (stat != CPX_STAT_OPTIMAL_INFEAS)) {
			std::cout << "lpopt(): " << qc.stat_str(stat)
				<< std::endl;
			exit(1);
		}
	
		qc.getx(&xstar[0], 0, n - 1);
	}
	
	double obj = qc.getobjval();
	double gap = info.mip_value - info.lp_value;
	double gc = 100.0 * (obj - info.lp_value) / gap;
	
	std::cout << "Separating splits: " << splsep << " / " << splcomp
		<< "; obj: " << obj << " (" << gc << "%)" << std::endl;

	p::values["gc"][rank]["splits"] = gc;
	p::values["gc"] = gc;

	// separate two-rows in rounds
	int tworow_sep = 0;
	
	for (int round = 0; round < 15; round++) {
		int cutcomp = 0;
		int cutsep = 0;
		
		for (int k = 0; k < nsep; k++) {
			// late init
			int km = k;
			if (km >= models_max_mem) {
				km = models_max_mem;
				
				p::values["timer"]["sep_total"].shadow_start();
				p::values["timer"]["sep_init"].start();
				sep[km].clear();
				sep[km].init(tab, x0,
					*rowp[sel[k].i1], *rowp[sel[k].i2],
					bhead[sel[k].i1], bhead[sel[k].i2]);
				p::values["timer"]["sep_init"].stop();
				p::values["timer"]["sep_total"].shadow_stop();
			}
			
			// separation
			p::constraint c;
			std::string cname;
			
			p::format(cname, "R%02dr%02dm%02d", rank, round, k);
			
			p::values["timer"]["sep_total"].start();
			int r = sep[km].separate(c, xstar, cname);
			p::values["timer"]["sep_total"].stop();
			
			if (r == 0) {
				p::values["tworow_cuts"]["computed"]++;
				cutcomp++;
				
				// Checks:
				// Before elimination:
				// dbl rhs2 = sep[k].mip_rhs(c);
				// After elimination:
				// dbl erhs = best_rhs(initial, ce.row);
#ifdef DEBUG_MODELS
				std::cout << rowp[sel[k].i1]->row << " = "
					<< rowp[sel[k].i1]->rhs << std::endl;
				std::cout << rowp[sel[k].i2]->row << " = "
					<< rowp[sel[k].i2]->rhs << std::endl;
#endif
#ifdef DEBUG_LOCAL_VALIDITY
				double rhs = c.rhs;
				double rhs2 = sep[km].mip_rhs(c);
#endif
#ifdef DEBUG_CUT
				std::cout << c.row << " >= " << c.rhs << std::endl;
#endif
				el.eliminate(c);
#ifdef DEBUG_CUT
				std::cout << c.row << " >= " << c.rhs << std::endl;
#endif
				double fxv = c.violation(xstar);
#ifdef DEBUG_GLOBAL_VALIDITY
				p::dvec xmip;
				double erhs = best_rhs(initial, c.row, xmip);
#endif
#ifdef DEBUG_VIOLATION
				std::cout << "v=" << fxv << "; "
#endif
#ifdef DEBUG_SPECIFIC_VALIDITY
					<< "s=" << c.slack(x_opt) << "; "
#endif
#ifdef DEBUG_LOCAL_VALIDITY
					<< "rhs=" << rhs
					<< " (max " << rhs2 << "); "
#endif
#ifdef DEBUG_GLOBAL_VALIDITY
					<< "erhs=" << c.rhs
					<< " (max " << erhs << ")"
#endif
#ifdef DEBUG_VIOLATION
					<< std::endl;
#endif
				
#ifdef DEBUG_LOCAL_VALIDITY
				if (rhs > rhs2 + 2.0e-5) {
					std::cout << c.name
						<< " is invalid_2r" << std::endl;
					exit(1);
				}
#endif
				
#ifdef DEBUG_GLOBAL_VALIDITY
				if (c.rhs > erhs + 2.0e-5) {
					std::cout << c.name
						<< " is invalid_mip"
						<< std::endl;
					std::cout << "x" << bhead[sel[k].i1]
						<< " = "
						<< xmip[bhead[sel[k].i1]]
						<< "; x" << bhead[sel[k].i2]
						<< " = "
						<< xmip[bhead[sel[k].i2]]
						<< std::endl;
					for (int l = 0; l < c.row.nz(); l++) {
						std::cout
							<< xmip[c.row.index(l)]
							<< "*" << c.row.value(l)
							<< " ";
					}
					std::cout << std::endl;
					exit(1);
				}
#endif

				// Hackish way to keep only cuts on split sets
				if (p::values["config"]["only_split_sets"]) {
					int lfbt = sep[km].get_lfb_type();
					
					if (lfbt != PXX_TWOROW_SPLIT)
						fxv = -1.0;
				}
				
				bool used = false;
			
				if ((c.dynamism(p::tolerance(0.0)) <= 1e6)
				&&  (fxv >= 1e-6)) {
					p::values["tworow_cuts"]["separating"]++;
					pool.add(c);
					cutsep++;
					used = true;
				}

				int lvl = p::values["config"]["analyze_lfb"];
				if ((lvl >= 2) || ((lvl >= 1) && (used))) {
					int lfbt = sep[km].get_lfb_type();
					p::values["tworow"]["types"][lfbt]++;
				}
			}
		}
		
		// Reopt after two-rows
		if (cutsep > 0) {
			qc.add_purge(pool, xstar, 1e-6, 1e-3, 5);
			p::values["timer"]["lp"].start();
			qc.lpopt();
			p::values["timer"]["lp"].stop();
		
			int stat = qc.getstat();
		
			if ((stat != CPX_STAT_OPTIMAL)
			&&  (stat != CPX_STAT_OPTIMAL_INFEAS)) {
				std::cout << "lpopt(): " << qc.stat_str(stat)
					<< std::endl;
				exit(1);
			}
			//qc.writeprob("withcuts.lp", NULL);
		
			qc.getx(&xstar[0], 0, n - 1);
		}
		
		double obj = qc.getobjval();
		double gap = info.mip_value - info.lp_value;
		double gc = 100.0 * (obj - info.lp_value) / gap;
		
		std::cout << "Separating cuts: " << cutsep << " / " << cutcomp
			<< "; obj: " << obj << " (" << gc << "%)" << std::endl;

		p::values["gc"][rank]["all"] = gc;
		p::values["gc"] = gc;

		tworow_sep += cutsep;
		if (cutsep == 0)
			break;
	}
	
	// remove nontight cuts
	qc.add_purge(pool, xstar, 1e-6, 1e-3, 0);
	
	return(splsep + tworow_sep);
}

// ************************************************************************
// Optimize, get tableau, separate, and loop.
// ************************************************************************
static void main_loop(const std::string &fname)
{
	// read file
	p::cplex_model initial;
	
	if (initial.readcopyprob(fname.c_str(), NULL)) {
		std::cout << fname << ": failed to load problem" << std::endl;
		return;
	}

	// lookup mipdb
	p::mipdb_entry info(fname);
	
	if (info.known) {
		std::cout << "Detected " << info.name << " from "
			<< info.source << "; LP = " << info.lp_value
			<< ", MIP = " << info.mip_value << std::endl;
	} else {
		std::cout << "Problem " << info.name << " unknown" << std::endl;
	}
	
	p::values["problem"]["name"] = info.name;
	
	// Find a feasible solution
	p::cplex_model qc(initial);
	int n = qc.getnumcols();
	p::dvec x_opt(n);
	
#ifdef DEBUG_SPECIFIC_VALIDITY
	qc.getenv().setintparam(CPX_PARAM_NODELIM, 100);
	qc.mipopt();
	
	int stat = qc.getstat();
	
	if (stat == CPXMIP_NODE_LIM_INFEAS) {
		std::cout << fname << ": could not find solution" << std::endl;
		return;
	} else if (stat == CPXMIP_NODE_LIM_FEAS) {
		std::cout << fname << ": found feasible solution" << std::endl;
	} else if ((stat == CPXMIP_OPTIMAL) || (stat == CPXMIP_OPTIMAL_INFEAS)
	|| (stat == CPXMIP_OPTIMAL_TOL)) {
		std::cout << fname << ": found optimal solution" << std::endl;
	}
	
	qc.getx(&x_opt[0], 0, n - 1);
	
	qc.getenv().setintparam(CPX_PARAM_NODELIM, 2100000000);
#endif

	// Main loop
	int rank_max = p::values["config"]["rank_max"];
	
	p::values["timer"]["global"].start();
	p::elim_ruleset el;

	for (int rank = 1; rank <= rank_max; rank++) {
		// put cplex_model in standard form
		p::model std;
		
		p::values["timer"]["std"].start();
		qc.export_model(std);
		std.standard(el, PXX_STD_INT_SLACKS);
		qc.import_model(std);
		p::values["timer"]["std"].stop();
		// qc.writeprob("out.lp", NULL);
		
		// solve cplex_model
		p::values["timer"]["lp"].start();
		qc.make_LP();
		qc.getenv().setintparam(CPX_PARAM_ADVIND, 0);
		qc.lpopt();
		qc.getenv().setintparam(CPX_PARAM_ADVIND, 1);
		p::values["timer"]["lp"].stop();
		
		// consistency check
		if ((info.lp_known) && (rank == 1)) {
			double ic = fabs(qc.getobjval() - info.lp_value);
			double ric = (fabs(info.lp_value) < 0.1)
				? ic : ic / fabs(info.lp_value);

			if ((ic > 0.005) || (ric > 5e-6)) {
				std::cout << p::strprintf("LP value is %.12f, "
					"expected %.12f", qc.getobjval(),
					info.lp_value) << std::endl;
				std::cout << "delta = " << ic
					<< ", rdelta = " << ric << std::endl;
				exit(1);
			} else {
				info.lp_value = qc.getobjval();
				info.lp_known = true;
				
				if (!info.mip_known)
					info.mip_value = 1e+20;
			}
		}
		
		// fetch tableau
		p::model tab;
		
		p::values["timer"]["tableau"].start();
		qc.export_tableau(tab, p::tolerance(1e-9));
		tab.ctype = std.ctype;
		p::values["timer"]["tableau"].stop();

		int sep = tableau_iter(initial, qc, el, x_opt, tab, info, rank);

		qc.ctype = std.ctype;
		qc.make_MIP();
		
		if (sep == 0)
			break;
	}
	
	p::values["timer"]["global"].stop();
}	

// ************************************************************************
// 
// ************************************************************************
#define HELPMSG	\
"Usage: heusel.run [options] [problem]\n" \
"\n" \
"Options:\n" \
"  --row-use-max, --models-max, --models-max-mem, --row-density-max,\n" \
"  --splits-enable, --rank-max, --analyze-lfb, --only-split-sets\n"


// ************************************************************************
// 
// ************************************************************************
int main(int argc, char **argv)
{
	p::values["config"]["row_use_max"] = 4;
	p::values["config"]["models_max"] = 5000;
	p::values["config"]["models_max_mem"] = 500;
	p::values["config"]["row_density_max"] = 0.4;
	p::values["config"]["splits_enable"] = 1;
	p::values["config"]["rank_max"] = 5;
	p::values["config"]["analyze_lfb"] = 0;
	p::values["config"]["only_split_sets"] = 0;
	
	p::values["config"].parse_cmdline(argc, argv);
	
	if (p::values["config"] != 1) {
		std::cerr << HELPMSG << std::endl;
		return(0);
	}
	
	std::string fname = p::values["config"][0];
	
	main_loop(fname);

	std::cout << p::values << std::endl;
	
	std::cout
		<< p::values["problem"]["name"]
		<< p::strprintf(" %.3f", p::values["timer"]["global"].total_s())
		<< " " << p::values["timer"]["sep_total"].calls()
		<< " " << p::values["tworow"]["add"].calls()
		<< " " << p::values["tworow"]["iter_max"]
		<< p::strprintf(" %.3f",
			p::values["timer"]["sep_total"].total_s())
		<< " " << p::values["tworow_cuts"]["computed"]
		<< " " << p::values["tworow_cuts"]["separating"]
		<< " " << p::values["timer"]["split"].calls()
		<< p::strprintf(" %.3f", p::values["timer"]["split"].total_s())
		<< " " << p::values["splits"]["separating"]
		<< p::strprintf(" %.2f", p::values["gc"].get_double());
	for (int rank = 1; rank <= p::values["config"]["rank_max"]; rank++) {
		std::cout << p::strprintf(" %.2f %.2f",
			p::values["gc"][rank]["splits"].get_double(),
			p::values["gc"][rank]["all"].get_double());
	}
	
	for (int lfbt = 0; lfbt <= 6; lfbt++)
		std::cout << " " << p::values["tworow"]["types"][lfbt];
	
	std::cout << std::endl;
	return(0);
}

