#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <queue>
#include <cmath>
#include <cstdlib>
#include "pxx/integers.hpp"
#include "pxx/strprintf.hpp"
#include "pxx/timer.hpp"
#include "pxx/value.hpp"
#include "pxx/triangular.hpp"
#include "pxx/mipdb.hpp"
#include "pxx/cplex_model.hpp"
#include "pxx/onerow.hpp"
#include "pxx/tworow.hpp"

// ************************************************************************
// 
// ************************************************************************
#define HELPMSG	\
"Usage: maxlfs.run [options] [problem]\n" \
"\n" \
"Options:\n" \
"  none yet\n"

// ************************************************************************
// 
// ************************************************************************
/*
	[ r0_0 | r1_0 ]  ( s_0 ) = ( x_0 )
	[ r0_1 | r1_1 ]  ( s_1 ) = ( x_1 )
	<=>
	( r0_0 s_0 + r1_0 s_1 ) = ( x_0 )
	( r0_1 s_0 + r1_1 s_1 ) = ( x_1 )
	<=>
	r0 s_0 + r1 s_1 = x
	<=>
	          1/            [  r1_1  -r1_0 ]  ( x_0 ) = ( s_0 )
	r0_0 r1_1 - r0_1 r1_0   [ -r0_1   r0_0 ]  ( x_1 ) = ( s_1 )
*/

#define CONE_OK			0
#define CONE_DEGENERATE		1
int cone_s(p::dvec &s, const p::dvec &r0, const p::dvec &r1, const p::dvec &x)
{
	double epsilon = 1e-6;

	double det = r0[0] * r1[1] - r0[1] * r1[0];
	
	if ((det >= -epsilon) && (det <= epsilon)) {
		s[0] = 0.0;
		s[1] = 0.0;
		return(CONE_DEGENERATE);
	}
	
	p::dvec k0(2), k1(2);
	
	k0[0] =  r1[1] / det;
	k0[1] = -r0[1] / det;
	k1[0] = -r1[0] / det;
	k1[1] =  r0[0] / det;
	
	s = k0 * x[0] + k1 * x[1];
	
	return(CONE_OK);
}
// ************************************************************************
// 
// ************************************************************************
int merge_xy(p::dvec &z, const p::dvec &a, const p::dvec &x,
	const p::dvec &b, const p::dvec &y)
{
	double epsilon = 1e-6;
	
	p::dvec r0 = x - a;
	p::dvec r1 = y - b;
	p::dvec w = b - a;
	
	double det = r0[0] * r1[1] - r0[1] * r1[0];
	
	if ((det >= -epsilon) && (det <= epsilon)) {
		std::cerr << " det = " << det << std::endl;
		return(-1);
	}
	
	p::dvec k0(2), k1(2);
	
	k0[0] =  r1[1] / det;
	k0[1] = -r0[1] / det;
	k1[0] = -r1[0] / det;
	k1[1] =  r0[0] / det;
	
	p::dvec l = k0 * w[0] + k1 * w[1];
	
	if ((l[0] <= epsilon) || (l[1] >= -epsilon)) {
		std::cerr << " ab is " << l << std::endl;
		return(-1);
	}
	
	z = a + r0 * l[0];
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
class icside {

public:
	double a0;
	p::dvec v0, v1;
	p::dvec t;
	p::dvec d;
	p::dvec o;

	bool operator<(const icside &s) const;
};

bool icside::operator<(const icside &s) const
{
	return(a0 < s.a0);
}

// ************************************************************************
// 
// ************************************************************************
class icmodel {

public:
	icmodel();
	~icmodel();

	void reset();
	int read(std::istream &ifs);
	int write(std::ostream &ofs);
	void dump();
	void build();
	void consider_point(p::dvec y, int level);
	int extend(p::value &stats);
	int extend_quad(int j0);
	int extend_tri(int j0);
	int check_det_sub(p::value &stats, int j0, int j1, int j2);
	int check_det(p::value &stats);
	int check_o(p::value &stats);
	int check_lf(p::value &stats);
	int store();
public:
	std::string name;
	p::dvec f;
	std::vector<p::dvec> r;
	std::vector<p::dvec> v;
	std::vector<p::dvec> t;
	
	std::vector<icside> side;
};

// ************************************************************************
// 
// ************************************************************************
icmodel::icmodel()
{
}

icmodel::~icmodel()
{
}

void icmodel::reset()
{
	f.clear();
	r.clear();
	v.clear();
	t.clear();
	side.clear();
}

// ************************************************************************
// 
// ************************************************************************
void icmodel::dump()
{
	std::cerr << "f = [" << f << " ]';" << std::endl;
	
	for (size_t j = 0; j < v.size(); j++) {
		std::cerr << "v" << j << " = [" << v[j] << " ]';" << std::endl;
	}
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::write(std::ostream &ofs)
{
	ofs << name << " f=" << f << " nr= " << r.size() << " r=";
	
	for (size_t j = 0; j < r.size(); j++)
		ofs << r[j];
	
	ofs << " nv= " << v.size() << " v=";

	for (size_t j = 0; j < v.size(); j++)
		ofs << v[j];
	
	ofs << " nt= " << t.size() << " t=";

	for (size_t j = 0; j < t.size(); j++)
		ofs << t[j];
	
	ofs << std::endl;
	
	return(!ofs.good());
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::read(std::istream &ifs)
{
	reset();
	
	std::string tk;
	
	// Name
	ifs >> name;
	if (!ifs.good())
		return(0);
	
	// f
	ifs >> tk;
	if (tk != "f=")
		return(-1);
	
	f.resize(2);
	ifs >> f[0];
	ifs >> f[1];
	
	// r
	ifs >> tk;
	if (tk != "nr=")
		return(-1);

	int nr;
	ifs >> nr;
	
	ifs >> tk;
	if (tk != "r=")
		return(-1);
	
	r.resize(nr);
	
	for (int j = 0; j < nr; j++) {
		r[j].resize(2);
		ifs >> r[j][0];
		ifs >> r[j][1];
	}
	
	// v
	ifs >> tk;
	if (tk != "nv=")
		return(-1);

	int nv;
	ifs >> nv;
	
	ifs >> tk;
	if (tk != "v=")
		return(-1);
	
	v.resize(nv);
	
	for (int j = 0; j < nv; j++) {
		v[j].resize(2);
		ifs >> v[j][0];
		ifs >> v[j][1];
	}
	
	// t
	ifs >> tk;
	if (tk != "nt=")
		return(-1);

	int nt;
	ifs >> nt;
	
	ifs >> tk;
	if (tk != "t=")
		return(-1);
	
	t.resize(nt);
	
	for (int j = 0; j < nt; j++) {
		t[j].resize(2);
		ifs >> t[j][0];
		ifs >> t[j][1];
	}
	
	return(1);
}

// ************************************************************************
// 
// ************************************************************************
void icmodel::build()
{
	side.clear();
	side.resize(v.size());
	
	for (size_t j = 0; j < side.size(); j++) {
		p::dvec r = v[j] - f;
		side[j].v0 = v[j];
		side[j].a0 = atan2(r[1], r[0]);
	}
	
	std::sort(side.begin(), side.end());
	
	for (size_t j = 0; j < side.size(); j++) {
		size_t u = (j + 1) % side.size();
		side[j].v1 = side[u].v0;
	}
	
	for (size_t i = 0; i < t.size(); i++) {
		consider_point(t[i], 0);
	}
}

// ************************************************************************
// 
// ************************************************************************
void icmodel::consider_point(p::dvec y, int level)
{
	int incones = 0;
	
	for (size_t j = 0; j < side.size(); j++) {
		p::dvec s;
		int c = cone_s(s, side[j].v0 - f, side[j].v1 - f, y - f);
		
		if (c == CONE_DEGENERATE) {
			std::cerr << name
				<< ": Degenerate cone." << std::endl;
			exit(1);
		}
		
		if ((s[0] < -1e-6) || (s[1] < -1e-6))
			continue;
		
		incones++;
		
		if (s[0] + s[1] < 1.0 - 1e-4) {
			std::cerr << name 
				<< ": Warning: inside point: ("
				<< s[0] << ", " << s[1] << "); sum = "
				<< s[0] + s[1] << std::endl;
		}
		
		if (s[0] + s[1] > 1.0 + 0.02)
			continue;
		
		if ((side[j].o.empty())
		&&  (s[0] >= 1e-6) && (s[1] >= 1e-6)) {
			side[j].o = y;
		}

		if (side[j].t.empty()) {
			side[j].t = y;
		} else if (side[j].d.empty()) {
			p::dvec q = y - side[j].t;
		
			long gcd = p::gcd(q[0], q[1]);
			
			q = q / gcd;
			
			side[j].d = q;

			if ((level == 0) && (side[j].o.empty())) {
				consider_point(y + q, level + 1);
				consider_point(y - q, level + 1);
			}
		}
	}
	
	if ((level == 0) && (incones == 0)) {
		std::cerr << name
			<< ": warning: no cone found for ("
			<< y << ")."
			<< std::endl;
		dump();
	}
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::store()
{
	for (size_t j = 0; j < side.size(); j++) {
		if (side[j].o.empty()) {
			std::cerr << "Empty side!" << std::endl;
			exit(1);
		}
	}
	
	v.clear();
	t.clear();
	
	v.resize(side.size());
	t.resize(side.size());
	
	for (size_t j = 0; j < side.size(); j++) {
		v[j] = side[j].v0;
		t[j] = side[j].o;
	}
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::extend(p::value &stats)
{
	int weaksides = 0;
	int jw = -1;
	
	for (size_t j = 0; j < side.size(); j++) {
		if (side[j].o.empty()) {
			weaksides++;
			jw = j;
		}
	}
	
	if (v.size() == 3) {
		if (weaksides == 0) {
			stats["3_max"]++;
			if (store())
				return(-1);
			return(0);
		}
		if (weaksides > 1) {
			//std::cerr << "T weak" << std::endl;
			//dump();
			stats["3_weak"]++;
			return(-1);
		}

		if (extend_tri(jw)) {
			std::cerr << "T ext failed" << std::endl;
			//dump();
			stats["3_fail"]++;
			return(-1);
		}
		
		stats["3_ext"]++;
		return(0);
	}
	
	if (v.size() == 4) {
		if (weaksides == 0) {
			stats["4_max"]++;
			if (store())
				return(-1);
			return(0);
		}
	
		if (weaksides > 1) {
			stats["4_weak"]++;
			return(-1);
		}
		
		if (extend_quad(jw)) {
			stats["4_fail"]++;
			//std::cerr << "Q2 ext failed: split" << std::endl;
			//dump();
			return(-1);
		}
	
		stats["4_ext"]++;
		return(0);
	}
	
	stats[">4"]++;
	return(-1);
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::extend_quad(int jw)
{
	for (int j = jw; j < 3; j++) {
		side[j] = side[j + 1];
	}
	
	side.resize(3);
	
	int j1 = jw % 3;
	int j0 = (j1 + 2) % 3;
	
	p::dvec z;
	
	if (merge_xy(z, side[j0].v0, side[j0].v1, side[j1].v1, side[j1].v0)) {
		// split
		//dump();
		return(-1);
	}

	side[j0].v1 = z;
	side[j1].v0 = z;
	
	if (store())
		return(-1);
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::extend_tri(int j0)
{
	std::vector<p::dvec> lt;

	for (int pm = 0; pm < 2; pm++) {
		int j1, j2;
		
		if (pm == 0) {
			j1 = (j0 + 1) % 3;
			j2 = (j0 + 2) % 3;
		} else {
			j1 = (j0 + 2) % 3;
			j2 = (j0 + 1) % 3;
		}
		
		if (side[j1].o.empty()) {
			std::cerr << "BUG: more than one weak side: "
				<< j1 << " in addition to " << j0 << std::endl;
			exit(1);
		}
		
		if (side[j2].d.empty())
			continue;
		
		lt.push_back(side[j1].o + side[j2].d);
		lt.push_back(side[j1].o - side[j2].d);
	}
	
	p::dvec a = side[(j0 + 1) % 3].v1;
	p::dvec r0 = side[j0].v0 - a;
	p::dvec r1 = side[j0].v1 - a;

	//std::cerr << "Extend from" << a << " to" << side[j0].v0
	//	<< " and" << side[j0].v1 << std::endl;

	p::dvec x_min;
	double sum_min = 1e+20;
	
	for (size_t k = 0; k < lt.size(); k++) {
		p::dvec s;
		int c = cone_s(s, r0, r1, lt[k] - a);
		
		if (c == CONE_DEGENERATE)
			continue;
		
		if ((s[0] < 1e-6) || (s[1] < 1e-6))
			continue;

		double sum = s[0] + s[1];
		
		if ((sum < 1.0 - 1e-4)
		&&  (s[0] > 1e-4) && (s[1] > 1e-4)) {
			std::cerr << "Warning!!! Integer point"
				<< lt[k] << " inside" << a << " + cone("
				<< r0 << "," << r1 << " )"
				<< "; sum = " << s[0] + s[1]
				<< std::endl;
			dump();
			return(-1);
		}
		
		if (sum < sum_min) {
			//std::cerr << "Selected x =" << x << std::endl;
			x_min = lt[k];
			sum_min = sum;
		}
	}
	
	if (x_min.size() == 0) {
		std::cerr << "Could not find in-cone integer point"
			<< std::endl;
		std::cerr << "Extend from" << a << " to" << side[j0].v0
			<< " and" << side[j0].v1 << std::endl;
		std::cerr << "Trying: " << std::endl;
		for (size_t k = 0; k < lt.size(); k++) {
			std::cerr << "  x =" << lt[k] << std::endl;
		}
		dump();
		return(-1);
	}
	
	side[j0].v0 = a + r0 * sum_min;
	side[j0].v1 = a + r1 * sum_min;
	side[j0].o = x_min;
	
	int jm = (j0 + 2) % 3;
	int j1 = (j0 + 1) % 3;
	
	side[jm].v1 = side[j0].v0;
	side[j1].v0 = side[j0].v1;
	
	if (store())
		return(-1);
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::check_det_sub(p::value &stats, int j0, int j1, int j2)
{
	p::dvec r0 = t[j1] - t[j0];
	p::dvec r1 = t[j2] - t[j0];
	
	double det = r0[0] * r1[1] - r0[1] * r1[0];
	double epsilon = 0.01;
	
	det = fabs(det);
	
	if ((det < 1.0 - epsilon) || (det > 1.0 + epsilon)) {
		std::cerr << "det = " << det << std::endl;
		dump();
		write(std::cerr);
		stats["!unimod"]++;
		return(-1);
	}

	return(0);
}

int icmodel::check_det(p::value &stats)
{
	if (t.size() < 3)
		return(-1);
	
	if (check_det_sub(stats, 0, 1, 2))
		return(-1);

	if (t.size() == 3)
		return(0);


	if (check_det_sub(stats, 0, 1, 3))
		return(-1);

	if (check_det_sub(stats, 0, 2, 3))
		return(-1);

	if (check_det_sub(stats, 1, 2, 3))
		return(-1);

	return(0);
}

// ************************************************************************
// 
// ************************************************************************
int icmodel::check_o(p::value &stats)
{
	for (size_t j = 0; j < side.size(); j++) {
		if (side[j].o.empty()) {
			stats["o_empty"]++;
			return(-1);
		}
	
		p::dvec s;
		int c = cone_s(s,
			side[j].v0 - f, side[j].v1 - f,
			side[j].o - f);

		if (c == CONE_DEGENERATE) {
			stats["o_degen"]++;
			return(-1);
		}
		
		if ((s[0] < 1e-6) || (s[1] < 1e-6)) {
			stats["o_out"]++;
			return(-1);
		}
		
		double sum = s[0] + s[1];
		
		if (sum < 1.0 - 1e-4) {
			stats["o_near"]++;
			return(-1);
		}

		if (sum > 1.0 + 0.02) {
			stats["o_far"]++;
			/*std::cerr << "Sides:" << std::endl;
			for (size_t l = 0; l < side.size(); l++) {
				std::cerr << "\tv0 = " << side[l].v0 << std::endl;
				std::cerr << "\tv1 = " << side[l].v1 << std::endl;
			}*/
			std::cerr << "  o = " << side[j].o << std::endl;
			std::cerr << "  v0 = " << side[j].v0 << std::endl;
			std::cerr << "  v1 = " << side[j].v1 << std::endl;
			std::cerr << "  sum = " << sum << std::endl;
			dump();
			return(-1);
		}
	}
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
int iip(p::dvec a, p::dvec b, p::dvec c, p::dvec &x)
{
	if (a[1] > b[1])
		a.swap(b);

	if (b[1] > c[1])
		b.swap(c);

	if (a[1] > b[1])
		a.swap(b);
	
	double y0 = ceil(a[1]);
	double y1 = floor(b[1]);

	if (b[1] > a[1] + 1e-8)
	for (double y = y0; y <= y1; y += 1.0) {
		double x0 = a[0] + (b[0] - a[0]) / (b[1] - a[1]) * (y - a[1]);
		double x1 = a[0] + (c[0] - a[0]) / (c[1] - a[1]) * (y - a[1]);
		
		if (x0 > x1) {
			double t = x1;
			x1 = x0;
			x0 = t;
		}
		
		if (ceil(x0) <= floor(x1)) {
			x.resize(2);
			x[0] = ceil(x0);
			x[1] = y;
			return(1);
		}
	}
	
	double y2 = ceil(b[1]);
	double y3 = floor(c[1]);

	if (c[1] > b[1] + 1e-8)
	for (double y = y2; y <= y3; y++) {
		double x0 = b[0] + (c[0] - b[0]) / (c[1] - b[1]) * (y - b[1]);
		double x1 = a[0] + (c[0] - a[0]) / (c[1] - a[1]) * (y - a[1]);

		if (x0 > x1) {
			double t = x1;
			x1 = x0;
			x0 = t;
		}
		
		if (ceil(x0) <= floor(x1)) {
			x.resize(2);
			x[0] = ceil(x0);
			x[1] = y;
			return(1);
		}
	}
	
	return(0);
}

int icmodel::check_lf(p::value &stats)
{
	p::dvec x;
	
	for (size_t j = 0; j < side.size(); j++) {
		if (iip(f,
		f + (side[j].v0 - f) * 0.97,
		f + (side[j].v1 - f) * 0.97, x)) {
			stats["lf"]++;
			std::cerr << "  Interior point: " << x << std::endl;
			dump();
			return(-1);
		}
	}
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
int main(int argc, char **argv)
{
	p::values["config"].parse_cmdline(argc, argv);
	
	if (p::values["config"] != 1) {
		std::cerr << HELPMSG << std::endl;
		return(0);
	}
	
	std::string fname = p::values["config"][0];
	
	std::ifstream ifs;
	
	ifs.open(fname, std::ifstream::in);
	
	if (!ifs.good()) {
		std::cerr << fname << ": failed to open." << std::endl;
		return(-1);
	}

	int icount = 0;
	int ocount = 0;
	int r = 0;
	p::value stats;
	
	while (1) {
		icmodel a, old;
		
		r = a.read(ifs);
		if (r <= 0)
			break;
		
		icount++;

		a.build();

		old = a;
		
		//std::cerr << "sides " << a.v.size() << std::endl;
		
		if (a.extend(stats))
			continue;
		
		if (a.check_det(stats))
			continue;
		
		if (a.check_o(stats))
			continue;

		if (a.check_lf(stats)) {
			std::cerr << "Old was:" << std::endl;
			old.dump();
			continue;
		}

		ocount++;
		
		a.write(std::cout);
	}

	std::cerr << stats << std::endl;
	
	if (r)
		std::cerr << "Incomplete model read." << std::endl;
	
	std::cerr << icount << " models read." << std::endl;
	std::cerr << ocount << " models written." << std::endl;
	
	return(0);
}

