#include <iostream>
#include <vector>
#include <queue>
#include <cmath>
#include "pxx/strprintf.hpp"
#include "pxx/timer.hpp"
#include "pxx/value.hpp"
#include "pxx/mipdb.hpp"
#include "pxx/cplex_model.hpp"
#include "pxx/onerow.hpp"
#include "pxx/tworow.hpp"
#include "concave.hpp"

// ************************************************************************
// 
// ************************************************************************
class triangle : public objective {
	
public:
	triangle(int n);
	virtual ~triangle();

	virtual double value(const p::dvec &x);
	virtual p::dvec gradient(const p::dvec &x);

public:
	p::dvec apex;
	p::dvec amplitude;
	std::vector<bool> left;
	p::dvec orig_obj;
	double lambda;
};

triangle::triangle(int n)
	: apex(n), amplitude(n), left(n)
{
}

triangle::~triangle()
{
}

double triangle::value(const p::dvec &x)
{
	int n = apex.size();
	double r = 0.0;
	
	for (int j = 0; j < n; j++) {
		if (x[j] <= apex[j]) {
			r += x[j] / apex[j] * amplitude[j];
		} else {
			r += (1.0 - x[j]) / (1.0 - apex[j]) * amplitude[j];
		}
	}
	
	return(r);
}

p::dvec triangle::gradient(const p::dvec &x)
{
	int n = apex.size();
	p::dvec g(n);
	bool go_left;
	
	for (int j = 0; j < n; j++) {
		go_left = ( (x[j] == apex[j]) ? left[j] :
			   ((x[j]  < apex[j]) ? true : false));

		g[j] = (go_left)
			? (1.0 / apex[j] * amplitude[j])
			: -1.0 / (1.0 - apex[j]) * amplitude[j];
	}
	
	g = g * (1.0 - lambda) + orig_obj * (lambda);
	
	return(g);
}

// ************************************************************************
// 
// ************************************************************************
#define HELPMSG	\
"Usage: primal.run [options] [problem]\n" \
"\n" \
"Options:\n" \
"  --presolve\n"


// ************************************************************************
// 
// ************************************************************************
#define INT_TOL		1e-6
#define LAMBDA_INIT	0.0
#define STUCK_THRESHOLD	1e-5
#define LAMBDA_MULT	1.0
#define LEFT_SWITCH_FRAC	0.5
int handle(p::cplex_model &o, double preoffs)
{
	concave_prob cp(o);
	int n = cp.n;
	int nint = 0;
	triangle f(n);
	p::dvec x0(n);
	
	for (int j = 0; j < n; j++) {
		if ((o.ctype[j] == 'I') || (o.ctype[j] == 'B')) {
			nint++;
			if ((o.lower[j] != 0.0) || (o.upper[j] != 1.0)) {
				std::cout << p::strprintf(
					"\tInteger column %d: %g <= %s <= %g\n",
					j, o.lower[j], o.xnames[j], o.upper[j]);
				return(1);
			}
		}
	}
	
	std::cout << "Instance is suitable, n = " << n << std::endl;
	
	for (int j = 0; j < n; j++) {
		f.left[j] = (o.obj[j] > 0.0);
		//f.left[j] = (cp.x[j] < 0.5);
		
		// pre-invert
		f.left[j] = !f.left[j];
	}
	
	f.lambda = LAMBDA_INIT;
	f.orig_obj = o.obj;
	
	x0 = cp.x;
	int it = 0;
	int nfrac = nint;
	while (nfrac > 0) {
		nfrac = 0;
		
		for (int j = 0; j < n; j++) {
			if (o.ctype[j] == 'C') {
				f.apex[j] = -1.0;
				f.amplitude[j] = 0.0;
			} else if ((x0[j] <= INT_TOL)
			|| (x0[j] >= 1.0 - INT_TOL)) {
				f.apex[j] = 0.5;
				f.amplitude[j] = 100.0 / nint;
			} else {
				f.apex[j] = x0[j];
				f.amplitude[j] = 100.0 / nint;
				if (urand(0.0, 1.0) < LEFT_SWITCH_FRAC)
					f.left[j] = !f.left[j];
				nfrac++;
			}
		}
		
		std::cout << p::strprintf("****** Iteration %d "
			"(%d / %d fractional), orig objval: %g",
			it++, nfrac, nint, x0 * o.obj)
			<< std::endl;
		
		
		/*std::cout << "Fractionals: ";
		
		for (int j = 0; j < n; j++) {
			if ((o.ctype[j] != 'C')
			&&  (x0[j] > INT_TOL)
			&&  (x0[j] < 1.0 - INT_TOL)) {
				std::cout << p::strprintf(
					" x%d = %g (grad %g)\n",
					j, x0[j], f.gradient(x0)[j]);
			}
		}*/
		
		cp.opt(f);
		
		if ((x0 - cp.x).norm() <= STUCK_THRESHOLD) {
			f.lambda *= LAMBDA_MULT;
			std::cout << "\tDecreasing lambda to " <<
				f.lambda << std::endl;
			if (f.lambda < 0.1)
				f.lambda = 0.0;
		}
		
		x0 = cp.x;
	}
	
	return(0);
}

// ************************************************************************
// 
// ************************************************************************
int main(int argc, char **argv)
{
	p::values["config"]["presolve"] = 0;
	p::values["config"].parse_cmdline(argc, argv);
	
	if (p::values["config"] != 1) {
		std::cout << HELPMSG << std::endl;
		return(0);
	}
	
	std::string fname = p::values["config"][0];
	
	p::cplex_model orig;
	int r;
	
	if (orig.readcopyprob(fname.c_str(), NULL)) {
		std::cout << fname << ": could not read file\n" << std::endl;
		return(1);
	}
	
	if (p::values["config"]["presolve"]) {
		orig.presolve();
		double offs = orig.getobjoffset();
		p::cplex_model pre(orig.getredlp());
		
		r = handle(pre, offs);
	} else {
		r = handle(orig, 0.0);
	}
	
	std::cout << p::values << std::endl;

	return(r);
}

