#include <iostream>
#include <vector>
#include <algorithm>
#include <queue>
#include <cmath>
#include <cstdlib>
#include "pxx/strprintf.hpp"
#include "pxx/timer.hpp"
#include "pxx/value.hpp"
#include "pxx/triangular.hpp"
#include "pxx/mipdb.hpp"
#include "pxx/cplex_model.hpp"
#include "pxx/onerow.hpp"
#include "pxx/tworow.hpp"

// ************************************************************************
// 
// ************************************************************************
#define HELPMSG	\
"Usage: ranshoot.run [options] [problem]\n" \
"\n" \
"Options:\n" \
"  none yet\n"

// ************************************************************************
// 
// ************************************************************************
class sol_info {
public:
	sol_info() {};
	sol_info(const p::dvec &x0, int cnt0 = 1) : x(x0), cnt(cnt0) {};

	bool operator<(const sol_info &b) const { return(cnt > b.cnt); };
public:
	p::dvec x;
	int cnt;
};

// ************************************************************************
// 
// ************************************************************************
#define SOL_DIFF_THRESHOLD	1e-5
#define SOL_123_N		19

bool sol_123(const p::dvec &x)
{
	static const double val[SOL_123_N] =
		{ 0.0, 1.0,
		  0.5,
		  1.0/3.0, 2.0/3.0,
		  0.25, 0.75,
		  0.2, 0.4, 0.6, 0.8,
		  1.0/6.0, 5.0/6.0,
		  1.0/7.0, 2.0/7.0, 3.0/7.0, 4.0/7.0, 5.0/7.0, 6.0/7.0,
		};
	
	for (size_t j = 0; j < x.size(); j++) {
		double v = x[j];
		bool eq123 = false;
		
		for (int i = 0; i < SOL_123_N; i++) {
			double d = fabs(v - val[i]);
		
			if (d <= SOL_DIFF_THRESHOLD) {
				eq123 = true;
				break;
			}
		}
		
		if (!eq123)
			return(false);
	}
	return(true);
}

bool sol_diff(const p::dvec &x, const p::dvec &y)
{
	for (size_t j = 0; j < x.size(); j++) {
		double d = fabs(x[j] - y[j]);
		
		if (d > SOL_DIFF_THRESHOLD)
			return(true);
	}
	return(false);
}

// ************************************************************************
// 
// ************************************************************************
int sol_add(std::vector<sol_info> &l, const p::dvec &x)
{
	for (std::vector<sol_info>::iterator i = l.begin(); i != l.end(); i++) {
		if (!sol_diff(i->x, x)) {
			i->cnt++;
			return(0);
		}
	}
	
	sol_info si(x, 1);
	
	l.push_back(si);
	return(1);
}

void sol_stats(std::vector<sol_info> &l)
{
	int o = 0;
	
	for (std::vector<sol_info>::iterator i = l.begin(); i != l.end(); i++) {
		std::cerr << o++ << ": " << i->cnt << " x: "
			<< i->x << std::endl;
	}
}

// ************************************************************************
// 
// ************************************************************************
int main(int argc, char **argv)
{
	p::values["config"].parse_cmdline(argc, argv);
	
	if (p::values["config"] != 1) {
		std::cerr << HELPMSG << std::endl;
		return(0);
	}
	
	std::string fname = p::values["config"][0];
	
	// read file
	p::cplex_model qc;
	
	if (qc.readcopyprob(fname.c_str(), NULL)) {
		std::cerr << fname << ": failed to load problem" << std::endl;
		return(1);
	}

	int n = qc.getnumcols();
	p::dvec x(n);
	std::vector<sol_info> l;
	int dup = 0;

	qc.make_LP();
	qc.obj.resize(n);
	
	std::cout << "CONV_SECTION" << std::endl;
	
	for (int t = 0; t < 100000000; t++) {
		for (int j = 0; j < n; j++) {
			qc.obj[j] = (double)rand() / (double)RAND_MAX;
		}
		qc.set_objective(qc.obj);
		qc.lpopt();
		qc.getx(&x[0], 0, n - 1);
		
		if (sol_add(l, x)) {
			std::cout << "(" << (l.size() - 1)  << ") "
				<< x << std::endl;
			if (!sol_123(x)) {
				std::cerr << "Strange sol: " << x << std::endl;
			}
		} else {
			dup++;
		}
		
		if ((t % 1000) == 0) {
			std::cerr << "Solutions: " << l.size()
				<< "; Duplicates: " << dup << std::endl;
		}
	}

	std::cerr << "Final results:" << std::endl;
	std::cerr << "Solutions: " << l.size() << "; Duplicates: " << dup
		<< std::endl;

	std::sort(l.begin(), l.end());
	std::cerr << "--- STATISTICS ---" << std::endl;
	sol_stats(l);
	
	return(0);
}

