#include <iostream>
#include <vector>
#include <cmath>
#include "pxx/model.hpp"
#include "pxx/cplex_model.hpp"
#include "pxx/mipdb.hpp"
#include "pxx/tolerance.hpp"
#include "pxx/strprintf.hpp"

// ***************************************************************************
// 
// ***************************************************************************
#define P_I	0
#define S_FREE	1
#define LIFTING	2
#define STRICT	3
#define NB_BNDS	4
#define N_RELAX	5

const char *relax_name[N_RELAX] = {
	"P_I",
	"S_FREE",
	"LIFTING",
	"STRICT",
	"NB_BNDS" };

/*
                    basic                   nonbasic
                int      cont            int        cont     
 0: P_I       drop bnds  drop bnds   drop int+nt  drop nt
 1: S-free               drop bnds   drop int+nt  drop nt
 2: lifting   drop bnds  drop bnds   drop nt      drop nt
 3: strict    drop nt    drop bnds   drop nt      drop nt
 4: NB bound  drop bnds  drop bnds   drop int


*/
int do_relax(p::mipdb_entry &ent, p::cplex_model &orig,
	p::dvec &xLP, std::vector<int> cstat)
{
	//int m = orig.getnumrows();
	int n = orig.getnumcols();
	
	// build relaxations
	orig.make_MIP();

	p::cplex_model relax[5] = orig;

	for (int j = 0; j < n; j++) {
		if (cstat[j] == CPX_BASIC) {
			if (orig.ctype[j] != 'C') {
				relax[P_I].lower[j] = -CPX_INFBOUND;
				relax[P_I].upper[j] =  CPX_INFBOUND;
				relax[LIFTING].lower[j] = -CPX_INFBOUND;
				relax[LIFTING].upper[j] =  CPX_INFBOUND;
				if (!orig.at_lb(xLP, j))
					relax[STRICT].lower[j] = -CPX_INFBOUND;
				if (!orig.at_ub(xLP, j))
					relax[STRICT].upper[j] =  CPX_INFBOUND;
				relax[NB_BNDS].lower[j] = -CPX_INFBOUND;
				relax[NB_BNDS].upper[j] =  CPX_INFBOUND;
			} else {
				relax[P_I].lower[j] = -CPX_INFBOUND;
				relax[P_I].upper[j] =  CPX_INFBOUND;
				relax[S_FREE].lower[j] = -CPX_INFBOUND;
				relax[S_FREE].upper[j] =  CPX_INFBOUND;
				relax[LIFTING].lower[j] = -CPX_INFBOUND;
				relax[LIFTING].upper[j] =  CPX_INFBOUND;
				relax[STRICT].lower[j] = -CPX_INFBOUND;
				relax[STRICT].upper[j] =  CPX_INFBOUND;
				relax[NB_BNDS].lower[j] = -CPX_INFBOUND;
				relax[NB_BNDS].upper[j] =  CPX_INFBOUND;
			}
		} else {
			relax[P_I].ctype[j] = 'C';
			relax[S_FREE].ctype[j] = 'C';
			relax[NB_BNDS].ctype[j] = 'C';
			
			if (cstat[j] == CPX_AT_LOWER) {
				relax[P_I].upper[j] =  CPX_INFBOUND;
				relax[S_FREE].upper[j] =  CPX_INFBOUND;
				relax[LIFTING].upper[j] =  CPX_INFBOUND;
				relax[STRICT].upper[j] =  CPX_INFBOUND;
			} else {
				relax[P_I].lower[j] = -CPX_INFBOUND;
				relax[S_FREE].lower[j] = -CPX_INFBOUND;
				relax[LIFTING].lower[j] = -CPX_INFBOUND;
				relax[STRICT].lower[j] = -CPX_INFBOUND;
			}
		}
	}

	// solve
	int has[N_RELAX];
	double optval[N_RELAX];
	double gc[N_RELAX];
	
	orig.getenv().setdblparam(CPX_PARAM_TILIM, 14400.0);
	//orig.getenv().setintparam(CPX_PARAM_SCRIND, CPX_ON);
	orig.getenv().setdblparam(CPX_PARAM_EPGAP, 1e-9);
	
	//orig.writeprob("orig.lp", NULL);

	for (int p = 0; p < N_RELAX; p++) {
		std::cout << p + 1 << "/" << N_RELAX << ": Solving "
			<< relax_name[p] << "..." << std::endl;
		relax[p].set_bounds(relax[p].lower, relax[p].upper);
		relax[p].set_ctype(&(relax[p].ctype[0]));
		relax[p].mipopt_failsafe();
		//relax[p].writeprob(p::strprintf("r%d.lp", p + 1).c_str(), NULL);
		
		int stat = relax[p].getstat();
		if ((stat == CPXMIP_OPTIMAL) || (stat == CPXMIP_OPTIMAL_TOL)) {
			has[p] = 1;
			optval[p] = relax[p].getobjval();
			gc[p] = 100.0 * (optval[p] - ent.lp_value)
				/ (ent.mip_value - ent.lp_value);
			std::cout << p::strprintf("\t%20.8f (%9.5f%%)",
				optval[p], gc[p]) << std::endl;
		} else {
			has[p] = 0;
			optval[p] = 0.0;
			std::cout << "\t" << relax[p].stat_str(stat)
				<< std::endl;
		}
		
		relax[p].clear();
	}
	
	// display
	std::cout << ent.name;
	
	for (int p = 0; p < N_RELAX; p++) {
		if (!has[p]) {
			std::cout << " ?";
		} else {
			if ((gc[p] <= 0.0) && (gc[p] >= -0.005))
				gc[p] = 0.0;
			
			std::cout << p::strprintf(" %.2f", gc[p]);
		}
	}
	
	std::cout << std::endl;
	
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
int run(std::string fname)
{
	// Detect instance and read file
	p::mipdb_entry ent;
	
	if (!ent.find(fname)) {
		std::cout << fname << ": unknown instance (" << ent.name
			<< ")" << std::endl;
		return(1);
	}
	
	if ((!ent.lp_known) || (!ent.mip_known)) {
		std::cout << ent.name << ": incomplete mipdb info"
			<< std::endl;
		return(1);
	}

	if (fabs(ent.mip_value - ent.lp_value) < 1e-5) {
		std::cout << ent.name << ": no integrality gap" << std::endl;
		return(1);
	}
	
	p::cplex_model orig;
	
	if (orig.readcopyprob(fname.c_str(), NULL)) {
		std::cout << fname << ": could not read problem" << std::endl;
		return(1);
	}
	
	// Put in standard form
	p::model std;
	p::elim_ruleset elim;
	
	orig.export_model(std);
	for (int j = 0; j < (int)std.ctype.size(); j++) {
		if (std.ctype[j] == 'B')
			std.ctype[j] = 'I';
	}
	std.standard(elim, PXX_STD_INT_SLACKS);
	orig.import_model(std);
	
	elim.clear();
	
	// Solve LP relaxation
	orig.getenv().setdblparam(CPX_PARAM_TILIM, 600.0);
	
	orig.make_LP();
	orig.lpopt();
	
	int stat = orig.getstat();
	
	if ((stat != CPX_STAT_OPTIMAL) && (stat != CPX_STAT_OPTIMAL_INFEAS)) {
		std::cout << ent.name << ": could not solve LP relaxation"
			<< std::endl;
		return(1);
	}
	
	std::cout << ent.name << ": LP objective is " << orig.getobjval()
		<< std::endl;
	std::cout << ent.name << ": mipdb: LP = " << ent.lp_value
		<< ", MIP = " << ent.mip_value << std::endl;
	
	if (!ent.consistent(orig.getobjval(), ent.lp_value, 0.005, 5e-6)) {
		std::cout << ent.name << ": inconsistent LP value" << std::endl;
		return(1);
	}
	
	int m = orig.getnumrows();
	int n = orig.getnumcols();
	std::vector<int> rstat(m), cstat(n);
	p::dvec xLP(n);
	
	orig.getx(&xLP[0], 0, n - 1);
	orig.getbase(&cstat[0], &rstat[0]);
	
	// pass on data
	return(do_relax(ent, orig, xLP, cstat));
}


// ***************************************************************************
// 
// ***************************************************************************
int main(int argc, char **argv)
{
	if (argc != 2) {
		std::cout << "Usage: relax.run [file]" << std::endl;
		return(0);
	}
	
	std::string fname(argv[1]);
	
	int r = run(fname);
	return(r);
}

