#include <iostream>
#include <fstream>
#include <cassert>
#include <vector>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include "pxx/dense_vector.hpp"
#include "pxx/sparse_vector.hpp"
#include "pxx/adjacency.hpp"
#include "pxx/forest.hpp"
#include "pxx/value.hpp"
#include "pxx/vp.hpp"

// ****************************************************************************
// 
// ****************************************************************************
#define HELPMSG	\
	"Usage: isvp.run [opts] [-g|-c adj] [-w weight]\n" \
	"\n" \
	"Files:\n" \
	"  -g adj          specify adjacency matrix file\n" \
	"  -c adj          specify complemented adjacency matrix file\n" \
	"  -w weight       specify node weight vector file\n" \
	"\n" \
	"General options:\n" \
	"  -b1             equivalent to: -pnd -er -eall\n" \
	"  -b2             equivalent to: -el -ende -endn -ede\n" \
	"  -iter n         specify max. iterations (default: 1G)\n" \
	"  -stall n        specify max. consecutive iterations without\n" \
	"                  improvement (default: 1G)\n" \
	"\n" \
	"Perturbation:\n" \
	"  -p min max      set perturbation magnitude (default: [-1, 2])\n" \
	"  -pnd            perturb costs (default: off)\n" \
	"  -upd            unperturb for degenerate pivots (default: off)\n" \
	"  -pd             (re)perturb for degenerate pivots (default: off)\n" \
	"\n" \
	"Entering variable:\n" \
	"  -el             select largest reduced cost, use variable index\n" \
	"                  or random to break ties, depending on -er flag\n" \
	"                  (default: use only variable index or random)\n" \
	"  -er             randomize variable selection (default: off)\n" \
	"  -ende           nondegenerate pivot-in edges (default: off)\n" \
	"  -endn           nondegenerate pivot-in nodes (default: off)\n" \
	"  -ede            degenerate pivot-in edges (default: off)\n" \
	"  -edn            degenerate pivot-in nodes (default: off)\n" \
	"  -eall           equivalent to: -ende -endn -ede -edn\n" \
	"\n"

int main(int argc, char **argv)
{
	double p_min = -1.0, p_max = 2.0;
	int flags = 0;
	int max_iter = 1<<30;
	int max_stalls = 1<<30;
	bool complement = false;
	std::string adjfile;
	std::string weightfile;
	
	for (int t = 1; t < argc; t++) {
		std::string arg = argv[t];
		
		if ((arg == "-g") || (arg == "-c")) {
			if ((++t) >= argc) {
				std::cout << "expected: file name" << std::endl;
				return(1);
			}
			adjfile = argv[t];
			complement = (arg == "-c");
		} else if (arg == "-w") {
			if ((++t) >= argc) {
				std::cout << "expected: file name" << std::endl;
				return(1);
			}
			weightfile = argv[t];
		} else if (arg == "-b1") {
			flags |= PXX_VP_PERTURB_ND | PXX_VP_ENTER_RANDOM
				| PXX_VP_ENTER_ND_EDGE | PXX_VP_ENTER_ND_NODE
				| PXX_VP_ENTER_D_EDGE | PXX_VP_ENTER_D_NODE;
		} else if (arg == "-b2") {
			flags |= PXX_VP_ENTER_LARGEST
				| PXX_VP_ENTER_ND_EDGE | PXX_VP_ENTER_ND_NODE
				| PXX_VP_ENTER_D_EDGE;
		} else if (arg == "-iter") {
			if ((++t) >= argc) {
				std::cout << "expected: iteration count"
					<< std::endl;
				return(1);
			}
			max_iter = strtol(argv[t], NULL, 0);
		} else if (arg == "-stall") {
			if ((++t) >= argc) {
				std::cout << "expected: iteration count"
					<< std::endl;
				return(1);
			}
			max_stalls = strtol(argv[t], NULL, 0);
		} else if (arg == "-p") {
			if (t + 2 >= argc) {
				std::cout << "expected: perturbation"
					<< std::endl;
				return(1);
			}
			p_min = strtod(argv[t + 1], NULL);
			p_max = strtod(argv[t + 2], NULL);
			
			t += 2;
		} else if (arg == "-pnd") {
			flags |= PXX_VP_PERTURB_ND;
		} else if (arg == "-upd") {
			flags |= PXX_VP_UNPERTURB_D;
		} else if (arg == "-pd") {
			flags |= PXX_VP_PERTURB_D;
		} else if (arg == "-el") {
			flags |= PXX_VP_ENTER_LARGEST;
		} else if (arg == "-er") {
			flags |= PXX_VP_ENTER_RANDOM;
		} else if (arg == "-ende") {
			flags |= PXX_VP_ENTER_ND_EDGE;
		} else if (arg == "-endn") {
			flags |= PXX_VP_ENTER_ND_NODE;
		} else if (arg == "-ede") {
			flags |= PXX_VP_ENTER_D_EDGE;
		} else if (arg == "-edn") {
			flags |= PXX_VP_ENTER_D_NODE;
		} else if (arg == "-eall") {
			flags |=  PXX_VP_ENTER_ND_EDGE | PXX_VP_ENTER_ND_NODE
				| PXX_VP_ENTER_D_EDGE | PXX_VP_ENTER_D_NODE;
		} else if (arg == "--help") {
			std::cout << HELPMSG << std::endl;
		} else {
			std::cout << "unknown parameter: " << arg << std::endl;
			return(1);
		}
	}
	
	if (adjfile.empty()) {
		std::cout << HELPMSG << std::endl;
		return(1);
	}
	
	p::values["global"].start();
	p::values[" files"].start();
	
	p::adjacency adj;

	if (adj.read_simple(adjfile.c_str()))
		return(1);
	
	if (complement) {
		std::cout << "Complement graph: ";
		adj.complement();
	} else {
		std::cout << "Original graph: ";
	}
	
	int n = adj.vertices();

	p::dvec weight(n);
	
	if (!weightfile.empty()) {
		std::ifstream f(weightfile.c_str());
	
		if (!f) {
			std::cerr << weightfile << ": unable to load weights"
				<< std::endl;
			return(1);
		}

		f >> weight;
	} else {
		weight.set(1.0);
	}

	p::values[" files"].stop();
	p::values[" computation"].start();
	
	int weight_sum = weight.sum();
	adj.pack();
	//p::adjacency_pre g(adj);
	
	std::cout << "n = " << n
		<< "; weight sum = " << weight_sum << std::endl;
		// << "; components: " << g.components() << std::endl;
	
	// ********************************
	std::vector<int> sol(n);
	std::fill(sol.begin(), sol.end(), -1);
	
	// for (int p = 0; p < g.components(); p++) {
		/*std::vector<int> idx;
		p::adjacency a;
		p::dvec w;

		g.build_component(p, a, idx);
		
		int q = a.vertices();
		
		w.resize(q);
		for (int j = 0; j < q; j++)
			w[j] = weight[idx[j]];
		
		std::cout << "Component " << p << ": n = " << q << std::endl;
		//std::cout << a;
		*/
		p::adjacency &a = adj;
		p::dvec &w = weight;
		/* --- */
		
		p::vp vp(a, w);
		
		vp.set_flags(flags);
		vp.set_max_iter(max_iter);
		vp.set_max_stalls(max_stalls);
		vp.set_perturbation(p_min, p_max);
		vp.solve();
		
		/*for (int j = 0; j < q; j++)
			sol[idx[j]] = vp[j];*/
		for (int j = 0; j < n; j++)
			sol[j] = vp[j];
		/* --- */
	//}

	p::values[" computation"].stop();
	p::values["global"].stop();
	
	// std::cerr << p::values << std::endl;
	printf("%.3f\n", p::values[" computation"].total_s());
	
	return(0);
}


