/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "forest.hpp"

namespace p {

// **************************************************************************
// 
// **************************************************************************
forest::forest()
	: n(0), f(0)
{
}

forest::forest(int nodes)
	: n(0), f(0)
{
	resize(nodes);
}

forest::forest(const forest &sf)
{
	resize(sf.n);
	
	for (int i = 0; i < n; i++)
		f[i] = sf.f[i];
}

forest::~forest()
{
	if (n)
		delete[] f;
}

int forest::size() const
{
	return(n);
}

void forest::resize(int nodes)
{
	if (n)
		delete[] f;
	
	n = nodes;
	f = new node[n];
}

const node &forest::operator[](int i) const
{
	return(f[i]);
}

node &forest::operator[](int i)
{
	return(f[i]);
}

// **************************************************************************
// 
// **************************************************************************
void forest::singletons()
{
	for (int i = 0; i < n; i++) {
		f[i].value = 0;
		f[i].meta = i;
		f[i].parent = -1;
		f[i].first_child = -1;
		f[i].prev_sibling = -1;
		f[i].next_sibling = -1;
	}
}

// **************************************************************************
// 
// **************************************************************************
bool forest::descending(int i, int j) const
{
	return(f[j].parent == i);
}

bool forest::ascending(int i, int j) const
{
	return(f[i].parent == j);
}

bool forest::linked(int i, int j) const
{
	return((f[j].parent == i) || (f[i].parent == j));
}

int forest::root(int i) const
{
	while (f[i].parent != -1)
		i = f[i].parent;
	
	return(i);
}

// **************************************************************************
// 
// **************************************************************************
void forest::reroot(int i)
{
	int k = i;
	int k_p = -1;
	int k_next = -1;
	int k_prev = -1;
	
	while (k != -1) {
		int p = f[k].parent;
		int next = f[k].next_sibling;
		int prev = f[k].prev_sibling;
		int fc = f[k].first_child;
		
		// detach k from parent and siblings
		if (p != -1) {
			if (f[p].first_child == k)
				f[p].first_child = next;
		
			if (next != -1)
				f[next].prev_sibling = prev;

			if (prev != -1)
				f[prev].next_sibling = next;
		}
		
		// update k's ancestry
		f[k].parent = k_p;
		f[k].next_sibling = k_next;
		f[k].prev_sibling = k_prev;
		
		// update k's children
		if (p != -1) {
			f[k].first_child = p;
			k_p = k;
			k_next = fc;
			k_prev = -1;
		
			if (fc != -1)
				f[fc].prev_sibling = p;
		}
		
		// switch to parent
		k = p;
	}
}

// **************************************************************************
// 
// **************************************************************************
void forest::attach(int child, int parent)
{
	reroot(child);
	
	int fc = f[parent].first_child;
	
	f[child].parent = parent;
	f[child].prev_sibling = -1;
	f[child].next_sibling = fc;
	
	if (fc != -1)
		f[fc].prev_sibling = child;
	
	f[parent].first_child = child;
}


void forest::link(int i, int j)
{
	if (f[i].parent == -1) {
		attach(i, j);
	} else {
		attach(j, i);
	}
}

// **************************************************************************
// 
// **************************************************************************
void forest::unlink(int i, int j)
{
	if (f[i].parent == j) {
		int t = j; j = i; i = t;
	}
	
	if (f[j].parent != i)
		return;
	
	int next = f[j].next_sibling;
	int prev = f[j].prev_sibling;
	
	if (f[i].first_child == j)
		f[i].first_child = next;
	
	if (next != -1)
		f[next].prev_sibling = prev;
	
	if (prev != -1)
		f[prev].next_sibling = next;
	
	f[j].parent = -1;
	f[j].next_sibling = -1;
	f[j].prev_sibling = -1;
}


// **************************************************************************
// 
// **************************************************************************
void forest::complement(int i)
{
	f[i].value = !f[i].value;
}

// **************************************************************************
// 
// **************************************************************************
void forest::tree_complement(int root)
{
	for (forest::it j(*this, root); *j != -1; j++) {
		f[*j].value = !f[*j].value;
	}
}

// **************************************************************************
// 
// **************************************************************************
void forest::tree_meta(int root, int value)
{
	for (forest::it j(*this, root); *j != -1; j++) {
		f[*j].meta = value;
	}
}

// **************************************************************************
// 
// **************************************************************************
forest::it::it(const forest &sf, int i0)
	: f(sf), first(i0), i(i0), d(0)
{
}

forest::it::it(const forest &sf, int root, int i0)
	: f(sf), first(root), i(i0), d(0)
{
}

forest::it::it(const it &si)
	: f(si.f), first(si.first), i(si.i), d(si.d)
{
}

forest::it::~it()
{
}


forest::it &forest::it::operator=(const forest::it &si)
{
	first = si.first;
	i = si.i;
	d = si.d;
	return(*this);
}

forest::it &forest::it::operator=(int i0)
{
	first = i = i0;
	d = 0;
	return(*this);
}

int forest::it::operator*() const
{
	return(i);
}

int forest::it::depth() const
{
	return(d);
}

int forest::it::operator++(int)
{
	int prev_i = i;
	operator++();
	return(prev_i);
}

int forest::it::operator++()
{
	if (i == -1)
		return(-1);
	
	if (f[i].first_child != -1) {
		i = f[i].first_child;
		d++;
		return(i);
	}
	
	while (i != first) {
		if (f[i].next_sibling != -1) {
			i = f[i].next_sibling;
			return(i);
		}
		i = f[i].parent;
		d--;
	}
	
	i = -1;
	return(-1);
}


// **************************************************************************
// 
// **************************************************************************
std::ostream &operator<<(std::ostream &os, const forest &f)
{
	for (int r = 0; r < f.size(); r++) {
		if (f[r].parent == -1) {
			for (forest::it it(f, r); *it != -1; it++) {
				for (int l = 0; l < it.depth(); l++) {
					os << "    ";
				}
				os << ((it.depth() == 0) ? "*-- " : "+-- ")
					<< "(" << *it << ") = "
					<< f[*it].value
					<< ", meta " << f[*it].meta << ""
					<< std::endl;
			}
		}
	}
	
	os << "FLAT:" << std::endl;
	
	for (int r = 0; r < f.size(); r++) {
		os << r << ": par " << f[r].parent
			<< ", fir " << f[r].first_child
			<< ", nex " << f[r].next_sibling
			<< ", prv " << f[r].prev_sibling
			<< ", value " << f[r].value
			<< ", meta " << f[r].meta << std::endl;
	}
	
	return(os);
}

// **************************************************************************
// 
// **************************************************************************
}


