/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "strprintf.hpp"
#include "model.hpp"


namespace p {

// ***************************************************************************
// 
// ***************************************************************************
model::model()
{
}

model::model(const model &src)
	: lower(src.lower), upper(src.upper), obj(src.obj),
	ctype(src.ctype), cname(src.cname), constraints(src.constraints)
{
}

model::~model()
{
}

void model::clear()
{
	obj.clear();
	lower.clear();
	upper.clear();
	ctype.clear();
	cname.clear();
	constraints.clear();
}

model &model::operator=(const model &src)
{
	lower = src.lower;
	upper = src.upper;
	obj = src.obj;
	ctype = src.ctype;
	cname = src.cname;
	constraints = src.constraints;
	return(*this);
}

constraint &model::add_constraint()
{
	constraints.push_back(constraint());
	return(constraints.back());
}

bool model::int_slack(const constraint &c) const
{
	if (int_tol.fractional(c.rhs))
		return(false);
	
	for (int l = 0; l < c.row.nz(); l++) {
		int j = c.row.index(l);
		double v = c.row.value(l);
		
		if (sparse_tol.nonzero(v)) {
			if ((ctype[j] != 'I') && (ctype[j] != 'B'))
				return(false);
			
			if (int_tol.fractional(v))
				return(false);
		}
	}
	
	return(true);
}

void model::standard(elim_ruleset &el, int opts)
{
	for (std::list<constraint>::iterator it = constraints.begin();
	it != constraints.end(); it++) {
		if ((it->sense == 'G') || (it->sense == 'L')) {
			el.eliminate(*it);
			
			int j = obj.size();
			
			elim_rule &e = el.add();
			e.slack = j;
			e.initial = *it;
			//el.eliminate(e.initial);
			
			obj.push_back(0.0);
			lower.push_back(0.0);
			upper.push_back(PXX_INF_BOUND);
			ctype.push_back('C');
			
			if ((opts & PXX_STD_INT_SLACKS) && (int_slack(*it)))
				ctype[j] = 'I';
			
			if (it->name.empty()) {
				cname.push_back(strprintf("_s%d", j));
			} else {
				cname.push_back(
					strprintf("_s%s", it->name.c_str()));
			}
			
			it->row.push(j, (it->sense == 'G') ? -1.0 : 1.0);
			it->sense = 'E';
		}
	}
	
	int n = obj.size();
	
	for (std::list<constraint>::iterator it = constraints.begin();
	it != constraints.end(); it++) {
		it->row.resize(n);
	}
}

// ***************************************************************************
// 
// ***************************************************************************
void elim_rule::eliminate(constraint &r) const
{
	double a_s = r.row[slack];
	
	if (sparse_tol.zero(a_s))
		return;
	
	if (initial.sense == 'L')
		a_s = -a_s;
	
	r.row += initial.row * a_s;
	r.rhs += initial.rhs * a_s;
	
	r.row.set_z(slack);
}

elim_rule &elim_ruleset::add()
{
	push_back(elim_rule());
	return(back());
}

void elim_ruleset::eliminate(constraint &r) const
{
	const_iterator it = begin();
	
	int l = 0;
	while (l < r.row.nz()) {
		int j = r.row.index(l);
		
		while ((it != end()) && (it->slack < j))
			it++;
		
		if (it == end()) {
			return;
		} else if (it->slack == j) {
			// std::cout << "* Eliminating x" << j << std::endl;
			
			double a_s = r.row.value(l);
			
			if (it->initial.sense == 'L')
				a_s = -a_s;
			
			r.row += it->initial.row * a_s;
			r.rhs += it->initial.rhs * a_s;
			
			// after the addition, l may become invalid
			l = r.row.locate(j);
			if (r.row.is(l, j))
				r.row.remove(l);

			it++;
		} else { // (it->slack > j)
			l++;
		}
	}
}

// ***************************************************************************
// 
// ***************************************************************************
} // namespace

