/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cmath>
#include "rational.hpp"

namespace p {

// **************************************************************************
// 
// **************************************************************************
rational::rational()
	: num(0), den(1)
{
}

rational::rational(const rational &src)
	: num(src.num), den(src.den)
{
}

rational::rational(const double &dbl)
{
	approx(dbl, PXX_RATIONAL_DEFAULT_MAX_COEF);
}

rational::~rational()
{
}

rational &rational::operator=(const rational &src)
{
	num = src.num;
	den = src.den;
	return(*this);
}

rational &rational::operator=(const double &dbl)
{
	approx(dbl, PXX_RATIONAL_DEFAULT_MAX_COEF);
	return(*this);
}

// **************************************************************************
// 
// **************************************************************************
void rational::approx(double f, unsigned int max_coef)
{
	double dmax = (double)max_coef + 1.0;
	unsigned int p[2], q[2];
	bool neg = false;

	if (f < 0.0) {
		f = -f;
		neg = true;
	}
	
	p[0] = 0;
	q[0] = 1;
	p[1] = 1;
	q[1] = 0;
	
	while (1) {
		double ff = floor(f);

		if (ff >= dmax)
			break;
		
		f = 1.0 / (f - ff);
		
		double dpn = ff * (double)p[1] + (double)p[0];
		double dqn = ff * (double)q[1] + (double)q[0];

		if ((dpn >= dmax) || (dqn >= dmax))
			break;
		
		unsigned int a = ff;
		unsigned int pn = a * p[1] + p[0];
		unsigned int qn = a * q[1] + q[0];
		
		// printf("cf: %ld/%ld (={%.10f, %.10e} ~{%.10f, %.10e})\n",
		//	pn, qn,
		//	(double)pn / (double)qn,
		//	(double)pn / (double)qn,
		//	f0 - (double)pn / (double)qn,
		//	f0 - (double)pn / (double)qn);

		p[0] = p[1];
		q[0] = q[1];
		p[1] = pn;
		q[1] = qn;
	}
	
	num = (neg) ? -p[1] : p[1];
	den = q[1];
}

// **************************************************************************
// 
// **************************************************************************
double rational::get_double() const
{
	return((double)num / (double)den);
}



// **************************************************************************
// 
// **************************************************************************
std::ostream &operator<<(std::ostream &os, const rational &q)
{
	os << q.num << "/" << q.den;
	
	return(os);
}


// **************************************************************************
// 
// **************************************************************************

} // namespace


