/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cmath>
#include "sparse_vector.hpp"

namespace p {

// *************************************************************************
// 
// *************************************************************************
__sref::__sref(svec &vec, int idx)
	: vector(vec), index(idx)
{
}

__sref::__sref(const __sref &src)
	: vector(src.vector), index(src.index)
{
}

__sref::~__sref()
{
}

__sref::operator double() const
{
	return(vector.get(index));
}

double __sref::operator=(const __sref &src) const
{
	double scalar = src;
	// Special speedup for appending
	if ((vector.nz() == 0) || (vector.index(vector.nz() - 1) < index))
		vector.push(index, scalar);
	else
		vector.set(index, scalar);
	
	return(scalar);
}

double __sref::operator=(double scalar) const
{
	// Special speedup for appending
	if ((vector.nz() == 0) || (vector.index(vector.nz() - 1) < index))
		vector.push(index, scalar);
	else
		vector.set(index, scalar);
	
	return(scalar);
}

double __sref::operator+=(double scalar) const
{
	int l = vector.locate(index);
	
	if (!vector.is(l, index))
		vector.insert(l, index, 0.0);
	
	return(vector.value(l) += scalar);
}

double __sref::operator-=(double scalar) const
{
	int l = vector.locate(index);
	
	if (!vector.is(l, index))
		vector.insert(l, index, 0.0);
	
	return(vector.value(l) -= scalar);
}

double __sref::operator*=(double scalar) const
{
	int l = vector.locate(index);
	
	if (!vector.is(l, index))
		return(0.0);

	return(vector.value(l) *= scalar);
}

double __sref::operator/=(double scalar) const
{
	int l = vector.locate(index);
	
	if (!vector.is(l, index))
		return(0.0);
	
	return(vector.value(l) /= scalar);
}



// *************************************************************************
// 
// *************************************************************************
svec::svec()
	: n(0)
{
}

svec::svec(const svec &src)
	: n(src.n), data(src.data)
{
}

svec::svec(const dvec &src)
	: n(0)
{
	gather(src);
}

svec::svec(int n0)
	: n(n0)
{
}

svec::svec(int n0, const double *val)
	: n(0)
{
	gather(n0, val);
}

svec::svec(int n0, int nz, const int *idx, const double *val)
	: n(n0), data(nz)
{
	set(nz, idx, val);
}

svec::~svec()
{
}

// *************************************************************************
// 
// *************************************************************************
void svec::clear()
{
	n = 0;
	data.clear();
}

int svec::size() const
{
	return(n);
}

void svec::resize(int n0)
{
	n = n0;
}

void svec::gather(int n0, const double *val, const tolerance &tol)
{
	n = n0;
	
	for (int i = 0; i < n; i++) {
		push(i, val[i], tol);
	}
}

void svec::gather(const dvec &src, const tolerance &tol)
{
	gather(src.size(), &src[0], tol);
}

void svec::spread(double *val) const
{
	for (int i = 0; i < n; i++)
		val[i] = 0.0;
	
	for (int l = 0; l < (int)data.size(); l++)
		val[data[l].index] = data[l].value;
}

void svec::spread(dvec &dst) const
{
	dst.resize(n);
	
	spread(&dst[0]);
}

dvec svec::dense() const
{
	dvec r;
	
	spread(r);
	
	return(r);
}

int svec::nz() const
{
	return(data.size());
}


// *************************************************************************
// 
// *************************************************************************
svec svec::operator+(const svec &src) const
{
	svec r(std::max(n, src.n));

	int l = 0, sl = 0;
	int i = nindex(0, r.n), si = src.nindex(0, r.n);
	
	while ((i < n) || (si < src.n)) {
		if (i == si) {
			double v = data[l].value + src.data[sl].value;
			
			r.push(i, v);
			
			i = nindex(++l, r.n);
			si = src.nindex(++sl, r.n);
		} else if (i < si) {
			r.push_nz(i, data[l].value);
			i = nindex(++l, r.n);
		} else {
			r.push_nz(si, src.data[sl].value);
			si = src.nindex(++sl, r.n);
		}
	}
	
	return(r);
}

svec svec::operator-(const svec &src) const
{
	svec r(std::max(n, src.n));

	int l = 0, sl = 0;
	int i = nindex(0, r.n), si = src.nindex(0, r.n);
	
	while ((i < n) || (si < src.n)) {
		if (i == si) {
			double v = data[l].value - src.data[sl].value;
			
			r.push(i, v);
			
			i = nindex(++l, r.n);
			si = src.nindex(++sl, r.n);
		} else if (i < si) {
			r.push_nz(i, data[l].value);
			i = nindex(++l, r.n);
		} else {
			r.push_nz(si, -src.data[sl].value);
			si = src.nindex(++sl, r.n);
		}
	}
	
	return(r);
}

double svec::operator*(const svec &src) const
{
	double v = 0.0;
	
	int l = 0, sl = 0;
	int i = nindex(0), si = src.nindex(0);
	
	while ((i < n) && (si < src.n)) {
		if (i == si) {
			v += data[l].value * src.data[sl].value;
			i = nindex(++l);
			si = src.nindex(++sl);
		} else if (i < si) {
			i = nindex(++l);
		} else {
			si = src.nindex(++sl);
		}
	}
	
	return(v);
}

double svec::operator*(const dvec &src) const
{
	double v = 0.0;
	
	for (int l = 0; l < (int)data.size(); l++) {
		int j = data[l].index;
		
		if (j < (int)src.size())
			v += data[l].value * src[j];
	}
	
	return(v);
}

svec svec::operator+() const
{
	return(*this);
}

svec svec::operator-() const
{
	svec r(n);
	
	r.data.resize(data.size());
	
	for (int l = 0; l < (int)data.size(); l++) {
		r.data[l].index = data[l].index;
		r.data[l].value = -data[l].value;
	}
	
	return(r);
}

svec svec::operator*(double scalar) const
{
	svec r(n);
	
	for (int l = 0; l < (int)data.size(); l++) {
		double v = data[l].value * scalar;
		
		r.push(data[l].index, v);
	}
	
	return(r);
}

svec svec::operator/(double scalar) const
{
	svec r(n);
	
	for (int l = 0; l < (int)data.size(); l++) {
		double v = data[l].value / scalar;

		r.push(data[l].index, v);
	}
	
	return(r);
}

// *************************************************************************
// 
// *************************************************************************
svec &svec::operator=(const svec &src)
{
	n = src.n;
	data = src.data;
	
	return(*this);
}

svec &svec::operator+=(const svec &src)
{
	n = std::max(n, src.n);

	int l = 0, sl = 0;
	int i = nindex(0), si = src.nindex(0, n);
	
	while ((i < n) || (si < src.n)) {
		if (i == si) {
			double v = data[l].value + src.data[sl].value;
			
			if (sparse_tol.nonzero(v)) {
				data[l].value = v;
				l++;
			} else {
				remove(l);
			}
			i = nindex(l);
			si = src.nindex(++sl, n);
		} else if (i < si) {
			i = nindex(++l);
		} else {
			insert(l++, si, src.data[sl].value);
			si = src.nindex(++sl, n);
		}
	}

	return(*this);
}

svec &svec::operator-=(const svec &src)
{
	n = std::max(n, src.n);

	int l = 0, sl = 0;
	int i = nindex(0), si = src.nindex(0, n);
	
	while ((i < n) || (si < src.n)) {
		if (i == si) {
			double v = data[l].value - src.data[sl].value;
			
			if (sparse_tol.nonzero(v)) {
				data[l].value = v;
				l++;
			} else {
				remove(l);
			}
			i = nindex(l);
			si = src.nindex(++sl, n);
		} else if (i < si) {
			i = nindex(++l);
		} else {
			insert(l++, si, -src.data[sl].value);
			si = src.nindex(++sl, n);
		}
	}

	return(*this);
}

svec &svec::operator*=(double scalar)
{
	int l = 0;
	
	while (l < (int)data.size()) {
		double v = data[l].value * scalar;
		
		if (sparse_tol.nonzero(v)) {
			data[l].value = v;
			l++;
		} else {
			remove(l);
		}
	}

	return(*this);
}

svec &svec::operator/=(double scalar)
{
	int l = 0;
	
	while (l < (int)data.size()) {
		double v = data[l].value / scalar;
		
		if (sparse_tol.nonzero(v)) {
			data[l].value = v;
			l++;
		} else {
			remove(l);
		}
	}

	return(*this);
}

// *************************************************************************
// 
// *************************************************************************
// used only when the svec is explicitely defined as const
double svec::operator[](int idx) const
{
	return(data[idx].value);
}

__sref svec::operator[](int idx)
{
	return(__sref(*this, idx));
}

// *************************************************************************
// 
// *************************************************************************
double svec::sqnorm() const
{
	double v = 0.0;
	
	for (int l = 0; l < (int)data.size(); l++) {
		double t = data[l].value;
		v += t * t;
	}
	
	return(v);
}

double svec::norm() const
{
	return(sqrt(sqnorm()));
}

// *************************************************************************
// 
// *************************************************************************
double svec::get(int idx) const
{
	int l = locate(idx);
	
	if (is(l, idx))
		return(data[l].value);
	
	return(0.0);
}

void svec::get(int *idx, double *val) const
{
	int nz = data.size();
	
	for (int l = 0; l < nz; l++) {
		idx[l] = data[l].index;
		val[l] = data[l].value;
	}
}

void svec::set(int idx, const double value, const tolerance &tol)
{
	if (tol.zero(value)) {
		set_z(idx);
	} else {
		set_nz(idx, value);
	}
}

void svec::set_nz(int idx, const double value)
{
	int l = locate(idx);
	
	if (is(l, idx)) {
		data[l].value = value;
	} else {
		insert(l, idx, value);
	}
}

void svec::set_z(int idx)
{
	int l = locate(idx);
	
	if (is(l, idx)) {
		remove(l);
	}
}

void svec::set(int nz, const int *idx, const double *val)
{
	data.resize(nz);
	
	for (int l = 0; l < nz; l++) {
		data[l].index = idx[l];
		data[l].value = val[l];
	}
}

void svec::push_nz(int idx, const double value)
{
	el e;
	
	e.index = idx;
	e.value = value;
	
	data.push_back(e);
}

void svec::push(int idx, const double value, const tolerance &tol)
{
	if (tol.nonzero(value)) {
		push_nz(idx, value);
	}
}

int svec::locate(int idx) const
{
	return(locate(idx, 0, data.size()));
}

#define PXX_LOCATE_LINEAR_THRESHOLD	8
int svec::locate(int idx, int l0, int l1) const
{
	int i0 = l0, i1 = l1;
	
	while ((i1 - i0) > PXX_LOCATE_LINEAR_THRESHOLD) {
		int im = (i0 + i1) >> 1;
		
		if (data[im].index == idx)
			return(im);
		else if (data[im].index < idx)
			i0 = im + 1;
		else
			i1 = im;
	}
	
	while ((i0 < i1) && (data[i0].index < idx))
		i0++;
	
	return(i0);
}

bool svec::is(int offset, int idx) const
{
	return((offset < (int)data.size()) && (data[offset].index == idx));
}

void svec::insert(int offset, int idx, const double value)
{
	el e;
	
	e.index = idx;
	e.value = value;
	
	data.insert(data.begin() + offset, e);
}

void svec::remove(int offset)
{
	data.erase(data.begin() + offset);
}

int svec::nindex(int offset, int n0) const
{
	return( (offset >= (int)data.size()) ? n0 : data[offset].index );
}

int svec::nindex(int offset) const
{
	return( (offset >= (int)data.size()) ? n : data[offset].index );
}

double svec::nvalue(int offset) const
{
	return( (offset >= (int)data.size()) ? 0.0 : data[offset].value );
}

int &svec::index(int offset)
{
	return(data[offset].index);
}

double &svec::value(int offset)
{
	return(data[offset].value);
}

int svec::index(int offset) const
{
	return(data[offset].index);
}

double svec::value(int offset) const
{
	return(data[offset].value);
}


// *************************************************************************
// 
// *************************************************************************
std::ostream &operator<<(std::ostream &os, const svec &v)
{
	/*int i0 = 0;
	
	for (int l = 0; l <= v.nz(); l++) {
		int i = v.nindex(l);
		
		for (int j = i0; j < i; j++)
			os << " .0";
		
		if (i < v.size())
			os << " " << v.nvalue(l);
		
		i0 = i;
	}*/
	
	for (int l = 0; l < v.nz(); l++) {
		os << "[" << l << "](" << v.index(l) << ":" << v.value(l)
			<< ") ";
	}
	return(os);
}

// *************************************************************************
// 
// *************************************************************************

} // namespace

