/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef PXX_SPARSE_VECTOR_HPP
#define PXX_SPARSE_VECTOR_HPP
#include <iostream>
#include <vector>
#include "tolerance.hpp"
#include "dense_vector.hpp"

namespace p {

// **************************************************
class __sref;
class svec;

// **************************************************
class svec {

public:
	svec();
	svec(const svec &src);
	svec(int n);
	svec(int n, const double *val);
	svec(const dvec &src);
	svec(int n, int nz, const int *idx, const double *val);
	~svec();
	
	void clear();
	
	int size() const;
	void resize(int n);
	
	void gather(int n, const double *val, const tolerance &t = sparse_tol);
	void gather(const dvec &src, const tolerance &t = sparse_tol);
	void spread(double *val) const;
	void spread(dvec &dst) const;
	dvec dense() const;
	int nz() const;
	
	svec operator+(const svec &src) const;
	svec operator-(const svec &src) const;
	double operator*(const svec &src) const;
	double operator*(const dvec &src) const;
	
	svec operator+() const;
	svec operator-() const;
	svec operator*(double scalar) const;
	svec operator/(double scalar) const;
	
	svec &operator=(const svec &src);
	svec &operator+=(const svec &src);
	svec &operator-=(const svec &src);
	svec &operator*=(double scalar);
	svec &operator/=(double scalar);
	
	double operator[](int idx) const;
	__sref operator[](int idx);
	
	double sqnorm() const;
	double norm() const;
	
	double get(int idx) const;
	void get(int *idx, double *value) const;
	void set_nz(int idx, const double value);
	void set_z(int idx);
	void set(int idx, const double value, const tolerance &t = sparse_tol);
	void set(int nz, const int *idx, const double *value);
	void push_nz(int idx, const double value);
	void push(int idx, const double value, const tolerance &t = sparse_tol);
	
	// Low-level
	int locate(int idx) const;
	int locate(int idx, int l0, int l1) const;
	bool is(int offset, int idx) const;
	void insert(int offset, int idx, const double value);
	void remove(int offset);
	
	int nindex(int offset) const;
	int nindex(int offset, int n) const;
	double nvalue(int offset) const;
	
	int &index(int offset);
	double &value(int offset);
	
	int index(int offset) const;
	double value(int offset) const;
private:
	int n;
	
	struct el {
		int index;
		double value;
	};
	
	std::vector<el> data;
};

std::ostream &operator<<(std::ostream &os, const svec &v);


// **************************************************
class __sref {

public:
	__sref(svec &vec, int idx);
	__sref(const __sref &src);
	~__sref();
	
	operator double() const;
	double operator=(const __sref &src) const;
	double operator=(double scalar) const;
	double operator+=(double scalar) const;
	double operator-=(double scalar) const;
	double operator*=(double scalar) const;
	double operator/=(double scalar) const;
	
	svec &vector;
	int index;
};

// **************************************************

} // namespace

#endif

