/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/time.h>
#include <cmath>
#include <string>
#include "strprintf.hpp"
#include "timer.hpp"

namespace p {

// ***************************************************************************
// 
// ***************************************************************************
// from libp, from <sys/time.h>:
#define p_time_now(t)		gettimeofday(t, NULL)
#define p_time_zero(tvp)	((tvp)->tv_sec = (tvp)->tv_usec = 0)
#define p_time_add(dst, a, b) \
	do { \
		(dst)->tv_sec = (a)->tv_sec + (b)->tv_sec; \
		(dst)->tv_usec = (a)->tv_usec + (b)->tv_usec; \
		if ((dst)->tv_usec >= 1000000) { \
			(dst)->tv_sec++; \
			(dst)->tv_usec -= 1000000; \
		} \
	} while (0)
#define p_time_sub(dst, a, b) \
	do { \
		(dst)->tv_sec = (a)->tv_sec - (b)->tv_sec; \
		(dst)->tv_usec = (a)->tv_usec - (b)->tv_usec; \
		if ((dst)->tv_usec < 0) { \
			(dst)->tv_sec--; \
			(dst)->tv_usec += 1000000; \
		} \
	} while (0)

// ***************************************************************************
// 
// ***************************************************************************
timer::timer()
{
	clear();
}

timer::timer(const timer &src)
{
	total = src.total;
	begin = src.begin;
	nc = src.nc;
	on = src.on;
}

timer::~timer()
{
}

timer &timer::operator=(const timer &src)
{
	total = src.total;
	begin = src.begin;
	nc = src.nc;
	return(*this);
}

void timer::clear()
{
	p_time_zero(&total);
	on = true;
	nc = 0;
}

void timer::enable()
{
	on = true;
}

void timer::disable()
{
	on = false;
}

bool timer::enabled() const
{
	return(on);
}

void timer::shadow_start()
{
	if (!on)
		return;
	
	p_time_now(&begin);
}

void timer::shadow_stop()
{
	if (!on)
		return;
	
	struct timeval end, delta;
	
	p_time_now(&end);
	p_time_sub(&delta, &end, &begin);
	p_time_add(&total, &total, &delta);
}

void timer::start()
{
	shadow_start();
}

void timer::stop()
{
	shadow_stop();
	nc++;
}

double timer::elapsed_s() const
{
	if (!on)
		return(0.0);

	struct timeval end, delta;
	
	p_time_now(&end);
	p_time_sub(&delta, &end, &begin);
	
	return(1.0e-6 * delta.tv_usec + (double)delta.tv_sec);
}

double timer::total_s() const
{
	return(1.0e-6 * total.tv_usec + (double)total.tv_sec);
}

int timer::calls() const
{
	return(nc);
}

// ***************************************************************************
// 
// ***************************************************************************
std::ostream &operator<<(std::ostream &os, const timer &t)
{
	if (t.enabled()) {
		os << strprintf("%5d %12.6f %12.6f",
			t.calls(), t.total_s(),
			(t.calls() == 0)
				? 0.0 : t.total_s() / (double)t.calls());
	} else {
		os << strprintf("%5d - -", t.calls());
	}

	return(os);
}

// ***************************************************************************
// 
// ***************************************************************************

} // namespace
