/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "triangular.hpp"

namespace p {

// ***************************************************************************
// generic
// ***************************************************************************
template<typename T>
triangular<T>::triangular()
	: n(0)
{
}

template<typename T>
triangular<T>::triangular(int nV)
	: n(nV), values(nV * (nV + 1) / 2)
{
}

template<typename T>
triangular<T>::triangular(const triangular &src)
	: n(src.n), values(src.values)
{
}

template<typename T>
triangular<T>::~triangular()
{
}

template<typename T>
triangular<T> &triangular<T>::operator=(const triangular &src)
{
	n = src.n;
	values = src.values;
	
	return(*this);
}

template<typename T>
void triangular<T>::clear()
{
	n = 0;
	values.clear();
}

template<typename T>
void triangular<T>::resize(int nV)
{
	n = nV;
	values.resize(nV * (nV + 1) / 2);
}

template<typename T>
int  triangular<T>::size() const
{
	return(n);
}

template<typename T>
int  triangular<T>::elements() const
{
	return(values.size());
}

template<typename T>
int  triangular<T>::offset(int i, int j) const
{
	if (j > i)
		std::swap(i, j);

	return((i * (i + 1)) / 2 + j);
}

template<typename T>
T &triangular<T>::array(int offs)
{
	return(values[offs]);
}

template<typename T>
const T &triangular<T>::array(int offs) const
{
	return(values[offs]);
}


template<typename T>
void triangular<T>::set(T v)
{
	std::fill(values.begin(), values.end(), v);
}

template<typename T>
void triangular<T>::set(int i, int j, T v)
{
	values[offset(i, j)] = v;
}

template<typename T>
T    triangular<T>::get(int i, int j) const
{
	return(values[offset(i, j)]);
}

template<typename T>
T   &triangular<T>::operator() (int i, int j)
{
	return(values[offset(i, j)]);
}

template<typename T>
T   triangular<T>::schur_product(const triangular &src) const
{
	T v = 0;
	size_t e = std::min(values.size(), src.values.size());
	
	for (size_t t = 0; t < e; t++)
		v += values[t] * src.values[t];
	
	return(v);
}

template class triangular<double>;
template class triangular<int>;
template class triangular<char>;

} // namespace
