/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cmath>
#include <vector>
#include <algorithm>
#include "strprintf.hpp"
#include "value.hpp"
#include "z2.hpp"
#include "rational.hpp"
#include "tworow.hpp"

#ifndef M_PI
#define M_PI		3.14159265358979323846
#endif

//#define DEBUG_EXPAND
//#define DEBUG_PROCESS
#define DEBUG_SPLIT
//#define DEBUG_DISPLAY
//#define DEBUG_ANALYZE

#define ANALYZE_HACKS
//#define USE_BOUNDS

namespace p {

// **************************************************************************
// 
// **************************************************************************
static int solve2x2(double x[2],
	const double c0[2], const double c1[2], const double b[2])
{
	double det;
	
	det = (c0[0] * c1[1]) - (c0[1] * c1[0]);
	
	if (fabs(det) <= PXX_TWOROW_SINGULAR_DET)
		return(-1);
	
	x[0] = ((c1[1] * b[0]) - (c1[0] * b[1])) / det;
	x[1] = ((c0[0] * b[1]) - (c0[1] * b[0])) / det;
	
	return(0);
}

static double atan2_2pi(double y, double x)
{
	double ra = atan2(y, x);

	if (ra < 0.0)
		ra += 2.0 * M_PI;
	
	return(ra);
}

// **************************************************************************
// 
// **************************************************************************
bool tworow_nonbasic::operator<(const tworow_nonbasic &o) const
{
	return(angle < o.angle);
}

// **************************************************************************
// 
// **************************************************************************
tworow::tworow()
{
}

tworow::tworow(const tworow &src)
	: col_order(src.col_order), col(src.col), ray(src.ray),
	point(src.point), ctype(src.ctype),
	p(src.p), ineq(src.ineq)
{
	f[0] = src.f[0];
	f[1] = src.f[1];
	fl[0] = src.fl[0];
	fl[1] = src.fl[1];
	fu[0] = src.fu[0];
	fu[1] = src.fu[1];
	orig_n = src.orig_n;
	basic1 = src.basic1;
	basic2 = src.basic2;
}

tworow::~tworow()
{
}

// **************************************************************************
// 
// **************************************************************************
void tworow::clear()
{
	col_order.clear();
	col.clear();
	ray.clear();
	point.clear();
	p.clear();
	ineq.clear();
#ifdef PXX_TWOROW_BUILD_MIP
	mip.clear();
#endif
}


// **************************************************************************
// 
// **************************************************************************
void tworow::init(model &q, const dvec &x0,
	const constraint &c1, const constraint &c2, int b1, int b2)
{
	// problem info
	orig_n = q.obj.size();
	basic1 = b1;
	basic2 = b2;
	ctype = q.ctype;

#ifdef PXX_TWOROW_BUILD_MIP
	mip.prepare_columns(orig_n);
	mip.lower = q.lower;
	mip.upper = q.upper;
	std::fill(mip.ctype.begin(), mip.ctype.end(), 'C');
	mip.ctype[b1] = 'I';
	mip.ctype[b2] = 'I';
	// mip.ctype = q.ctype; // 'lift' integer nonbasics
	mip.xstr = q.cname;
	for (int j = 0; j < orig_n; j++) {
		mip.xnames[j] = (char *)mip.xstr[j].c_str();
	}
	for (int j = 0; j < orig_n; j++) { // drop loose bound
		if ((j != b1) && (j != b2)) {
			double dist_l = x0[j] - q.lower[j];
			double dist_u = q.upper[j] - x0[j];
			bool at_upper = (dist_l > dist_u);
			
			if (at_upper)
				mip.lower[j] = -CPX_INFBOUND;
			else
				mip.upper[j] =  CPX_INFBOUND;
		}
	}

	mip.commit_columns();

	mip.add_constraint(c1);
	mip.add_constraint(c2);
#endif

	// fill out the col field
	int l1 = 0, l2 = 0;
	int j1 = c1.row.nindex(0);
	int j2 = c2.row.nindex(0);
	
	while ((j1 < c1.row.size()) || (j2 < c2.row.size())) {
		int j;
		double r[2];
				
		if (j1 < j2) {
			j = j1;
			r[0] = c1.row.nvalue(l1);
			r[1] = 0.0;
			j1 = c1.row.nindex(++l1);
		} else if (j2 < j1) {
			j = j2;
			r[0] = 0.0;
			r[1] = c2.row.nvalue(l2);
			j2 = c2.row.nindex(++l2);
		} else {
			j = j1;
			r[0] = c1.row.nvalue(l1);
			r[1] = c2.row.nvalue(l2);
			j1 = c1.row.nindex(++l1);
			j2 = c2.row.nindex(++l2);
		}

		if ((j != b1) && (j != b2)) {
			int nbidx = col.size();
			col.resize(nbidx + 1);
			tworow_nonbasic &v = col[nbidx];
		
			double dist_l = x0[j] - q.lower[j];
			double dist_u = q.upper[j] - x0[j];
			
			v.column = j;
			v.at_upper = (dist_l > dist_u);
			v.r[0] = (v.at_upper) ? r[0] : -r[0];
			v.r[1] = (v.at_upper) ? r[1] : -r[1];
			v.angle = atan2_2pi(v.r[1], v.r[0]);
			v.norm = sqrt(v.r[0] * v.r[0] + v.r[1] * v.r[1]);
			v.ray = -1;
			v.init_idx = nbidx;
			v.lower = q.lower[j];
			v.upper = q.upper[j];
			
			/*std::cout << nbidx << ": c" << j << " = ["
				<< r[0] << ", " << r[1] << "]; norm = "
				<< v.norm << " bounds [" << v.lower
				<< ", " << v.upper << "]; x0 = " << x0[j]
				<< std::endl;*/
		}
	}
	
	// sort col, saving initial order
	std::sort(col.begin(), col.end());
	
	col_order.resize(col.size());
	
	for (int ao = 0; ao < (int)col.size(); ao++)
		col_order[col[ao].init_idx] = ao;
	
	// fill out ray field
	for (int k = 0; k < (int)col.size(); k++) {
		int r0 = (int)ray.size() - 1;
		int r;
		
		if (r0 < 0) {
			r = -1;
		} else if (col[k].angle >= ray[0].angle + 2.0 * M_PI
					- PXX_TWOROW_ANGLE_TOL) {
			r = 0;
		} else if (col[k].angle <= ray[r0].angle
					+ PXX_TWOROW_ANGLE_TOL) {
			r = r0;
		} else {
			r = -1;
		}
		
		if (r == -1) {
			r = ray.size();
			ray.resize(r + 1);
			tworow_ray &h = ray[r];
			
			h.r[0] = col[k].r[0] / col[k].norm;
			h.r[1] = col[k].r[1] / col[k].norm;
			h.angle = col[k].angle;
			h.flags = 0;
			h.convex_constr = -1;
		}
		
		col[k].ray = r;
	}
	
	// add artificial rays
	for (int r1 = 0; r1 < (int)ray.size(); r1++) {
		int r2 = (r1 + 1) % (int)ray.size();
		
		double span = ray[r2].angle - ray[r1].angle;
		if (span < 0.0)
			span += 2.0 * M_PI;
		
		if (span >= 0.9 * M_PI) {
			tworow_ray a;
			
			if (span >= 1.8 * M_PI)
				a.angle = ray[r1].angle + span / 3.0;
			else
				a.angle = ray[r1].angle + span / 2.0;
			
			if (a.angle > 2.0 * M_PI)
				a.angle -= 2.0 * M_PI;
			
			a.r[0] = cos(a.angle);
			a.r[1] = sin(a.angle);
			a.flags = PXX_TWOROW_RAY_ART;
			a.convex_constr = -1;
			
			ray.insert(ray.begin() + (r1 + 1), a);

			for (int k = 0; k < (int)col.size(); k++) {
				if (col[k].ray >= r1 + 1) {
					col[k].ray++;
				}
			}
		}
	}
		
	/*for (int r = 0; r < (int)ray.size(); r++) {
		std::cout << r << ": [" << ray[r].r[0] << ", " << ray[r].r[1]
			<< "] angle=" << ray[r].angle << std::endl;
	}*/
	
	// initialize columns
	int n = ray.size();
	
	p.prepare_columns(n);
	p.lower.set(PXX_TWOROW_ALPHA_LB);
	p.upper.set(CPX_INFBOUND);
	p.xstr.resize(n);
	
	for (int i = 0; i < n; i++) {
		p.ctype[i] = 'C';
		format(p.xstr[i], "a%d", i);
		p.xnames[i] = (char *)p.xstr[i].c_str();
	}
	
	p.commit_columns();
	
	// write down convexity constraints
	int nc = 0;
	
	for (int i1 = 0; i1 < n; i1++) {
		int i2 = (i1 + 1) % n;
		int i3 = (i1 + 2) % n;
		
		double l[2];
		
		if ((solve2x2(l, ray[i1].r, ray[i3].r, ray[i2].r) == 0)
		&&  (l[0] >= 0.0) && (l[1] >= 0.0)) {
			constraint c;

			format(c.name, "conv_r%d", i2);
			c.row.resize(n);
			c.row[i1] = l[0];
			c.row[i2] = -1.0;
			c.row[i3] = l[1];
			
			c.sense = 'G';
			c.rhs = 0.0;
			
			p.add_constraint(c);
			
			ray[i2].convex_constr = nc;
			nc++;
		}
	}
	
	// right-hand side
	f[0] = x0[b1];
	f[1] = x0[b2];
#ifdef USE_BOUNDS
	fl[0] = (q.lower[b1] <= (double)(-1 << 24))
		? (-1 << 24) : int_tol.ceil(q.lower[b1]);
	fl[1] = (q.lower[b2] <= (double)(-1 << 24))
		? (-1 << 24) : int_tol.ceil(q.lower[b2]);
	fu[0] = (q.upper[b1] >= (double)( 1 << 24))
		? ( 1 << 24) : int_tol.floor(q.upper[b1]);
	fu[1] = (q.upper[b2] >= (double)( 1 << 24))
		? ( 1 << 24) : int_tol.floor(q.upper[b2]);
#else
	fl[0] = fl[1] = (-1 << 24);
	fu[0] = fu[1] = ( 1 << 24);
#endif

	// the initial points (the set S in the paper)
	int ff0 = floor(f[0]);
	int ff1 = floor(f[1]);
	
	add_point(ff0 + 0, ff1 + 0);
	add_point(ff0 + 0, ff1 + 1);
	add_point(ff0 + 1, ff1 + 0);
	add_point(ff0 + 1, ff1 + 1);
	
	p.add_all(ineq);
}


// **************************************************************************
// 
// **************************************************************************
int tworow::separate(constraint &c, const dvec &xstar,
	const std::string &cname)
{
	int n = ray.size();
	dvec a(n);
	
	lfb_type = PXX_TWOROW_INVALID;
	
	if (n < 3)
		return(PXX_TWOROW_RAY_ERROR);

	// compute objective function
	p.obj.set(0.0);
	
	for (int k = 0; k < (int)col.size(); k++) {
		int r = col[k].ray;
		int j = col[k].column;
		
		double s = (col[k].at_upper)
			? (col[k].upper - xstar[j])
			: (xstar[j] - col[k].lower);
		p.obj[r] += col[k].norm * s;
	}
	
	for (int i = 0; i < n; i++) {
		if (p.obj[i] < 0.0)
			p.obj[i] = 0.0;
	}

	if (p.obj.sqnorm() <= PXX_TWOROW_0_OBJ)
		p.obj.set(1.0);
	
	p.set_objective(p.obj);

#ifdef DEBUG_PROCESS
	std::cout << "------------------------------- " << cname << std::endl;
	std::cout << "objective: " << p.obj << std::endl;
#endif

	// row generation loop
	int iter = 0;
	while (1) {
		p::values["tworow"]["lp"].start();
		p.lpopt();
		p::values["tworow"]["lp"].stop();
		
		int stat = p.getstat();
		
		if ((stat != CPX_STAT_OPTIMAL)
		&&  (stat != CPX_STAT_OPTIMAL_INFEAS)) {
			std::cout << "lpopt(): "
				<< p.stat_str(stat) << std::endl;
			std::cout << "note: f = (" << f[0] << ", " << f[1]
				<< ")" << std::endl;
			//p.writeprob("buggy_polar.lp", NULL);
			return(PXX_TWOROW_LP_ERROR);
		}
		
		p.getx(&a[0], 0, n - 1);
		
		p::values["tworow"]["display"].start();
#ifdef DEBUG_DISPLAY
		display(strprintf("%si%02d", cname.c_str(), iter), a);
#endif
		p::values["tworow"]["display"].stop();
		
		p::values["tworow"]["count"].start();
		std::vector<r2_point> v(n);

		for (int j = 0; j < n; j++) {
			v[j].x[0] = f[0] + ray[j].r[0] / a[j];
			v[j].x[1] = f[1] + ray[j].r[1] / a[j];
		}

#ifdef DEBUG_PROCESS	
		for (std::list<cut>::iterator it = ineq.pool.begin();
		it != ineq.pool.end(); it++) {
			std::cout << it->name << ": slack = " << it->slack(a)
				<< std::endl;
		}
		std::cout << "Count..." << std::endl;
#endif
		polytope2 La;
		int in = La.count_interior_pre(v, f, fl, fu, 1, 200);
#ifdef DEBUG_PROCESS	
		if (in == 0)
			std::cout << "Done counting: " << in << std::endl;
		else
			std::cout << "Done counting: " << in
				<< ": (" << La.point[0].x[0]
				<< "," << La.point[0].x[1] << ")" << std::endl;
#endif
		p::values["tworow"]["count"].stop();
		
		if (in == 0)
			break;
		
		if (in < 0) {
			std::cout << "tworow: could no check Lalpha"
				<< std::endl;
			return(PXX_TWOROW_CHECK_FAILED);
		}
		
		if (iter >= PXX_TWOROW_MAX_ITER) {
			std::cout << "tworow: iteration limit hit"
				<< std::endl;
			return(PXX_TWOROW_CHECK_FAILED);
		}

		p::values["tworow"]["add"].start();
		add_point(La.point[0].x);

		p.add_purge(ineq, a,
			PXX_TWOROW_PT_MIN_VIOL, PXX_TWOROW_PT_MAX_SLACK,
			PXX_TWOROW_PT_MAX_AGE);
		
		p::values["tworow"]["add"].stop();

		if (iter + 1 > p::values["tworow"]["iter_max"])
			p::values["tworow"]["iter_max"] = iter + 1;

		iter++;
	}
	
	// get bounds, generate splits
	p::values["tworow"]["split"].start();
	std::vector<int> cstat(n);
	std::vector<int> rstat(p.getnumrows());
	bool tight_bound = false;
	int split_j = -1;
	
	p.getbase(&cstat[0], &rstat[0]);

#ifndef PXX_SPLIT2_DISABLE
	for (int j = 0; j < n; j++) {
		if (cstat[j] == CPX_AT_LOWER) {
			tight_bound = true;
			
			if (split(a, j) == 0) {
				split_j = j;
				break;
			}
		}
	}
#endif

	p::values["tworow"]["split"].stop();
	if (tight_bound) {
		if (split_j == -1) {
#ifdef DEBUG_SPLIT
			std::cout << "Split: Failed" << std::endl;
#endif
			return(PXX_TWOROW_TIGHT_BOUND);
		} else {
			lfb_type = PXX_TWOROW_SPLIT;
#ifdef DEBUG_SPLIT
			std::cout << "Split: on r" << split_j << std::endl;
#endif
#ifdef DEBUG_DISPLAY
			display(strprintf("%ss%d", cname.c_str(), split_j), a);
#endif
		}
	} else {
		lfb_type = PXX_TWOROW_VALID;
#ifdef DEBUG_SPLIT
		std::cout << "Split: no tight bound" << std::endl;
#endif
	}
	
	/*if (split_j == -1) {
		for (int j = 0; j < n; j++) {
			if (a[j] >= (1.0 / PXX_TWOROW_MIN_WIDTH))
				return(PXX_TWOROW_F_ON_BORDER);
		}
	}*/
	
	// export result to original space
	p::values["tworow"]["export"].start();
	c.name = cname;
	c.row.clear();
	c.row.resize(orig_n);
	c.sense = 'G';
	c.rhs = 1.0;
	
#ifdef DEBUG_EXPAND
	std::cout << "Basis is x" << basic1 << ", x" << basic2 << std::endl;
#endif

	for (int ko = 0; ko < (int)col.size(); ko++) {
		int k = col_order[ko];
		int j = col[k].column;
		int r = col[k].ray;
		
		double aj;
		
		if (col[k].at_upper) {
			aj = -a[r] * col[k].norm;
			/*if ((aj < -1.0) && (ctype[j] != 'C'))
				aj = -1.0;*/
#ifdef DEBUG_EXPAND
			std::cout << "x" << j << ": alpha" << r << " = "
				<< a[r] << " becomes " << aj
				<< "; rhs += " << aj * col[k].upper
				<< " (ub = " << col[k].upper
				<< ")" << std::endl;
#endif
			c.rhs += aj * col[k].upper;
		} else {
			aj =  a[r] * col[k].norm;
			/*if ((aj > 1.0) && (ctype[j] != 'C'))
				aj = 1.0;*/

#ifdef DEBUG_EXPAND
			std::cout << "x" << j << ": alpha" << r << " = "
				<< a[r] << " becomes " << aj
				<< "; rhs += " << aj * col[k].lower
				<< " (lb = " << col[k].lower
				<< ")" << std::endl;
#endif
			c.rhs += aj * col[k].lower;
		}

		// Push also almost-zero values. We want the dynamism
		// test to catch these.
		c.row.push(j, aj, p::tolerance(0.0));
	}
	p::values["tworow"]["export"].stop();

	return(PXX_TWOROW_OK);
}


// **************************************************************************
// 
// **************************************************************************
bool tworow::ray_in_span(double s[2], int i1, int i2, const double y[2],
	bool force)
{
	if (ray[i1].flags & PXX_TWOROW_RAY_NOCONE) {
		std::cout << "WARNING: flag RAY_NOCONE detected" << std::endl;
		return(false);
	}
	
	if (solve2x2(s, ray[i1].r, ray[i2].r, y)) {
		std::cout << "WARNING: singular ray matrix (r" << i1
			<< ", r" << i2 << ")" << std::endl;
		return(false);
	}
	
	if ((!force) && ((s[0] < 0.0) || (s[1] < 0.0))) {
		//std::cout << "[" << s[0] << ", " << s[1] << "]" << std::endl;
		return(false);
	}
	
	// clamp to zero
	// (amounts to strengthening the constraint, provided s1,s2 >= 0)
	if (s[0] < PXX_TWOROW_ZERO_S)
		s[0] = 0.0;

	if (s[1] < PXX_TWOROW_ZERO_S)
		s[1] = 0.0;
	
	return(true);
}

// **************************************************************************
// 
// **************************************************************************
static bool angle_between(double a, double al, double au)
{
	if (al <= au)
		return((a >= al) && (a <= au));
	else
		return((a >= al) || (a <= au));
}

bool tworow::find_span(double s[2], int &i1_r, int &i2_r, const int x[2])
{
	double y[2];
	int n = ray.size();
	
	y[0] = x[0] - f[0];
	y[1] = x[1] - f[1];
	
	if (n <= 20) {
		for (int i1 = 0; i1 < n; i1++) {
			int i2 = (i1 + 1) % n;
			
			if (ray_in_span(s, i1, i2, y)) {
				i1_r = i1;
				i2_r = i2;
				return(true);
			}
		}
		
		std::cout << "find_span(): O(n): WARNING: no span for ("
			<< x[0] << "," << x[1] << ")" << std::endl;
		return(false);
	}
	
	double angle = atan2_2pi(y[1], y[0]);
	
	int i1 = 0;
	int i2 = n;
	
	while (i2 > i1 + 1) {
		int im = (i1 + i2) / 2;
		
		if (angle_between(angle, ray[i1].angle, ray[im].angle)) {
			i2 = im;
		} else {
			i1 = im;
		}
	}
	
	i2 %= n;

	if (ray_in_span(s, i1, i2, y, true)) {
		i1_r = i1;
		i2_r = i2;
		return(true);
	}
	
	std::cout << "find_span(): O(log(n)): WARNING: no span for ("
		<< x[0] << "," << x[1] << ")" << std::endl;
	return(false);
}

// **************************************************************************
// 
// **************************************************************************
void tworow::add_point(const int x[2])
{
	double rhs = 1.0;
	
	for (int p = 0; p < (int)point.size(); p++) {
		if ((point[p].x[0] == x[0]) && (point[p].x[1] == x[1])) {
#ifdef DEBUG_PROCESS
			std::cout << "add_point(): strengthening" << std::endl;
#endif
			rhs = point[p].ineq->rhs *= PXX_TWOROW_PT_RHS_MUL;
			break;
		}
	}

	double s[2];
	int i1, i2;
	
	if (!find_span(s, i1, i2, x))
		return;
	
	int idx = ineq.pool.size();	
	cut &c = ineq.add();

	format(c.name, "pt%d", idx);
	c.row[i1] = s[0];
	c.row[i2] = s[1];
	c.sense = 'G';
	c.rhs = rhs;
	
	point.resize(idx + 1);
	point[idx].x[0] = x[0];
	point[idx].x[1] = x[1];
	point[idx].s[0] = s[0];
	point[idx].s[1] = s[1];
	point[idx].ray[0] = i1;
	point[idx].ray[1] = i2;
	point[idx].ineq = &c;

#ifdef DEBUG_PROCESS
	std::cout << c.name << ": " << s[0] << " s" << i1 << " + "
		<< s[1] << " s" << i2 << " >= " << rhs << std::endl;
#endif
}

void tworow::add_point(int x1, int x2)
{
	int x[2];
	
	x[0] = x1;
	x[1] = x2;
	add_point(x);
}

// **************************************************************************
// 
// **************************************************************************
int tworow::split(dvec &alpha_r, int sj)
{
	int n = ray.size();

	// Rationalize splitting direction
	double slope;
	rational rr;
	double pi[2];
	
	if (fabs(ray[sj].r[0]) < fabs(ray[sj].r[1])) {
		slope = ray[sj].r[0] / ray[sj].r[1];
		rr.approx(slope, PXX_SPLIT2_COEF_MAX);
		pi[0] =  (double)rr.den;
		pi[1] = -(double)rr.num;
	} else {
		slope = ray[sj].r[1] / ray[sj].r[0];
		rr.approx(slope, PXX_SPLIT2_COEF_MAX);
		pi[0] =  (double)rr.num;
		pi[1] = -(double)rr.den;
	}
	
#ifdef DEBUG_SPLIT
	std::cout << "Split: (" << ray[sj].r[0] << ", " << ray[sj].r[1]
		<< ") approximated by (" << -pi[1] << ", " << pi[0]
		<< ")" << std::endl;
#endif
	if (fabs(slope - rr.get_double()) > PXX_SPLIT2_SLOPE_TOL) {
#ifdef DEBUG_SPLIT
		std::cout << "Split: Rational approx failed: "
			<< slope << " != " << rr.get_double() << std::endl;
#endif
		return(-1);
	}
	
	// Compute -frac(pi^T f) and 1-frac(pi^T f)
	double pif = pi[0] * f[0] + pi[1] * f[1];
	double frac = pif - floor(pif);
	double frac_minus = -frac;
	double frac_plus =  1.0 - frac;
	
	if ((frac <= PXX_SPLIT2_MIN_FRAC)
	||  (frac >= 1.0 - PXX_SPLIT2_MIN_FRAC)) {
#ifdef DEBUG_SPLIT
		std::cout << "Split: Fractional test failed" << std::endl;
#endif
		return(-1);
	}
	
	p::dvec alpha(n);

	for (int j = 0; j < n; j++) {
		double pirj = pi[0] * ray[j].r[0] + pi[1] * ray[j].r[1];
		
		if (pirj < 0.0) {
			alpha[j] = pirj / frac_minus;
		} else {
			alpha[j] = pirj / frac_plus;
		}
		
		if (!(ray[j].flags & PXX_TWOROW_RAY_ART)
		&&  (alpha[j] >= (1.0 / PXX_SPLIT2_MIN_WIDTH))) {
#ifdef DEBUG_SPLIT
			std::cout << "Split on r" << sj
				<< ": Too thin (alpha" << j << " = "
				<< alpha[j] << ")" << std::endl;
#endif
			return(-1);
		}
		
		if (alpha[j] <= PXX_SPLIT2_ZERO_ALPHA) {
			alpha[j] = 0.0;
		}
	}

	alpha_r = alpha;
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
void tworow::display(const std::string &cname, const dvec &a)
{
	std::cout << "DISPLAY " << cname << std::endl;

	std::string fname = "out/" + cname + ".svg";
	
	FILE *fp = fopen(fname.c_str(), "w");

	if (!fp)
		return;
	
	int n = ray.size();
	
	double ff[2];
	
	ff[0] = f[0] - floor(f[0]);
	ff[1] = f[1] - floor(f[1]);
	
	double gw = 8.0;
	double gh = 8.0;
	double gx0 = -(gw - 1.0) / 2.0;
	double gy0 = -(gh - 1.0) / 2.0;
	double gx1 = 1.0 + (gw - 1.0) / 2.0;
	double gy1 = 1.0 + (gh - 1.0) / 2.0;
	
	fprintf(fp,
		"<?xml version=\"1.0\" standalone=\"no\"?>\n"
		"<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \n"
		"  \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n"
		"<svg width=\"%g\" height=\"%g\" viewBox=\"%g %g %g %g\" "
			"version=\"1.1\"\n"
		"     xmlns=\"http://www.w3.org/2000/svg\">\n",
		640.0, 640.0, gx0, -gy1, gw, gh);

	fprintf(fp,
		"  <defs>\n"
		"    <marker id=\"Triangle\"\n"
		"      viewBox=\"0 0 10 10\" refX=\"7\" refY=\"5\" \n"
		"      markerUnits=\"strokeWidth\"\n"
		"      markerWidth=\"8\" markerHeight=\"8\"\n"
		"      orient=\"auto\">\n"
		"      <path d=\"M 0 2 L 10 5 L 0 8 L 3 5 z\" />\n"
		"    </marker>\n"
		"  </defs>\n\n");
	
	fprintf(fp, "<g stroke-width=\"%g\">\n\n", gw / 600.0);
	
	// Compute v
	double v[n][2];
	
	for (int i = 0; i < n; i++) {
		if (a[i] < 1.0e-3) {
			v[i][0] = ff[0];
			v[i][1] = ff[1];
		} else {
			v[i][0] = ff[0] + ray[i].r[0] / a[i];
			v[i][1] = ff[1] + ray[i].r[1] / a[i];
		}
	}
	
	// L_a
	fprintf(fp, "<g fill=\"#ddd\" stroke=\"black\">\n");
	
	fprintf(fp, "<path d=\"");
	for (int i0 = 0; i0 < n; i0++) {
		int i1 = (i0 + 1) % n;
		int i2 = (i0 + 2) % n;
		
		if (a[i1] < 1e-3) {
			fprintf(fp, "%c %g %g L %g %g ",
				(i0 == 0) ? 'M' : 'L',
				v[i0][0] + ray[i1].r[0] * 100.0,
				-(v[i0][1] + ray[i1].r[1] * 100.0),
				v[i2][0] + ray[i1].r[0] * 100.0,
				-(v[i2][1] + ray[i1].r[1] * 100.0));
		} else {
			fprintf(fp, "%c %g %g ",
				(i0 == 0) ? 'M' : 'L',
				v[i1][0], -v[i1][1]);
		}
	}
	fprintf(fp, "z\" />\n");
	fprintf(fp, "</g>\n\n");
	
	// Integer points
	int i0 = floor(gx0);
	int j0 = floor(gy0);
	int i1 = ceil(gx1);
	int j1 = ceil(gy1);
	if (i1 - i0 > 40) {
		i0 = -20;
		i1 =  20;
	}
	if (j1 - j0 > 40) {
		j0 = -20;
		j1 =  20;
	}
	
	// Begin part that requires text
	fprintf(fp, "<g>\n");

	fprintf(fp, "<g >\n");
	for (int j = j0; j < j1; j++)
	for (int i = i0; i < i1; i++) {
		char *pcolor;
		
		if (((double)i >= fl[0] - floor(f[0]))
		&&  ((double)i <= fu[0] - floor(f[0]))
		&&  ((double)j >= fl[1] - floor(f[1]))
		&&  ((double)j <= fu[1] - floor(f[1])))
			pcolor = (char *)"#000";
		else
			pcolor = (char *)"#ff0";
		
		fprintf(fp, "  <circle cx=\"%d\" cy=\"%d\" r=\"%g\" "
			"stroke=\"%s\" fill=\"#aaa\" />\n",
			i, -j, gw / 300.0,
			pcolor);
	}
	fprintf(fp, "</g>\n\n");
	
	// Rays
	for (int i = 0; i < n; i++) {
		double x[2];
		
		x[0] = ff[0] + 2.0 * ray[i].r[0];
		x[1] = ff[1] + 2.0 * ray[i].r[1];
		fprintf(fp, "  <path d=\"M %g %g L %g %g\" "
			"stroke=\"%s\" "
			"marker-end=\"url(#Triangle)\" />\n",
			ff[0], -ff[1], x[0], -x[1],
			(ray[i].flags & PXX_TWOROW_RAY_ART)
				? "green" : "black");
			
		fprintf(fp, "  <g stroke=\"none\" fill=\"black\">\n");
		fprintf(fp, "  <text x=\"%g\" y=\"%g\" font-size=\"%g\">"
			"r%d</text>\n", // "r%d:(%g,%g),a%g</text>",
			x[0], -x[1], gw / 50.0,
			i /*, ray[i].r[0], ray[i].r[1], ray[i].angle*/);
		fprintf(fp, "  </g>\n");
	}
	
	// Message
	fprintf(fp, "<g stroke=\"none\" fill=\"black\">\n");
	fprintf(fp, "  <text x=\"%.5f\" y=\"%.5f\" font-size=\"%g\">"
		"%s</text>\n", gx0 + 0.25, -(gy1 - 0.25),
		gw / 50.0, cname.c_str());
	fprintf(fp, "  <text x=\"0\" y=\"0\" font-size=\"%g\">"
			"(%d, %d)</text>\n", gw / 50.0,
			(int)floor(f[0]), (int)floor(f[1]));
	fprintf(fp, "</g>\n");
	
	// End part that requires text
	fprintf(fp, "</g>\n\n");


	// Constraining points
	fprintf(fp, "<g stroke=\"red\" fill=\"#aaa\">\n");
	for (int i = 0; i < (int)point.size(); i++) {
		double x[2];
		
		x[0] = (double)point[i].x[0] - floor(f[0]);
		x[1] = (double)point[i].x[1] - floor(f[1]);
		fprintf(fp, "  <circle cx=\"%g\" cy=\"%g\" r=\"%g\" />\n",
			x[0], -x[1], gw / 300.0);
	}
	fprintf(fp, "</g>\n\n");

	// Vertices of L_a
	fprintf(fp, "<g stroke=\"green\" fill=\"#aaa\">\n");
	for (int i = 0; i < n; i++) {
		fprintf(fp, "  <circle cx=\"%g\" cy=\"%g\" r=\"%g\" />\n",
			v[i][0], -v[i][1], gw / 300.0);
	}
	fprintf(fp, "</g>\n\n");

	// Point f
	fprintf(fp, "<g stroke=\"blue\" fill=\"#aaa\">\n");
	fprintf(fp, "  <circle cx=\"%g\" cy=\"%g\" r=\"%g\" />\n",
		ff[0], -ff[1], gw / 300.0);
	fprintf(fp, "</g>\n\n");
	
	// tail
	fprintf(fp, "</g>\n</svg>\n");

	fclose(fp);
}

// **************************************************************************
// 
// **************************************************************************
#ifdef PXX_TWOROW_BUILD_MIP
double tworow::mip_rhs(const constraint &c)
{
	mip.obj = c.row;
	mip.set_objective(mip.obj);
	mip.mipopt();
	int stat = mip.getstat();
	
	if ((stat != CPXMIP_OPTIMAL) && (stat != CPXMIP_OPTIMAL_INFEAS)
	&&  (stat != CPXMIP_OPTIMAL_TOL)) {
		std::cout << "mip_rhs(): " << mip.stat_str(stat) << std::endl;
		return(-CPX_INFBOUND);
	}
	
	double objval = mip.getobjval();
	
	if (sparse_tol.exclusive_ge(c.rhs, objval)) {
		double x1, x2;
		mip.getx(&x1, basic1, basic1);
		mip.getx(&x2, basic2, basic2);
		std::cout << "cut close to (" << x1 << ", " << x2 << ")"
			<< std::endl;
	}
	return(objval);
}
#endif


// **************************************************************************
// 
// **************************************************************************
int tworow::get_lfb_type()
{
	if ((lfb_type == PXX_TWOROW_SPLIT)
	||  (lfb_type == PXX_TWOROW_INVALID)) {
#ifdef DEBUG_ANALYZE
		std::cout << "TYPE " << lfb_type << std::endl;
#endif
		return(lfb_type);
	}
	
	// get alpha
	int n = ray.size();
	dvec a(n);
	
	p.getx(&a[0], 0, n - 1);
	
	// find rays supporting a vertex
	std::vector<int> vray;
	
	for (int i = 0; i < n; i++) {
		int ci = ray[i].convex_constr;
		if (ci == -1) {
#ifdef DEBUG_ANALYZE
			std::cout << "vertex on r" << i
				<< " (no convexity constr.)" << std::endl;
#endif
			vray.push_back(i);
		} else {
			double slack = 0.0;
			
			p.getslack(&slack, ci, ci);
			slack = -slack;
			
			if (slack > PXX_TWOROW_SLACK_CONSTR) {
#ifdef DEBUG_ANALYZE
				std::cout << "vertex on r" << i
					<< " (slack = " << slack << ")"
					<< std::endl;
#endif
				vray.push_back(i);
			} else {
#ifdef DEBUG_ANALYZE
				std::cout << "       [no vertex on r" << i
					<< " (slack = " << slack << ")]"
					<< std::endl;
#endif
			}
		}
	}
	
	int vs = vray.size();
	
	if (vs < 3) {
#ifdef DEBUG_ANALYZE
		std::cout << "TYPE 0 (" << vs << " vertices)" << std::endl;
#endif
#ifdef ANALYZE_HACKS
		lfb_type = PXX_TWOROW_SPLIT;
#endif
		return(lfb_type);
	}
	
	// count tight points
	std::vector<int> xcount(vs);
	std::fill(xcount.begin(), xcount.end(), 0);
	
	for (int p = 0; p < (int)point.size(); p++) {
		// verify that the point is tight
		if (point[p].ineq->slack(a) > PXX_TWOROW_SLACK_CONSTR)
			continue;
		
		// verify that the point has no later duplicate
		int p2 = p + 1;
		for (; p2 < (int)point.size(); p2++) {
			if ((point[p].x[0] == point[p2].x[0])
			&&  (point[p].x[1] == point[p2].x[1]))
				break;
		}
		
		if (p2 < (int)point.size())
			continue;
		
		// find cone (enlarged, i.e. rays support a vertex each)
		int r0 = point[p].ray[0];
		int vci = -1;
		while(1) {
			for (int v = 0; v < vs; v++) {
				if (r0 == vray[v]) {
					vci = v;
					break;
				}
			}
			
			if (vci != -1)
				break;
			
			r0 = (r0 > 0) ? (r0 - 1) : n - 1;
		}
		
		int vcj = (vci + 1) % vs;
		int r1 = vray[vcj];
		
		// count that point for that cone
#ifdef DEBUG_ANALYZE
		std::cout << " (" << point[p].x[0] << ", " << point[p].x[1]
			<< ") belongs to cone(r" << r0 << ", r" << r1
			<< ")" << std::endl;
#endif
		xcount[vci]++;
		
		// if it is in two cones, count also for the other one
		if ((point[p].ray[0] == r0) && (point[p].s[1] == 0.0)) {
#ifdef DEBUG_ANALYZE
			std::cout << "   also in previous cone" << std::endl;
#endif
			xcount[(vci == 0) ? (vs - 1) : (vci - 1)]++;
		}

		if ((point[p].ray[1] == r1) && (point[p].s[0] == 0.0)) {
#ifdef DEBUG_ANALYZE
			std::cout << "   also in next cone" << std::endl;
#endif
			xcount[(vci + 1) % vs]++;
		}
	}
	
	// determine type
	std::sort(xcount.begin(), xcount.end());
	std::reverse(xcount.begin(), xcount.end());
	
	if (vs == 3) {
		if (xcount[2] >= 3)
			lfb_type = PXX_TWOROW_T1;
		else if ((xcount[0] >= 2) && (xcount[1] >= 1))
			lfb_type = PXX_TWOROW_T2;
		else if (xcount[2] >= 1)
			lfb_type = PXX_TWOROW_T3;
	} else if (vs == 4) {
		if ((xcount[0] >= 2) && (xcount[2] >= 1))
			lfb_type = PXX_TWOROW_Q1;
		else if (xcount[3] >= 1)
			lfb_type = PXX_TWOROW_Q2;
		else if (xcount[1] >= 2)
			lfb_type = PXX_TWOROW_SPLIT;
	}
	
#ifdef DEBUG_ANALYZE
	std::cout << "TYPE " << lfb_type << " (";
	
	for (int v = 0; v < vs; v++) {
		if (v)
			std::cout << " - ";
		
		std::cout << xcount[v];
	}
	
	std::cout << ")" << std::endl;
#endif

#ifdef ANALYZE_HACKS
	if (lfb_type == PXX_TWOROW_VALID) {
		lfb_type = PXX_TWOROW_SPLIT;
	}
#endif
	
	return(lfb_type);
}

// **************************************************************************
// 
// **************************************************************************
/*
  Format:
    "f=" f_0 f_1
    "nr=" nr
    "r=" r0_0 r0_1 ...
    "nv=" nv
    "v=" v0_0 v0_1 ...
    "nt=" nt
    "t=" t0_0 t0_1 ...
*/
int tworow::dump_lfb()
{
	if ((lfb_type == PXX_TWOROW_SPLIT)
	||  (lfb_type == PXX_TWOROW_INVALID)) {
#ifdef DEBUG_ANALYZE
		std::cout << "TYPE " << lfb_type << std::endl;
#endif
		return(lfb_type);
	}

	// get alpha
	int n = ray.size();
	dvec a(n);
	
	p.getx(&a[0], 0, n - 1);
	
	// find rays supporting a vertex
	std::vector<int> vray;
	
	for (int i = 0; i < n; i++) {
		int ci = ray[i].convex_constr;
		if (ci == -1) {
#ifdef DEBUG_ANALYZE
			std::cout << "vertex on r" << i
				<< " (no convexity constr.)" << std::endl;
#endif
			vray.push_back(i);
		} else {
			double slack = 0.0;
			
			p.getslack(&slack, ci, ci);
			slack = -slack;
			
			if (slack > PXX_TWOROW_SLACK_CONSTR) {
#ifdef DEBUG_ANALYZE
				std::cout << "vertex on r" << i
					<< " (slack = " << slack << ")"
					<< std::endl;
#endif
				vray.push_back(i);
			} else {
#ifdef DEBUG_ANALYZE
				std::cout << "       [no vertex on r" << i
					<< " (slack = " << slack << ")]"
					<< std::endl;
#endif
			}
		}
	}
	
	int vs = vray.size();
	
	if ((vs < 3) || (vs > 4)) {
#ifdef DEBUG_ANALYZE
		std::cout << "TYPE 0 (" << vs << " vertices)" << std::endl;
#endif
		return(lfb_type);
	}
	
	std::cout << strprintf("@ f= %.12g %.12g", f[0], f[1]);
	
	int int_col = 0;
	
	for (int k = 0; k < (int)col.size(); k++) {
		if (ctype[col[k].column] != 'C')
			int_col++;
	}
	
	std::cout << strprintf(" nr= %d r=", int_col);

	for (int k = 0; k < (int)col.size(); k++) {
		if (ctype[col[k].column] != 'C') {
			std::cout << strprintf(" %.12g %.12g",
				col[k].r[0], col[k].r[1]);
		}
	}
	
	std::cout << strprintf(" nv= %d v=", vs);
	
	for (int v = 0; v < vs; v++) {
		int j = vray[v];
		
		std::cout << strprintf(" %.12g %.12g",
			f[0] + ray[j].r[0] / a[j],
			f[1] + ray[j].r[1] / a[j]);
	}
	
#ifdef DEBUG_ANALYZE
	std::cout << std::endl;
#endif

	std::vector<z2_point> tp;

	// count tight points
	std::vector<int> xcount(vs);
	std::fill(xcount.begin(), xcount.end(), 0);
	
	for (int p = 0; p < (int)point.size(); p++) {
		// verify that the point is tight
		if (point[p].ineq->slack(a) > PXX_TWOROW_SLACK_CONSTR)
			continue;
		
		// verify that the point has no later duplicate
		int p2 = p + 1;
		for (; p2 < (int)point.size(); p2++) {
			if ((point[p].x[0] == point[p2].x[0])
			&&  (point[p].x[1] == point[p2].x[1]))
				break;
		}
		
		if (p2 < (int)point.size())
			continue;
		
		z2_point tpp;
		tpp.x[0] = point[p].x[0];
		tpp.x[1] = point[p].x[1];
		tp.push_back(tpp);
		
		// find cone (enlarged, i.e. rays support a vertex each)
		int r0 = point[p].ray[0];
		int vci = -1;
		while(1) {
			for (int v = 0; v < vs; v++) {
				if (r0 == vray[v]) {
					vci = v;
					break;
				}
			}
			
			if (vci != -1)
				break;
			
			r0 = (r0 > 0) ? (r0 - 1) : n - 1;
		}
		
		int vcj = (vci + 1) % vs;
		int r1 = vray[vcj];
		
		// count that point for that cone
#ifdef DEBUG_ANALYZE
		std::cout << " (" << point[p].x[0] << ", " << point[p].x[1]
			<< ") belongs to cone(r" << r0 << ", r" << r1
			<< ")" << std::endl;
#endif
		xcount[vci]++;
		
		// if it is in two cones, count also for the other one
		if ((point[p].ray[0] == r0) && (point[p].s[1] == 0.0)) {
#ifdef DEBUG_ANALYZE
			std::cout << "   also in previous cone" << std::endl;
#endif
			xcount[(vci == 0) ? (vs - 1) : (vci - 1)]++;
		}

		if ((point[p].ray[1] == r1) && (point[p].s[0] == 0.0)) {
#ifdef DEBUG_ANALYZE
			std::cout << "   also in next cone" << std::endl;
#endif
			xcount[(vci + 1) % vs]++;
		}
	}
	
	std::cout << strprintf(" nt= %d t=", tp.size());
	
	for (int k = 0; k < (int)tp.size(); k++) {
		std::cout << strprintf(" %d %d", tp[k].x[0], tp[k].x[1]);
	}
	
	std::cout << std::endl;
	
	// determine type
	std::sort(xcount.begin(), xcount.end());
	std::reverse(xcount.begin(), xcount.end());
	
	if (vs == 3) {
		if (xcount[2] >= 3)
			lfb_type = PXX_TWOROW_T1;
		else if ((xcount[0] >= 2) && (xcount[1] >= 1))
			lfb_type = PXX_TWOROW_T2;
		else if (xcount[2] >= 1)
			lfb_type = PXX_TWOROW_T3;
	} else if (vs == 4) {
		if ((xcount[0] >= 2) && (xcount[2] >= 1))
			lfb_type = PXX_TWOROW_Q1;
		else if (xcount[3] >= 1)
			lfb_type = PXX_TWOROW_Q2;
		else if (xcount[1] >= 2)
			lfb_type = PXX_TWOROW_SPLIT;
	}
	
#ifdef DEBUG_ANALYZE
	std::cout << "TYPE " << lfb_type << " (";
	
	for (int v = 0; v < vs; v++) {
		if (v)
			std::cout << " - ";
		
		std::cout << xcount[v];
	}
	
	std::cout << ")" << std::endl;
#endif

	return(lfb_type);
}

// **************************************************************************
// 
// **************************************************************************

} // namespace

