/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef PXX_TWOROW_HPP
#define PXX_TWOROW_HPP
#include "dense_vector.hpp"
#include "model.hpp"
#include "cut_pool.hpp"
#include "cplex_model.hpp"

namespace p {

// parameters: mip
//#define PXX_TWOROW_BUILD_MIP

// parameters: polar
#define PXX_TWOROW_ANGLE_TOL	1e-6
#define PXX_TWOROW_SINGULAR_DET	1e-6
#define PXX_TWOROW_ZERO_S	0.01
#define PXX_TWOROW_ALPHA_LB	(1.0 / 500.0)
#define PXX_TWOROW_MIN_WIDTH	1e-3
#define PXX_TWOROW_PT_RHS_MUL	1.03125
#define PXX_TWOROW_PT_MIN_VIOL	-1e-4
#define PXX_TWOROW_PT_MAX_SLACK	1.0e-2
#define PXX_TWOROW_SLACK_CONSTR	1e-5
#define PXX_TWOROW_PT_MAX_AGE	5
#define PXX_TWOROW_MAX_ITER	1000
#define PXX_TWOROW_0_OBJ	1e-4

// parameters: two-row split
//#define PXX_SPLIT2_DISABLE
#define PXX_SPLIT2_COEF_MAX	256
#define PXX_SPLIT2_SLOPE_TOL	1e-2
#define PXX_SPLIT2_MIN_FRAC	1e-4
#define PXX_SPLIT2_MIN_WIDTH	1e-5
#define PXX_SPLIT2_ZERO_ALPHA	1e-10

// return values
#define PXX_TWOROW_OK		0
#define PXX_TWOROW_LP_ERROR	-1
#define PXX_TWOROW_CHECK_FAILED	-2
#define PXX_TWOROW_TIGHT_BOUND	-3
#define PXX_TWOROW_F_ON_BORDER	-4
#define PXX_TWOROW_RAY_ERROR	-5

// lattice-free body types
#define PXX_TWOROW_INVALID	-1
#define PXX_TWOROW_VALID	0
#define PXX_TWOROW_T1		1
#define PXX_TWOROW_T2		2
#define PXX_TWOROW_T3		3
#define PXX_TWOROW_Q1		4
#define PXX_TWOROW_Q2		5
#define PXX_TWOROW_SPLIT	6

// ***************************************************
class tworow_nonbasic {

public:
	bool operator<(const tworow_nonbasic &) const;
	
public:
	double r[2];
	double angle;
	double norm;
	
	int column;
	bool at_upper;
	int ray;
	int init_idx;
	
	double lower, upper;
};

// ***************************************************
#define PXX_TWOROW_RAY_ART	1
#define PXX_TWOROW_RAY_NOCONE	2
struct tworow_ray {
	double r[2];
	double angle;
	int flags;
	int convex_constr;
};

// ***************************************************
struct tworow_point {
	int x[2];
	double s[2];
	int ray[2];
	cut *ineq;
};

// ***************************************************
class tworow {

public:
	tworow();
	tworow(const tworow &src);
	~tworow();
	
	void init(model &q, const dvec &x0,
		const constraint &c1, const constraint &c2, int b1, int b2);
	void clear();
	
	int separate(constraint &c_r, const dvec &xstar,
		const std::string &cname);

	int get_lfb_type();
	int dump_lfb();
	
#ifdef PXX_TWOROW_BUILD_MIP
	double mip_rhs(const constraint &c);
#endif

private:
	void add_point(const int x[2]);
	void add_point(int x1, int x2);
	bool ray_in_span(double s[2], int i1, int i2, const double r[2],
		bool force = false);
	bool find_span(double s[2], int &i1, int &i2, const int x[2]);
	int split(dvec &a, int r);
	void display(const std::string &cname, const dvec &alpha);
	
private:
	std::vector<int> col_order;
	std::vector<tworow_nonbasic> col;
	std::vector<tworow_ray> ray;
	std::vector<tworow_point> point;
	std::vector<char> ctype;
	cplex_model p;
	cut_pool ineq;
#ifdef PXX_TWOROW_BUILD_MIP
	cplex_model mip;
#endif
	double f[2];
	int fl[2], fu[2];
	int basic1, basic2, orig_n;
	int lfb_type;
};



// ***************************************************

} // namespace

#endif

