/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cmath>
#include <cstdlib>
#include <string>
#include "strprintf.hpp"
#include "value.hpp"

namespace p {

// ***************************************************************************
// 
// ***************************************************************************
value::value()
	: id(-1), name(), depth(0),
	has_dbl(false), has_str(false), has_tmr(false), has_sub(false),
	dbl(0.0)
{
}

value::value(int identifier, int d)
	: id(identifier), name(), depth(d),
	has_dbl(false), has_str(false), has_tmr(false), has_sub(false),
	dbl(0.0)
{
}

value::value(const std::string &vname, int d)
	: id(-1), name(vname), depth(d),
	has_dbl(false), has_str(false), has_tmr(false), has_sub(false),
	dbl(0.0)
{
}

value::value(const value &src)
	: id(src.id), name(src.name), depth(src.depth),
	has_dbl(src.has_dbl), has_str(src.has_str), has_tmr(src.has_tmr),
	has_sub(src.has_sub),
	dbl(src.dbl), str(src.str), tmr(src.tmr), l(src.l)
{
}

value::~value()
{
}

// ***************************************************************************
// 
// ***************************************************************************
value &value::operator=(const value &src)
{
	if (src.has_dbl) {
		dbl = src.dbl;
		has_dbl = true;
	}
	
	if (src.has_str) {
		str = src.str;
		has_str = true;
	}
	
	if (src.has_tmr) {
		tmr = src.tmr;
		has_tmr = true;
	}
	
	if (src.has_sub) {
		l = src.l;
		has_sub = true;
	}
	
	return(*this);
}

value &value::operator=(const std::string &v)
{
	has_str = true;
	str = v;
	return(*this);
}

value &value::operator=(const double &v)
{
	has_dbl = true;
	dbl = v;
	return(*this);
}

// ***************************************************************************
// 
// ***************************************************************************
void value::set_name(const std::string &s)
{
	name = s;
}

const std::string &value::get_name() const
{
	return(name);
}

void value::set_id(int identifier)
{
	id = identifier;
}

int value::get_id() const
{
	return(id);
}

void value::set_depth(int d)
{
	depth = d;
}

int value::get_depth() const
{
	return(depth);
}

// ***************************************************************************
// 
// ***************************************************************************
void value::clear()
{
	has_str = has_dbl = has_sub = false;
	dbl = 0.0;
	str.clear();
	tmr.clear();
	l.clear();
}

bool value::is_string() const
{
	return(has_str);
}

bool value::is_double() const
{
	return(has_dbl);
}

bool value::is_timer() const
{
	return(has_tmr);
}

bool value::is_folder() const
{
	return(has_sub);
}

// ***************************************************************************
// 
// ***************************************************************************
const std::string &value::get_string() const
{
	return(str);
}

const char *value::c_str() const
{
	return(str.c_str());
}

double value::get_double() const
{
	return(dbl);
}

const timer &value::get_timer() const
{
	return(tmr);
}

double value::conv_double() const
{
	if (has_dbl) {
		return(dbl);
	}
	
	if (has_str) {
		char *er;
		double d;
		
		d = strtod(str.c_str(), &er);
		return(d);
	}
	
	if (has_tmr) {
		return(tmr.total_s());
	}
	
	return(0.0);
}

std::string value::conv_string() const
{
	if (has_str) {
		return(str);
	}
	
	if (has_dbl) {
		return(strprintf("%g", dbl));
	}
	
	if (has_tmr) {
		return(strprintf("%g", tmr.total_s()));
	}
	
	return("");
}

value::operator double() const
{
	return(conv_double());
}

value::operator std::string() const
{
	return(conv_string());
}

value &value::operator+=(const std::string &v)
{
	has_str = true;
	str += v;
	return(*this);
}

value &value::operator+=(const double &v)
{
	has_dbl = true;
	dbl += v;
	return(*this);
}

value &value::operator-=(const double &v)
{
	has_dbl = true;
	dbl -= v;
	return(*this);
}

value &value::operator*=(const double &v)
{
	has_dbl = true;
	dbl *= v;
	return(*this);
}

value &value::operator/=(const double &v)
{
	has_dbl = true;
	dbl /= v;
	return(*this);
}

value &value::operator++()
{
	has_dbl = true;
	dbl += 1.0;
	return(*this);
}

int value::operator++(int)
{
	int r = dbl;
	has_dbl = true;
	dbl += 1.0;
	return(r);
}

value &value::operator--()
{
	has_dbl = true;
	dbl -= 1.0;
	return(*this);
}

int value::operator--(int)
{
	int r = dbl;
	has_dbl = true;
	dbl -= 1.0;
	return(r);
}


// ***************************************************************************
// 
// ***************************************************************************
void value::start()
{
	has_tmr = true;
	tmr.start();
}

void value::stop()
{
	has_tmr = true;
	tmr.stop();
}

void value::shadow_start()
{
	has_tmr = true;
	tmr.shadow_start();
}

void value::shadow_stop()
{
	has_tmr = true;
	tmr.shadow_stop();
}

double value::elapsed_s() const
{
	return(tmr.elapsed_s());
}

double value::total_s() const
{
	return(tmr.total_s());
}

int value::calls() const
{
	return(tmr.calls());
}

// ***************************************************************************
// 
// ***************************************************************************
value &value::add_entry(int identifier)
{
	has_sub = true;
	l.push_back(value(identifier, depth + 1));
	return(l.back());
}

value &value::add_entry(const std::string &vname)
{
	has_sub = true;
	l.push_back(value(vname, depth + 1));
	return(l.back());
}

value &value::operator[](int identifier)
{
	for (std::list<value>::iterator it = l.begin(); it != l.end(); it++) {
		if (it->get_id() == identifier) {
			return(*it);
		}
	}
	
	return(add_entry(identifier));
}

value &value::operator[](const std::string &vname)
{
	for (std::list<value>::iterator it = l.begin(); it != l.end(); it++) {
		if (it->get_name() == vname) {
			return(*it);
		}
	}
	
	return(add_entry(vname));
}

value &value::operator[](const char *vname)
{
	return(operator[](std::string(vname)));
}

// ***************************************************************************
// 
// ***************************************************************************
std::ostream &operator<<(std::ostream &os, const value &v)
{
	int pr = 0;
	
	if (v.is_string()) {
		os << v.get_string();
		pr++;
	}
	
	if (v.is_double()) {
		os << v.get_double();
		pr++;
	}
	
	if (v.is_timer()) {
		os << v.get_timer();
		pr++;
	}
	
	if (v.is_folder()) {
		int d = v.get_depth();
		if ((d < 0) || (d >= 10))
			d = 10;
		
		std::string tab(d * 2, ' ');
		std::string pre;
		
		for (std::list<value>::const_iterator it = v.l.begin();
		it != v.l.end(); it++) {
			if (it->get_name().empty())
				format(pre, "[%3d] ", it->get_id());
			else
				format(pre, "%-20s ", it->get_name().c_str());
			
			os << std::endl << tab << pre << *it;
		}
		
		pr++;
	}

	if (pr == 0) {
		os << "0";
	}
	
	return(os);
}

// ***************************************************************************
// 
// ***************************************************************************
void value::parse_cmdline(int argc, char **argv)
{
	int lone = 0;
	
	for (int t = 1; t < argc; t++) {
		std::string arg(argv[t]);
		
		if (arg[0] == '-') {
			while (arg[0] == '-')
				arg.erase(0, 1);
			
			for (size_t l = 0; l < arg.size(); l++) {
				if (arg[l] == '-')
					arg[l] = '_';
			}
			
			t++;
			if (t >= argc)
				return;
			
			char *e;
			double v = strtod(argv[t], &e);
			if (*e == 0) {
				operator[] (arg) = v;
			} else {
				operator[] (arg) = argv[t];
			}
		} else {
			operator[] (lone) = arg;
			lone++;
		}
	}
	
	has_dbl = true;
	dbl = lone;
}

// ***************************************************************************
// 
// ***************************************************************************
value values;

// ***************************************************************************
// 
// ***************************************************************************

} // namespace

