/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <algorithm>
#include <cassert>
#include <cstdio>
#include "vp.hpp"


namespace p {

// **************************************************************************
// 
// **************************************************************************
vp::vp(adjacency &a0, dvec &w0)
	: n(a0.vertices()), a(a0), w(w0), f(a0.vertices()),
	pr_min(-1.0), pr_max(2.0),
	flags(PXX_VP_PERTURB_ND | PXX_VP_ENTER_RANDOM
		| PXX_VP_ENTER_ND_EDGE | PXX_VP_ENTER_ND_NODE
		| PXX_VP_ENTER_D_EDGE | PXX_VP_ENTER_D_NODE),
	max_iter(1<<30), max_stalls(1<<30),
	iter(0), iter_nde(0), iter_ndn(0), iter_de(0), iter_dn(0)
{
}

vp::~vp()
{
}


// **************************************************************************
// 
// **************************************************************************
int vp::operator[] (int j) const
{
	return(f[j].value);
}

// **************************************************************************
// 
// **************************************************************************
bool vp::trivial_infeas(int child) const
{
	return(f[child].value == 0);
}

// **************************************************************************
// 
// **************************************************************************
bool vp::H_condition(int root) const
{
	for (forest::it h(f, root); *h != -1; h++)
	if (f[*h].value == 0) {
		for (adjacency::it i(a, *h); *i != -1; i++) {
			if (f[*i].meta == root) {
				if (f[*i].value == 0)
					return(false);
			} else {
				if (f[*i].value == 1)
					return(false);
			}
		}
	}
	
	return(true);
}

// **************************************************************************
// 
// **************************************************************************
double vp::node_rcost(int j)
{
	double r = 0.0;
	
	if (f[j].meta != j)
		return(0.0);
	
	for (forest::it i(f, j); *i != -1; i++) {
		if (f[*i].value) {
			r -= w[*i];
		} else {
			r += w[*i];
		}
	}
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
double vp::edge_rcost(int child)
{
	double r = 0.0;
	int parent = f[child].parent;
	
	if (parent == -1)
		return(0.0);
	
	f.unlink(parent, child);
	
	for (forest::it i(f, child); *i != -1; i++) {
		if (f[*i].value) {
			r -= w[*i];
		} else {
			r += w[*i];
		}
	}
	
	if (f[child].value == 0)
		r *= -1.0;
	
	f.attach(child, parent);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
class vp_nonbasic {
public:
	vp_nonbasic(int opts,
		int i = -1, bool e = false, double v = 0.0, int idx = 0)
		: flags(opts), node(i), edge(e), rcost(v), var_idx(idx) {};
	
	bool operator<(const vp_nonbasic &b) const;
	
	int flags;
	int node;
	bool edge;
	double rcost;
	int var_idx;
};

bool vp_nonbasic::operator<(const vp_nonbasic &b) const
{
	if (flags & PXX_VP_ENTER_LARGEST) {
		if (rcost > b.rcost)
			return(true);
		if (rcost < b.rcost)
			return(false);
	}
	
	if (var_idx < b.var_idx)
		return(true);
	return(false);
};

// **************************************************************************
// 
// **************************************************************************
void vp::include_all()
{
	ex.clear();
}

void vp::exclude(int i, int j)
{
	vp::edge e;
	
	if (i > j)
		std::swap(i, j);
	
	e.i = i;
	e.j = j;
	ex.push_back(e);
}

bool vp::excluded(int i, int j) const
{
	if (i > j)
		std::swap(i, j);
	
	for (int l = 0; l < (int)ex.size(); l++) {
		if ((i == ex[l].i) && (j == ex[l].j))
			return(true);
	}
	
	return(false);
}


// **************************************************************************
// 
// **************************************************************************
void vp::set_flags(int opts)
{
	flags = opts;
}

void vp::set_max_iter(int iter)
{
	max_iter = iter;
}

void vp::set_max_stalls(int iter)
{
	max_stalls = iter;
}

void vp::set_perturbation(double pmin, double pmax)
{
	pr_min = pmin;
	pr_max = pmax;
}

// **************************************************************************
// 
// **************************************************************************
static double drand(double a, double b)
{
	double r = (double)(random() % 65536) / 65535.0;
	
	return(a + r * (b - a));
}

bool vp::iteration()
{
	std::vector<vp_nonbasic> nb;
	
	p::dvec w0(w);
	
	if (flags & PXX_VP_PERTURB_ND) {
		for (int i = 0; i < n; i++)
			w[i] += drand(pr_min, pr_max);
	}
	
	for (int i = 0; i < n; i++) {
		if (f[i].meta == i) {
			nb.push_back(vp_nonbasic(flags,
				i, false, node_rcost(i), i));
		}
		
		if (f[i].parent != -1) {
			nb.push_back(vp_nonbasic(flags,
				i, true, edge_rcost(i),
				n + a.offset(i, f[i].parent)));
		}

		if (flags & PXX_VP_ENTER_RANDOM)
			nb.back().var_idx = random() % 100;
	}
	w = w0;
	
	std::sort(nb.begin(), nb.end());
	
	/*if (nb[0].rcost <= 0.0)
		return(false);*/
	
	// ******************************************************************
	// Nondegenerate improving pivots
	// ******************************************************************
	for (int v = 0; v < n; v++)
	if (nb[v].rcost > 0.0) {
		if ((nb[v].edge) && (flags & PXX_VP_ENTER_ND_EDGE)) {
			// nb variable is an edge
			int i = nb[v].node;
			int j = f[i].parent;
			
			f.unlink(i, j);
			
			bool feas = ((!trivial_infeas(i)) && (H_condition(i)));
			
			if (!feas) {
				f.attach(i, j);
				continue;
			}
			
			f.tree_complement(i);
			f.tree_meta(i, i);
			
			printf("[nd pivot] in e(%4d,%4d) "
				"out x%-4d        ", j, i, i);
			iter_nde++;
			//include_all();
			return(true);
		}
		
		if ((!nb[v].edge) && (flags & PXX_VP_ENTER_ND_NODE)) {
			// nb variable is a (root) node
			int i = nb[v].node;
			
			if (!H_condition(i))
				continue;
			
			if (f[i].first_child != -1) {
				// not singleton: change root
				f.tree_complement(i);
				f.tree_meta(i, f[i].first_child);
				f.reroot(f[i].first_child);
				printf("[nd pivot] in x%-4d        "
					"out x%-4d        ", i, f[i].meta);
			} else {
				// singleton: merge to another tree
				f.complement(i);
				
				adjacency::it j(a, i);
				while ((*j != -1) && (f[*j].value != 0))
					j++;

				assert(*j != -1);
				f.attach(i, *j);
				f[i].meta = f[*j].meta;

				printf("[nd pivot] in x%-4d        "
					"out e(%4d,%4d) ", i, i, *j);
			}
			iter_ndn++;
			//include_all();
			return(true);
		}
	}	

	// ******************************************************************
	// Degenerate "improving" pivots
	// ******************************************************************
	// std::cout << "degenerate pivot: " << std::endl;
	//return(false);
	//std::sort(nb.begin(), nb.end(), vp_order_edges_first);

	/*std::cout << "Reduced costs:" << std::endl;
	for (int v = 0; v < n; v++) {
		if (nb[v].edge) {
			std::cout << "e(" << nb[v].node << ","
				<< f[nb[v].node].parent << "): "
				<< nb[v].rcost << std::endl;
		} else {
			std::cout << "x" << nb[v].node << ": "
				<< nb[v].rcost << std::endl;
		}
	}*/

	if ((flags & PXX_VP_PERTURB_D) || (flags & PXX_VP_UNPERTURB_D)) {
		if (flags & PXX_VP_PERTURB_D) {
			for (int i = 0; i < n; i++)
				w[i] += drand(pr_min, pr_max);
		}
		
		nb.clear();
		
		for (int i = 0; i < n; i++) {
			if (f[i].meta == i) {
				nb.push_back(vp_nonbasic(flags,
					i, false, node_rcost(i), i));
			}
		
			if (f[i].parent != -1) {
				nb.push_back(vp_nonbasic(flags,
					i, true, edge_rcost(i),
					n + a.offset(i, f[i].parent)));
			}

			if (flags & PXX_VP_ENTER_RANDOM)
				nb.back().var_idx = random() % 100;
		}

		std::sort(nb.begin(), nb.end());
		w = w0;
	}
	
	for (int v = 0; v < n; v++)
	if (nb[v].rcost > 0.0) {
		if ((nb[v].edge) && (flags & PXX_VP_ENTER_D_EDGE)) {
			// nb variable is an edge
			int i = nb[v].node;
			int j = f[i].parent;
		
			f.unlink(i, j);
		
			int k;
			if (f[i].value == 0) {
				k = i;
			} else {
				k = f[i].first_child;
				f.reroot(k);
			}
			f.tree_meta(k, k);

			printf("[ d pivot] in e(%4d,%4d) "
				"out x%-4d        ", j, i, k);
			iter_de++;
			//exclude(i, j);
			return(true);
		}

		if ((!nb[v].edge) && (flags & PXX_VP_ENTER_D_NODE)) {
			// nb variable is a (root) node
			int i = nb[v].node;
		
			forest::it t(f, i);
			adjacency::it j;

			while (*t != -1) {
				j.of(a, *t);
				while (*j != -1) {
					if ((f[*j].meta != i)
					&&  (f[*j].value != f[*t].value)
					//&&  (!excluded(*j, *t))
					)
						break;
					j++;
				}
			
				if (*j != -1)
					break;
			
				t++;
			}
		
			if (*t == -1) {
				//std::cout << "Failed to merge trees"
				//	<< std::endl;
				continue;
			}
		
			f.tree_meta(i, f[*j].meta);
			f.attach(*t, *j);

			printf("[ d pivot] in x%-4d        "
				"out e(%4d,%4d) ", i, *t, *j);
			iter_dn++;
			return(true);
		}
	}
	
	printf("[no pivot] Failed to find a pivot           ");
	return(false);
}

// **************************************************************************
// 
// **************************************************************************
double vp::get_objval()
{
	double c = 0.0;
	for (int i = 0; i < n; i++)
		c += (double)f[i].value * w[i];
		
	return(c);
}

int vp::get_iter(int flags)
{
	if (flags == 0)
		return(iter);
	
	int r = 0;
	if (flags & PXX_VP_ENTER_ND_EDGE)
		r += iter_nde;
	if (flags & PXX_VP_ENTER_ND_NODE)
		r += iter_ndn;
	if (flags & PXX_VP_ENTER_D_EDGE)
		r += iter_de;
	if (flags & PXX_VP_ENTER_D_NODE)
		r += iter_dn;
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
void vp::solve()
{
	f.singletons();
	
	iter = 0;
	int stalls = 0;
	double c0 = -1.0;
	double cbest = -1.0;
	bool cont = true;
	while ((cont) && (iter < max_iter) && (stalls <= max_stalls)) {
		printf("%4d: ", iter);
		cont = iteration();

		double c = get_objval();

		printf("obj =%6g\n", c);
		
		if (c > c0) {
			stalls = 0;
			c0 = c;
		} else {
			stalls++;
		}
		
		if (c > cbest) {
			cbest = c;
		}
		
		iter++;
	}
	
	printf("iter degn    obj\n%4d %4d %6g\n",
		iter, iter_de + iter_dn, cbest);
}

// **************************************************************************
// 
// **************************************************************************
} // namespace

