/*
    This file is part of pxx -- modeling tools for optimization
    Copyright (C) 2011-2014  Laurent Poirrier

    libp is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pxx.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <iostream>
#include <cmath>
#include "tolerance.hpp"
#include "z2.hpp"

//#define DISCARD_BOUNDS
//#define DEBUG_Z2

namespace p {

// ***************************************************************************
// 
// ***************************************************************************
int polytope2::count_interior(const std::vector<r2_point> &v,
	double f[2], int fl[2], int fu[2], int max_pt, int max_iter)
{
	point.clear();

	int n = v.size();

	if (n <= 2)
		return(0);
	
	if ((f[0] < (double)fl[0]) || (f[0] > (double)fu[0])
	||  (f[1] < (double)fl[1]) || (f[1] > (double)fu[1])) {
		std::cerr << "check_interior(): f not within range"
			<< std::endl;
		return(-1);
	}

	double xmin = v[0].x[0];
	double xmax = v[0].x[0];
	
	for (int i = 1; i < n; i++) {
		if (v[i].x[0] < xmin)
			xmin = v[i].x[0];
		if (v[i].x[0] > xmax)
			xmax = v[i].x[0];
	}
	
	int ixmin = int_tol.floor(xmin) + 1;
	int ixmax = int_tol.ceil(xmax) - 1;
	
#ifndef DISCARD_BOUNDS
	if (ixmin < fl[0])
		ixmin = fl[0];
	if (ixmax > fu[0])
		ixmax = fu[0];
#endif

#ifdef DEBUG_Z2
	std::cout << "n = " << n << std::endl;
	std::cout << "Domain is [" << fl[0] << "," << fu[0] << "] x ["
		<< fl[1] << "," << fu[1] << "]" << std::endl;

	std::cout << "Range is (R) " << xmin << ", " << xmax << std::endl;
	std::cout << "Range is (Z) " << ixmin << ", " << ixmax << std::endl;
	
	std::cout << "f = (" << f[0] << "," << f[1] << ")" << std::endl;
	std::cout << "Vertices are:" << std::endl;
	for (int i = 0; i < n; i++) {
		std::cout << "(" << v[i].x[0] << "," << v[i].x[1] << ")"
			<< std::endl;
	}
#endif
	
	int cl = 0;
	int cu = 0;
	int iter = 0;

	for (int x = ixmin; x <= ixmax; x++) {
		int iter_l = 0;
		while (!((v[cl].x[0] <= (double)x)
		&&       (v[(cl + 1) % n].x[0] > (double)x))) {
			cl = (cl + 1 == n) ? 0 : cl + 1;
			if (iter_l++ > n) {
				std::cerr << "check_interior(): could not find"
					<< " lower edge!" << std::endl;
				return(-1);
			}
		}
#ifdef DEBUG_Z2
		std::cout << "\tfound lower in " << iter_l << std::endl;
#endif
		int iter_u = 0;
		while (!((v[cu].x[0] >= (double)x)
		&&       (v[(cu + 1) % n].x[0] < (double)x))) {
			cu = (cu == 0) ? n - 1 : cu - 1;
			if (iter_u++ > n) {
				std::cerr << "check_interior(): could not find"
					<< " upper edge!" << std::endl;
				return(-1);
			}
		}
		
#ifdef DEBUG_Z2
		std::cout << "\tfound upper in " << iter_u << std::endl;
#endif
		double yl = v[cl].x[1] +
			(v[(cl + 1) % n].x[1] - v[cl].x[1])
			/ (v[(cl + 1) % n].x[0] - v[cl].x[0])
			* ((double)x - v[cl].x[0]);
		
		double yu = v[cu].x[1] +
			(v[(cu + 1) % n].x[1] - v[cu].x[1])
			/ (v[(cu + 1) % n].x[0] - v[cu].x[0])
			* ((double)x - v[cu].x[0]);
		
		int iyl = int_tol.floor(yl) + 1;
		int iyr = int_tol.ceil(yu) - 1;

#ifndef DISCARD_BOUNDS
		if (iyl < fl[1])
			iyl = fl[1];
		if (iyr > fu[1])
			iyr = fu[1];
#endif
		
#ifdef DEBUG_Z2
		std::cout << "--- line " << x
			<< "; range is " << iyl << ", " << iyr
			<< "(" << yl << ", " << yu << ")" << std::endl;
#endif
		for (int y = iyl; y <= iyr; y++) {
			z2_point p;
			p.x[0] = x;
			p.x[1] = y;
			point.push_back(p);
			
#ifdef DEBUG_Z2
			std::cout << "      col " << y << std::endl;
#endif
			if ((int)point.size() >= max_pt)
				return(point.size());
			
			if (iter >= max_iter) {
#ifdef DEBUG_Z2
				std::cout << iter << " iterations reached"
					<< std::endl;
#endif
				return(-1);
			}
			
			iter++;
		}
	}
	
#ifdef DEBUG_Z2
	std::cout << "Return " << point.size() << std::endl;
#endif
	return(point.size());
}



// ***************************************************************************
// 
// ***************************************************************************
int polytope2::count_interior_pre(const std::vector<r2_point> &v,
	double f[2], int l[2], int u[2], int max_pt, int max_iter)
{
	int n = v.size();
	
	if (n <= 2)
		return(0);
	
	// Compute centroid
	double g[2];

	g[0] = 0.0;
	g[1] = 0.0;
	for (int i = 0; i < n; i++) {
		g[0] += v[i].x[0];
		g[1] += v[i].x[1];
	}
	g[0] /= (double)n;
	g[1] /= (double)n;
	
	// Flip to get g in the first quadrant
	int c[2], sign[2], order;
	
	sign[0] = (g[0] >= f[0]) ? 1 : -1;
	sign[1] = (g[1] >= f[1]) ? 1 : -1;
	
	if (fabs(g[0] - f[0]) >= fabs(g[1] - f[1])) {
		c[0] = 0;
		c[1] = 1;
		order = 1;
	} else {
		c[0] = 1;
		c[1] = 0;
		order = -1;
	}
	
	// build input
	std::vector<r2_point> tv(n);
	double tf[2];
	int tl[2], tu[2];
	
	if (sign[0] * sign[1] * order == 1) {
		for (int i = 0; i < n; i++) {
			tv[i].x[0] = v[i].x[c[0]] * sign[c[0]];
			tv[i].x[1] = v[i].x[c[1]] * sign[c[1]];
		}
	} else {
		for (int i = 0; i < n; i++) {
			tv[i].x[0] = v[n - i - 1].x[c[0]] * sign[c[0]];
			tv[i].x[1] = v[n - i - 1].x[c[1]] * sign[c[1]];
		}
	}
	
	tf[0] = f[c[0]] * sign[c[0]];
	tf[1] = f[c[1]] * sign[c[1]];
	
	if (sign[c[0]] == 1) {
		tl[0] = l[c[0]];
		tu[0] = u[c[0]];
	} else {
		tl[0] = -u[c[0]];
		tu[0] = -l[c[0]];
	}
	
	if (sign[c[1]] == 1) {
		tl[1] = l[c[1]];
		tu[1] = u[c[1]];
	} else {
		tl[1] = -u[c[1]];
		tu[1] = -l[c[1]];
	}
	
	// count
	int r = count_interior(tv, tf, tl, tu, max_pt, max_iter);
	
	// transform output
	for (int p = 0; p < (int)point.size(); p++) {
		int tx[2];
		tx[0] = point[p].x[0];
		tx[1] = point[p].x[1];
		point[p].x[c[0]] = tx[0] * sign[c[0]];
		point[p].x[c[1]] = tx[1] * sign[c[1]];
	}
	
	return(r);
}

// ***************************************************************************
// 
// ***************************************************************************

} // namespace
